/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <assert.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <enatsc3cap/enatsc3cap.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>

#include "enatsc3source.h"
#include "basebandfilesource.h"

static char *ensys = "bb2fmp4";
static char *ensubsys = "bbfilesrc";

#define MAX_PAYLOAD_SIZE 65536

#define MAX_BASEBAND_FILE_SOURCES 32

typedef struct
{
    struct _ENATSC3Source base;

    int32_t fOpened;
    int ifd;
    int32_t fRealtime;
    uint64_t plpIdMask;
    int32_t fPLPIdMaskChanged;

    uint64_t packetNumber;
    uint64_t nBytes;

    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback;
    ENATSC3SourceBasebandPacketCallbackFunction BasebandPacketCallback;
    void *pCallbackUserData;

    pthread_mutex_t lock;
} ENATSC3BasebandFileSource;

struct
{
    int32_t fInitialized;
    ENATSC3BasebandFileSource vSources[MAX_BASEBAND_FILE_SOURCES];

    pthread_mutex_t lock;

} atsc3BasebandFileSourceGlobalsG =
{
    .fInitialized = 0,
};

ENStatus
ATSC3BasebandFileSourceRun(ENATSC3BasebandFileSource *pSource, 
    int32_t *pFRunning)
{
    int64_t tOffsetSeconds;
    int64_t tOffsetNanoseconds;
    int32_t fHaveTimeOffset = 0;
    ENATSC3CapFileHeader fileHeader;
    uint8_t *pPayload = NULL;
    uint64_t plpIdMaskLocal;
    uint64_t llsValidMask = 0;
    int32_t lastCC = -1;
    ENStatus rval = ENStatus_Success;

    pSource->packetNumber = 0;
    pSource->nBytes = 0;

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENStatus_Failure;
        goto out;
    }

    plpIdMaskLocal = pSource->plpIdMask;

    if (read(pSource->ifd, &fileHeader, sizeof(fileHeader)) !=
        sizeof(fileHeader))
    {
        ERRORPRINT("Failed to read file header\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fileHeader.type != ENATSC3CapType_ATSC3_Baseband)
    {
        ERRORPRINT("Unhandled format %d or type %d\n",
            fileHeader.format, fileHeader.type);
        rval = ENStatus_Failure;
        goto out;
    }

    while (*pFRunning)
    {
        ENATSC3CapRecordHeaderV2 header;
        int nb;
        int32_t payloadLength;
        int32_t fCCError = 0;
        struct timespec packetTime;

        nb = read(pSource->ifd, &header, sizeof(header));
        if (nb != sizeof(header))
        {
            if (nb == 0)
            {
                break;
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading rzrcap packet header: %s\n",
                    strerror(errno));
                rval = ENStatus_Failure;
                break;
            }
            else
            {
                ERRORPRINT("Short read of rzrcap packet header (%d != %zd)\n",
                    nb, sizeof(header));
                rval = ENStatus_Failure;
                break;
            }
        }

        header.length = ntohl(header.length);
        header.rtwhen.sec = ENUtils_ntohll(header.rtwhen.sec);
        header.rtwhen.nsec = ntohl(header.rtwhen.nsec);
        header.monowhen.sec = ENUtils_ntohll(header.monowhen.sec);
        header.monowhen.nsec = ntohl(header.monowhen.nsec);

        if (lastCC != -1)
        {
            if (header.continuityCounter != ((lastCC + 1) & 0xff))
            {
                fCCError = 1;
            }
        }
        lastCC = header.continuityCounter;

        if (header.length > MAX_PAYLOAD_SIZE)
        {
            ERRORPRINT("Baseband packet too large (%d > %d)\n",
                header.length, MAX_PAYLOAD_SIZE);
            rval = ENStatus_Failure;
            break;
        }

        payloadLength = header.length;
        packetTime.tv_sec = header.rtwhen.sec;
        packetTime.tv_nsec = header.rtwhen.nsec;

        nb = read(pSource->ifd, pPayload, header.length);
        if (nb != header.length)
        {
            if (nb == 0)
            {
                ERRORPRINT("EOF reading packet payload\n");
            }
            else if (nb < 0)
            {
                ERRORPRINT("Error reading baseband packet: %s\n",
                    strerror(errno));
            }
            else
            {
                ERRORPRINT("Short read of baseband packet (%d != %zd)\n",
                    nb, sizeof(header));
            }

            rval = ENStatus_Failure;
            break;
        }

        if (pSource->fRealtime)
        {
            if (!fHaveTimeOffset)
            {
                struct timespec mononow;

                clock_gettime(CLOCK_MONOTONIC_RAW, &mononow);
                tOffsetSeconds = mononow.tv_sec;
                tOffsetSeconds -= header.monowhen.sec;
                tOffsetNanoseconds = mononow.tv_nsec;
                tOffsetNanoseconds -= header.monowhen.nsec; 

                fHaveTimeOffset = 1;
            }
            else
            {
                struct timespec mononow;
                int64_t delta;

                while (1)
                {
                    clock_gettime(CLOCK_MONOTONIC_RAW, &mononow);
                    delta = tOffsetSeconds - ((int64_t)mononow.tv_sec -
                        (int64_t)header.monowhen.sec);
                    delta *= NANOSECONDS_PER_SECOND;
                    delta += tOffsetNanoseconds - (mononow.tv_nsec -
                        header.monowhen.nsec);

                    if (delta >= NANOSECONDS_PER_SECOND / 100 / 2)
                    {
                        usleep(10000);
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }

        if (header.type == ENATSC3CapPayloadType_Demod_V2)
        {
            uint64_t currentLLSValidMask;
            ENATSC3CapDemodulatorRecordV2 *pDemodRecord;

            pDemodRecord = (ENATSC3CapDemodulatorRecordV2 *)pPayload;
            currentLLSValidMask = ENUtils_ntohll(pDemodRecord->atsc3.llsPresentMask);
            if (currentLLSValidMask != llsValidMask)
            {
                if (pSource->LLSPLPsCallback)
                {
                    pSource->LLSPLPsCallback(currentLLSValidMask, 
                        &packetTime, pSource->packetNumber,
                        pSource->pCallbackUserData);
                }
                llsValidMask = currentLLSValidMask;
            }
            continue;
        }

        if (header.type != ENATSC3CapPayloadType_ATSC3_Baseband)
        {
            ERRORPRINT("Unexpected capture record type %d\n",
                header.type);
            continue;
        }

        // See if the plp mask has changed
        if (pSource->fPLPIdMaskChanged)
        {
            if (pthread_mutex_lock(&pSource->lock))
            {
                ERRORPRINT("Failed to lock\n");
            }

            plpIdMaskLocal = pSource->plpIdMask;
            pSource->fPLPIdMaskChanged = 0;

            if (pthread_mutex_unlock(&pSource->lock))
            {
                ERRORPRINT("Failed to unlock\n");
            }
        }

        if (plpIdMaskLocal & (1ULL << header.plpId))
        {
            if (pSource->BasebandPacketCallback)
            {
                ENATCS3L1Time l1time;

                l1time.precision = header.broadcastTime.L1B_time_info_flag;
                l1time.seconds = htonl(header.broadcastTime.L1D_time_sec);
                l1time.milliseconds = htonl(header.broadcastTime.L1D_time_msec);
                l1time.microseconds = htonl(header.broadcastTime.L1D_time_usec);
                l1time.nanoseconds = htonl(header.broadcastTime.L1D_time_nsec);

                pSource->BasebandPacketCallback(header.plpId, &packetTime, 
                    pSource->packetNumber, header.fError, fCCError, &l1time,
                    pPayload, payloadLength, pSource->pCallbackUserData);
            }
        }

        pSource->packetNumber++;
        pSource->nBytes += payloadLength;
    }

out:
    if (pPayload)
    {
        free(pPayload);
    }

    return rval;
}

ENStatus
ATSC3BasebandFileSourceStart(ENATSC3Source hSource, int32_t *pFRunning)
{
    ENATSC3BasebandFileSource *pBasebandFileSource = 
        (ENATSC3BasebandFileSource *)hSource;
    ENStatus rval = ENStatus_Success;

    rval = ATSC3BasebandFileSourceRun(pBasebandFileSource, pFRunning);

    return rval;
}

ENStatus
ATSC3BasebandFileSourceClose(ENATSC3Source hSource)
{
    ENATSC3BasebandFileSource *pBasebandFileSource = 
        (ENATSC3BasebandFileSource *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&atsc3BasebandFileSourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pBasebandFileSource->fOpened = 0;

    if (pthread_mutex_unlock(&atsc3BasebandFileSourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ATSC3BasebandFileSourceSetPLPIds(ENATSC3Source hSource, uint64_t plpIdMask)
{
    ENATSC3BasebandFileSource *pBasebandFileSource =
        (ENATSC3BasebandFileSource *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pBasebandFileSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pBasebandFileSource->plpIdMask = plpIdMask;
    pBasebandFileSource->fPLPIdMaskChanged = 1;

    if (pthread_mutex_unlock(&pBasebandFileSource->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ATSC3BasebandFileSourceRegisterCallbacks(ENATSC3Source hSource,
    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback,
    ENATSC3SourceBasebandPacketCallbackFunction BasebandPacketCallback,
    ENATSC3SourceALPPacketCallbackFunction ALPPacketCallback,
    void *pUserData)
{
    ENATSC3BasebandFileSource *pBasebandFileSource =
        (ENATSC3BasebandFileSource *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pBasebandFileSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pBasebandFileSource->LLSPLPsCallback = LLSPLPsCallback;
    pBasebandFileSource->BasebandPacketCallback = BasebandPacketCallback;
    pBasebandFileSource->pCallbackUserData = pUserData;

    if (pthread_mutex_unlock(&pBasebandFileSource->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ENATSC3BasebandFileSourceInitialize(int32_t *pNDevices)
{
    int32_t i = 0;
    int32_t fHaveLock = 0;
    ENStatus rval = ENStatus_Success;
    if (pthread_mutex_init(&atsc3BasebandFileSourceGlobalsG.lock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveLock = 1;

    for (i = 0; i < numof(atsc3BasebandFileSourceGlobalsG.vSources); i++)
    {
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.Start =
            ATSC3BasebandFileSourceStart;
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.Close =
            ATSC3BasebandFileSourceClose;
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.SetPLPIds =
            ATSC3BasebandFileSourceSetPLPIds;
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.RegisterCallbacks =
            ATSC3BasebandFileSourceRegisterCallbacks;
        strncpy(atsc3BasebandFileSourceGlobalsG.vSources[i].base.strModel,
            "bbfile",
            sizeof(atsc3BasebandFileSourceGlobalsG.vSources[i].base.strModel) - 1);
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.strModel[sizeof(atsc3BasebandFileSourceGlobalsG.vSources[i].base.strModel) - 1] = 0;
        atsc3BasebandFileSourceGlobalsG.vSources[i].base.strSerial[0] = 0;
        atsc3BasebandFileSourceGlobalsG.vSources[i].fOpened = 0;
        if (pthread_mutex_init(
            &atsc3BasebandFileSourceGlobalsG.vSources[i].lock, NULL))
        {
            ERRORPRINT("Failed to init lock\n");
            for (i--; i >= 0; i--)
            {
                if (pthread_mutex_destroy(
                    &atsc3BasebandFileSourceGlobalsG.vSources[i].lock))
                {
                    ERRORPRINT("Failed to destroy lock\n");
                }
            }
            goto errout;
        }
    }

    *pNDevices = MAX_BASEBAND_FILE_SOURCES;

out:
    return rval;

errout:
    if (fHaveLock)
    {
        if (pthread_mutex_destroy(&atsc3BasebandFileSourceGlobalsG.lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    goto out;
}

ENStatus
ENATSC3BasebandFileSourceShutdown(void)
{
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0; i < numof(atsc3BasebandFileSourceGlobalsG.vSources); i++)
    {
        if (pthread_mutex_destroy(
            &atsc3BasebandFileSourceGlobalsG.vSources[i].lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (pthread_mutex_destroy(&atsc3BasebandFileSourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    return rval;
}

ENStatus
ENATSC3BasebandFileSourceOpen(int32_t deviceNumber, int ifd, int32_t fRealtime,
    ENATSC3Source *ppSource)
{
    ENStatus rval = ENStatus_Success;

    if (deviceNumber >= numof(atsc3BasebandFileSourceGlobalsG.vSources))
    {
        ERRORPRINT("Source %d out of range %d\n", deviceNumber,
            numof(atsc3BasebandFileSourceGlobalsG.vSources));
        rval = ENStatus_Failure;
        goto out;
    }

    if (pthread_mutex_lock(&atsc3BasebandFileSourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (atsc3BasebandFileSourceGlobalsG.vSources[deviceNumber].fOpened)
    {
        ERRORPRINT("Device %d already opened\n", deviceNumber);
        rval = ENStatus_Failure;
        goto out;
    }

    if (lseek(ifd, 0, SEEK_SET) != 0)
    {
        ERRORPRINT("Failed to seek input file: %s\n", strerror(errno));
        rval = ENStatus_Failure;
        goto out;
    }

    atsc3BasebandFileSourceGlobalsG.vSources[deviceNumber].fOpened = 1;
    atsc3BasebandFileSourceGlobalsG.vSources[deviceNumber].ifd = ifd;
    atsc3BasebandFileSourceGlobalsG.vSources[deviceNumber].fRealtime = 
        fRealtime;
    atsc3BasebandFileSourceGlobalsG.vSources[deviceNumber].plpIdMask = 0;
    atsc3BasebandFileSourceGlobalsG.vSources[deviceNumber].fPLPIdMaskChanged = 
        0;
    atsc3BasebandFileSourceGlobalsG.vSources[deviceNumber].LLSPLPsCallback =
        NULL;
    atsc3BasebandFileSourceGlobalsG.vSources[deviceNumber].pCallbackUserData =
        NULL;

    *ppSource = (ENATSC3Source)(atsc3BasebandFileSourceGlobalsG.vSources + deviceNumber);

out:
    if (pthread_mutex_unlock(&atsc3BasebandFileSourceGlobalsG.lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ENATSC3BasebandFileSourceGetDeviceInfo(int32_t deviceNumber,
    char *pModel, int32_t nModelBytes, char *pSerial, int32_t nSerialBytes)
{
    ENATSC3BasebandFileSource *pSource;
    ENStatus rval = ENStatus_Success;

    if (deviceNumber >= numof(atsc3BasebandFileSourceGlobalsG.vSources))
    {
        ERRORPRINT("Source %d out of range %d\n", deviceNumber,
            numof(atsc3BasebandFileSourceGlobalsG.vSources));
        rval = ENStatus_Failure;
        goto out;
    }

    pSource = atsc3BasebandFileSourceGlobalsG.vSources + deviceNumber;
    strncpy(pModel, pSource->base.strModel, nModelBytes - 1);
    pModel[nModelBytes - 1] = 0;
    strncpy(pSerial, pSource->base.strSerial, nSerialBytes - 1);
    pSerial[nSerialBytes - 1] = 0;

out:
    return rval;
}

