/*
 * Copyright © 2021 - 2022 Koherence, LLC (www.koherence.com).
 * All Rights Reserved.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/statvfs.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <assert.h>

#include <curl/curl.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <enatsc3cap/enatsc3cap.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3j288alp/libenatsc3j288alp.h>
#include <libenatsc3alp/libenatsc3alp.h>

#include <libhdhomerun/hdhomerun.h>

#include "enatsc3source.h"
#include "hdhrj288alpsource.h"

static char *ensys = "atsc32mpts";
static char *ensubsys = "j288alphdhrsrc";

#define MAX_SOURCES 32

#define INVALID_PLP_ID 255

#define HDHR_MAX_TUNERS 128
#define DISCOVER_DATA_EXPIRATION_SECONDS 5

static struct 
{
    char *pString;
    int32_t sString;
    ENATSC3Modulation mod;

} vModTableG[] = 
{
    {"qpsk", sizeof("qpsk") - 1, ENATSC3Modulation_QPSK},
    {"qam16", sizeof("qam16") - 1, ENATSC3Modulation_16QAM},
    {"qam64", sizeof("qam64") - 1, ENATSC3Modulation_64QAM},
    {"qam256", sizeof("qam256") - 1, ENATSC3Modulation_256QAM},
    {"qam1024", sizeof("qam1024") - 1, ENATSC3Modulation_1024QAM},
    {"qam4096", sizeof("qam4096") - 1, ENATSC3Modulation_4096QAM},
};

static struct 
{
    char *pString;
    int32_t sString;
    ENATSC3Coding cod;

} vCodTableG[] = 
{
    {"2/15", sizeof("2/15") - 1, ENATSC3Coding_2_15},
    {"3/15", sizeof("3/15") - 1, ENATSC3Coding_3_15},
    {"4/15", sizeof("4/15") - 1, ENATSC3Coding_4_15},
    {"5/15", sizeof("5/15") - 1, ENATSC3Coding_5_15},
    {"6/15", sizeof("6/15") - 1, ENATSC3Coding_6_15},
    {"7/15", sizeof("7/15") - 1, ENATSC3Coding_7_15},
    {"8/15", sizeof("8/15") - 1, ENATSC3Coding_8_15},
    {"9/15", sizeof("9/15") - 1, ENATSC3Coding_9_15},
    {"10/15", sizeof("10/15") - 1, ENATSC3Coding_10_15},
    {"11/15", sizeof("11/15") - 1, ENATSC3Coding_11_15},
    {"12/15", sizeof("12/15") - 1, ENATSC3Coding_12_15},
    {"13/15", sizeof("13/15") - 1, ENATSC3Coding_13_15},
};

typedef struct
{
    struct _ENATSC3Source base;

    int32_t fOpened;
    int32_t fRunning;

    ENVerbosity verbosity;

    struct hdhomerun_discover_device_t hdhr;

    // the tuner instance on the HDHR
    int32_t tunerNum;

    ENATSC3J288ALPHDHRSourceOperatingMode operatingMode;

    int32_t frequencyKHz;
    int32_t rfChannelNumber;

    // The currently tuned PLPs and their slot assignemnts. Most ATSC 3 demods
    // can handle only 4 PLPs, as that is all that is required by spec.
    uint8_t vPLPIds[4];
    uint64_t plpIdMask;
    uint64_t requestedPlpIdMask;
    int32_t singlePLPId;

    int32_t fPLPIdMaskChanged;
    pthread_mutex_t plpIdMaskChangedLock;

    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback;
    ENATSC3SourceALPPacketCallbackFunction ALPPacketCallback;
    void *pCallbackUserData;

    pthread_t transferThread;
    pthread_t plpStatusThread;

    ENATSC3J288ALPHandle_t hATSC3J288ALP;

    struct _stats
    {
        uint8_t masterLock;
        uint64_t plpValidMask;
        uint64_t plpLockedMask;
        uint64_t llsValidMask;
        int32_t snr;
        int32_t rssi;

        // ALP statistics
        uint64_t nALPPackets;

        struct _plpStats
        {
            ENATSC3Modulation modulation;
            ENATSC3Coding coding;

            // PLP ALP packets received
            uint32_t nALPPackets;
        } vPLPStats[64];
    } stats;

    pthread_mutex_t statsLock;

    ENATSC3J288PTPInfo ptpInfo;
    int32_t fHavePTPInfo;
    pthread_mutex_t ptpInfoLock;

    int curlSocket;
    pthread_mutex_t curlSocketLock;

    pthread_mutex_t lock;
} ENATSC3J288ALPHDHRSource;

struct
{
    int32_t fInitialized;

    ENATSC3J288ALPHDHRSource vSources[MAX_SOURCES];
    int32_t nSources;
} atsc3J288ALPHDHRSourceGlobalsG =
{
    .fInitialized = 0,
};

void 
ATSC3J288ALPHDHRSourcePTPCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, ENATSC3J288PTPInfo *pPTPInfo, void *pCbUserData)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)pCbUserData;

    if (pthread_mutex_lock(&pSource->ptpInfoLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pSource->ptpInfo = *pPTPInfo;
    pSource->fHavePTPInfo = 1;

    if (pthread_mutex_unlock(&pSource->ptpInfoLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }
}

void
ATSC3J288ALPHDHRSourceALPPacketCallback(uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity, const uint8_t *pPacket, 
    int32_t sPacket, void *pCbUserData)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)pCbUserData;

    // If only one PLP is requested then the sony demod does not include the 
    // extension header indicating the source PLP Id, so it arrives as 255.
    // We need to figure it out here.
    if (plpId == INVALID_PLP_ID)
    {
        if (pSource->singlePLPId != INVALID_PLP_ID)
        {
            plpId = pSource->singlePLPId;
        }
    }

    pSource->stats.nALPPackets++;
    if (plpId < numof(pSource->stats.vPLPStats))
    {
        pSource->stats.vPLPStats[plpId].nALPPackets++;
    }

    if (pSource->ALPPacketCallback)
    {
        pSource->ALPPacketCallback(plpId, pPacketTimeUTC, packetNumber,
            fError, fDiscontinuity, NULL, pPacket, sPacket,
            pSource->pCallbackUserData);

    }
}

static curl_socket_t 
storeCurlSocket(void *pCbUserData, curlsocktype purpose, 
    struct curl_sockaddr *addr) 
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)pCbUserData;

    if (pthread_mutex_lock(&pSource->curlSocketLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (pSource->fRunning)
    {
        pSource->curlSocket = socket(addr->family, addr->socktype, 
            addr->protocol);
    }
    else
    {
        ERRORPRINT("NOT RUNNING\n");
    }

    if (pthread_mutex_unlock(&pSource->curlSocketLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return pSource->curlSocket;
}

static void
closeCurlSocket(ENATSC3J288ALPHDHRSource *pSource)
{
    if (pthread_mutex_lock(&pSource->curlSocketLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (pSource->curlSocket != -1)
    {
        close(pSource->curlSocket);
        pSource->curlSocket = -1;
    }

    if (pthread_mutex_unlock(&pSource->curlSocketLock))
    {
        ERRORPRINT("Failed to lock\n");
    }
}

static size_t
curlWriteFunction(void *ptr, size_t size, size_t nmemb, void *pCbUserData)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)pCbUserData;
    struct timespec utcnow;
    size_t rval = size * nmemb;

    clock_gettime(CLOCK_REALTIME, &utcnow);
    if (ENATSC3J288ALPProcessJ288Packets(pSource->hATSC3J288ALP, 
        &utcnow, (uint8_t *)ptr, size * nmemb))
    {
        ERRORPRINT("Failed to process J288 packets\n");
    }

    return rval;
}

void *
ATSC3J288ALPHDHRSourceTransferThread(void *arg)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)arg;
    char strPLPIds[32] = "";
    int32_t off = 0;
    CURLcode res;
    char strError[CURL_ERROR_SIZE];
    char strURL[128];
    CURL *curl = NULL;
    int32_t i;

    for (i = 0; i < numof(pSource->vPLPIds); i++)
    {
        if (pSource->vPLPIds[i] != INVALID_PLP_ID)
        {
            int ret;

            ret = snprintf(strPLPIds + off, sizeof(strPLPIds) - off, "p%d",
                pSource->vPLPIds[i]);
            if (ret < 0)
            {
                ERRORPRINT("Error adding PLP %d\n", pSource->vPLPIds[i]);
                goto out;
            }
            off += ret;
            if (off >= sizeof(strPLPIds))
            {
                ERRORPRINT("PLP ID string too long\n");
                goto out;
            }
        }
    }
    
    if (strPLPIds[0] == 0)
    {
        strcpy(strPLPIds, "p0");
    }

    if (snprintf(strURL, sizeof(strURL),
        "http://%d.%d.%d.%d:5004/tuner%d/ch%d%s?duration=311040000&format=j288",
        (pSource->hdhr.ip_addr >> 24) & 0xff, 
        (pSource->hdhr.ip_addr >> 16) & 0xff, 
        (pSource->hdhr.ip_addr >> 8) & 0xff, 
        (pSource->hdhr.ip_addr) & 0xff, 
        pSource->tunerNum, pSource->rfChannelNumber, 
        strPLPIds) >=
        sizeof(strURL))
    {
        ERRORPRINT("URL too long\n");
        goto out;
    }

    curl = curl_easy_init();
    if (!curl)
    {
        ERRORPRINT("Out of memory\n");
        goto out;
    }

    if (curl_easy_setopt(curl, CURLOPT_OPENSOCKETFUNCTION, storeCurlSocket) != 
        CURLE_OK)
    {
        ERRORPRINT("Failed to set opensocketfunction\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_OPENSOCKETDATA, pSource) != CURLE_OK)
    {
        ERRORPRINT("Failed to set opensocketdata\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_URL, strURL) != CURLE_OK)
    {
        ERRORPRINT("Failed to set URL\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_USERAGENT, "encurl/7.29.0") != CURLE_OK)
    {
        ERRORPRINT("Failed to set useragent\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_ACCEPT_ENCODING, "") != CURLE_OK)
    {
        ERRORPRINT("Failed to set encoding\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_HTTPGET, 1L) != CURLE_OK)
    {
        ERRORPRINT("Failed to set httpget\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1) != 
        CURLE_OK)
    {
        ERRORPRINT("Failed to set httpver\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_BUFFERSIZE, 1400 * 188) != CURLE_OK)
    {
        ERRORPRINT("Failed to set buffersize\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curlWriteFunction) != 
        CURLE_OK)
    {
        ERRORPRINT("Failed to set writefunc\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_WRITEDATA, pSource) != CURLE_OK)
    {
        ERRORPRINT("Failed to set userdata\n");
        goto out;
    }
    if (curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, strError) != CURLE_OK)
    {
        ERRORPRINT("Failed to set errorbuffer\n");
        goto out;
    }

    if (ENATSC3J288ALPOpen(&pSource->hATSC3J288ALP))
    {
        ERRORPRINT("Failed to open J288 processor\n");
        goto out;
    }

    if (ENATSC3J288ALPRegisterCallbacks(pSource->hATSC3J288ALP,
        ATSC3J288ALPHDHRSourceALPPacketCallback, 
        NULL,
        ATSC3J288ALPHDHRSourcePTPCallback, pSource))
    {
        ERRORPRINT("Failed to register J288 callbacks\n");
    }

    res = curl_easy_perform(curl);
    if (res != CURLE_OK)
    {
        int curlSocket;

        if (pthread_mutex_lock(&pSource->curlSocketLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        curlSocket = pSource->curlSocket;

        if (pthread_mutex_unlock(&pSource->curlSocketLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }

        if (curlSocket != -1)
        {
            ERRORPRINT("CURL %s failed with %d (%s)\n", strURL, res, 
                curl_easy_strerror(res));
        }
    }

out:
    if (curl)
    {
        curl_easy_cleanup(curl);
    }

    if (pSource->hATSC3J288ALP)
    {
        if (ENATSC3J288ALPClose(pSource->hATSC3J288ALP))
        {
            ERRORPRINT("Failed to close J288 processor\n");
        }
        pSource->hATSC3J288ALP = NULL;
    }

    return NULL;
}

void *
PLPStatusThread(void *arg)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)arg;
    struct hdhomerun_device_t *pHDHomerun = NULL;
    char strPLPInfoPath[32];
    char strStatusPath[32];
    char strHDHR[32];
    int32_t hdhrErrorCount = 0;

    if (snprintf(strPLPInfoPath, sizeof(strPLPInfoPath), "/tuner%d/plpinfo",
        pSource->tunerNum) >= sizeof(strPLPInfoPath))
    {
        ERRORPRINT("Path too long\n");
        goto out;
    }

    if (snprintf(strStatusPath, sizeof(strStatusPath), "/tuner%d/status",
        pSource->tunerNum) >= sizeof(strStatusPath))
    {
        ERRORPRINT("Path too long\n");
        goto out;
    }

    if (snprintf(strHDHR, sizeof(strHDHR), "%s-%d", pSource->base.strSerial,
        pSource->tunerNum) >= sizeof(strHDHR))
    {
        ERRORPRINT("HDHR name too long\n");
        goto out;
    }

    while (pSource->fRunning)
    {
        char *pStatusString;
        char *pLockStatus;
        int32_t masterLock = 0;
        char *pPLPInfoString;
        char *pPLPInfoIter;
        char *pLLS;
        char *pPLPLock;
        char *pPLPMod;
        char *pPLPCod;
        ENATSC3Modulation vPLPMod[64];
        ENATSC3Coding vPLPCod[64];
        int32_t plpId;
        uint64_t plpValidMask = 0;
        uint64_t plpLockedMask = 0;
        uint64_t llsValidMask = 0;
        int32_t fParseError = 0;
        int32_t i;
        int hdret;

        if (!pHDHomerun)
        {
            pHDHomerun = hdhomerun_device_create_from_str(strHDHR, NULL);
            if (!pHDHomerun)
            {
                ERRORPRINT("Failed to create HDHR device for %s\n", strHDHR);
                sleep(1);
                continue;
            }
            hdhrErrorCount = 0;
        }

        hdret = hdhomerun_device_get_var(pHDHomerun, strStatusPath, 
            &pStatusString, NULL);
        if (hdret != 1)
        {
            ERRORPRINT("Failed to get tuner status (%d)\n", hdret);
            if (hdhrErrorCount++ >= 5)
            {
                ERRORPRINT("HDHomerun %s not responding, restarting\n", 
                    pSource->base.strSerial);
                hdhomerun_device_destroy(pHDHomerun);
                pHDHomerun = NULL;
                continue;
            }
            sleep(1);
            continue;
        }

        // Example tuner status:
        // ch=auto:6 lock=atsc3 ss=89(-55dBm) snq=63(16.6dB) seq=100 bps=0 pps=0
        pLockStatus = strstr(pStatusString, "lock=");
        if (pLockStatus && !strncmp(pLockStatus + 5, "atsc3", 5))
        {
            masterLock = 1;
        }

        hdret = hdhomerun_device_get_var(pHDHomerun, strPLPInfoPath, 
            &pPLPInfoString, NULL);
        if (hdret != 1)
        {
            ERRORPRINT("Failed to get plp info (%d)\n", hdret);
            hdhrErrorCount++;
            sleep(1);
            continue;
        }
        hdhrErrorCount = 0;

        // Ignore empty plp info. This can happen when there isn't lock, and
        // also for a time as a transfer is restarted.
        if (pPLPInfoString[0] == 0)
        {
            usleep(500000);
            continue;
        }

        // Example plp info:
        //  1: sfi=0 mod=qam256 cod=9/15 layer=core ti=none lls=0

        pPLPInfoIter = pPLPInfoString;
        while (*pPLPInfoIter != '\n' && *pPLPInfoIter)
        {
            char *pPLPInfoEnd;

            pPLPInfoEnd = strchr(pPLPInfoIter, '\n');
            if (!pPLPInfoEnd)
            {
                ERRORPRINT("Failed to parse tuner status string {%s}\n",
                    pPLPInfoIter);
                fParseError = 1;
                break;
            }

            plpId = atol(pPLPInfoIter);
            if (plpId < 0 || plpId > 63)
            {
                ERRORPRINT("Invalid plp %d (%s)\n", plpId,
                    pPLPInfoString);
            }
            else
            {
                plpValidMask |= ((int64_t)1 << plpId);

                pLLS = (char *)memmem(pPLPInfoIter,
                    (size_t)(pPLPInfoEnd - pPLPInfoIter), "lls=", 4);
                if (!pLLS)
                {
                    ERRORPRINT("Failed to find lls info {%s}\n",
                        pPLPInfoString);
                }
                else
                {
                    if (pLLS[4] == '0')
                    {
                        // Nothing to do
                    }
                    else if (pLLS[4] == '1')
                    {
                        llsValidMask |= ((int64_t)1 << plpId);
                    }
                    else
                    {
                        ERRORPRINT("Unexpected lls value %s (%s)\n",
                            pLLS, pPLPInfoString);
                    }
                }

                pPLPLock = memmem(pPLPInfoIter,
                    (size_t)(pPLPInfoEnd - pPLPInfoIter), "lock=", 5);
                if (!pPLPLock)
                {
                    // Lack of a "lock=" means the PLP is not tuned. 
                    // Implicitly it has no lock.
                }
                else
                {
                    if (pPLPLock[5] == '0')
                    {
                        // Nothing to do
                    }
                    else if (pPLPLock[5] == '1')
                    {
                        plpLockedMask |= ((int64_t)1 << plpId);
                    }
                    else
                    {
                        ERRORPRINT("Unexpected lock value %s (%s)\n",
                            pLLS, pPLPInfoString);
                    }
                }

                pPLPMod = memmem(pPLPInfoIter,
                    (size_t)(pPLPInfoEnd - pPLPInfoIter), "mod=", 4);
                if (!pPLPMod)
                {
                    ERRORPRINT("Failed to find modulation info {%s}\n",
                        pPLPInfoString);
                }
                else
                {
                    for (i = 0; i < numof(vModTableG); i++)
                    {
                        if (!strncmp(pPLPMod + 4, vModTableG[i].pString, 
                            vModTableG[i].sString))
                        {
                            vPLPMod[plpId] = vModTableG[i].mod;
                            break;
                        }
                    }

                    if (i == numof(vModTableG))
                    {
                        ERRORPRINT("Unexpected mod value %s (%s)\n",
                            pPLPMod, pPLPInfoString);
                        vPLPMod[plpId] = ENATSC3Modulation_Unknown;
                    }
                }

                pPLPCod = memmem(pPLPInfoIter,
                    (size_t)(pPLPInfoEnd - pPLPInfoIter), "cod=", 4);
                if (!pPLPCod)
                {
                    ERRORPRINT("Failed to find coding info {%s}\n",
                        pPLPInfoString);
                }
                else
                {
                    for (i = 0; i < numof(vCodTableG); i++)
                    {
                        if (!strncmp(pPLPCod + 4, vCodTableG[i].pString, 
                            vCodTableG[i].sString))
                        {
                            vPLPCod[plpId] = vCodTableG[i].cod;
                            break;
                        }
                    }

                    if (i == numof(vCodTableG))
                    {
                        ERRORPRINT("Unexpected cod value %s (%s)\n",
                            pPLPCod, pPLPInfoString);
                        vPLPCod[plpId] = ENATSC3Coding_Unknown;
                    }
                }
            }

            pPLPInfoIter = pPLPInfoEnd + 1;
        }

        if (!fParseError)
        {
            uint64_t mask;
            int32_t i;

            if (pthread_mutex_lock(&pSource->statsLock))
            {
                ERRORPRINT("Failed to lock\n");
            }

            pSource->stats.masterLock = masterLock;
            pSource->stats.plpValidMask = plpValidMask;
            pSource->stats.plpLockedMask = plpLockedMask;
            pSource->stats.llsValidMask = llsValidMask;

            for (i = 0, mask = plpValidMask; i < 64; i++, plpValidMask <<= 1)
            {
                if (mask & 1)
                {
                    pSource->stats.vPLPStats[i].modulation = vPLPMod[i];
                    pSource->stats.vPLPStats[i].coding = vPLPCod[i];
                }
            }

            if (pthread_mutex_unlock(&pSource->statsLock))
            {
                ERRORPRINT("Failed to unlock\n");
            }
        }

        usleep(1000000);
    }

out:
    if (pHDHomerun)
    {
        hdhomerun_device_destroy(pHDHomerun);
    }

    return NULL;
}

ENStatus
ATSC3J288ALPHDHRSourceRun(ENATSC3Source hSource, int32_t *pFRunning)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)hSource;
    int32_t i;
    int32_t j;
    uint64_t llsValidMask = 0;
    int32_t fHaveTransferThread = 0;
    int32_t fHavePLPStatusThread = 0;
    ENStatus rval = ENStatus_Success;

    memset(&pSource->stats, 0, sizeof(pSource->stats));

    pSource->fRunning = 1;

    if (pthread_create(&pSource->plpStatusThread, NULL, PLPStatusThread,
        pSource))
    {
        ERRORPRINT("Failed to start PLP status thread\n");
        rval = ENStatus_Failure;
        goto out;
    }
    fHavePLPStatusThread = 1;

    while (*pFRunning)
    {
        if (!fHaveTransferThread)
        {
            if (pthread_create(&pSource->transferThread, NULL, 
                ATSC3J288ALPHDHRSourceTransferThread, pSource))
            {
                ERRORPRINT("Failed to start traansfer thread\n");
                rval = ENStatus_Failure;
                goto out;
            }
            fHaveTransferThread = 1;
        }

        while (*pFRunning && fHaveTransferThread)
        {
            usleep(100000);

            if (pSource->stats.llsValidMask != llsValidMask)
            {
                struct timespec rtnow;

                clock_gettime(CLOCK_REALTIME, &rtnow);

                llsValidMask = pSource->stats.llsValidMask;
                pSource->LLSPLPsCallback(llsValidMask, 
                    &rtnow, pSource->stats.nALPPackets,
                    pSource->pCallbackUserData);
            }

            if (pSource->fPLPIdMaskChanged)
            {
                int32_t fPLPsChanged = 0;
                uint64_t plpIdMaskTemp;

                // We're going to snapshot the mask since processing the change 
                // could take a while and we don't want to hold the lock that
                // long. If someone changes the mask again we'll catch it the 
                // next time around.
                if (pthread_mutex_lock(&pSource->plpIdMaskChangedLock))
                {
                    ERRORPRINT("Failed to lock\n");
                }

                pSource->plpIdMask = pSource->requestedPlpIdMask;
                pSource->fPLPIdMaskChanged = 0;

                // First reconcile the current PLPs with the mask.
                plpIdMaskTemp = pSource->plpIdMask;
                for (i = 0; i < numof(pSource->vPLPIds); i++)
                {
                    if (pSource->vPLPIds[i] != INVALID_PLP_ID)
                    {
                        if (plpIdMaskTemp & 1ULL << pSource->vPLPIds[i])
                        {
                            // Already set, remove from mask
                            plpIdMaskTemp &= ~1ULL << pSource->vPLPIds[i];
                        }
                        else
                        {
                            // No longer set, free up slot
                            pSource->vPLPIds[i] = INVALID_PLP_ID;
                            fPLPsChanged = 1;
                        }
                    }
                }

                // Next assign new PLPs to slots
                for (i = 0; i < 64; i++)
                {
                    if (plpIdMaskTemp & 1)
                    {
                        for (j = 0; j < numof(pSource->vPLPIds); j++)
                        {
                            if (pSource->vPLPIds[j] == INVALID_PLP_ID)
                            {
                                pSource->vPLPIds[j] = i;
                                break;
                            }
                        }

                        if (j == numof(pSource->vPLPIds))
                        {
                            ERRORPRINT("Too many plps: couldn't set %d, mask %x, slots %d %d %d %d\n", 
                                i, pSource->plpIdMask, pSource->vPLPIds[0], 
                                pSource->vPLPIds[1], pSource->vPLPIds[2], 
                                pSource->vPLPIds[3]);
                        }
                        fPLPsChanged = 1;
                    }

                    plpIdMaskTemp >>=1;
                }

                // See if only a single PLP is active. This is required to
                // determine the PLP Id of ALP and Baseband packets. See the
                // ALP and Baseband packet callback functions.
                pSource->singlePLPId = INVALID_PLP_ID;
                for (i = 0, plpIdMaskTemp = pSource->plpIdMask; 
                     plpIdMaskTemp;
                     i++, plpIdMaskTemp >>= 1)
                {
                    if (plpIdMaskTemp & 1)
                    {
                        if (pSource->singlePLPId == INVALID_PLP_ID)
                        {
                            pSource->singlePLPId = i;
                        }
                        else
                        {
                            // More than one PLP has been requested.
                            pSource->singlePLPId = INVALID_PLP_ID;
                            break;
                        }
                    }
                }

                LOGPRINT("Setting PLP mask 0x%x slots %d %d %d %d\n",
                    pSource->plpIdMask, pSource->vPLPIds[0], 
                    pSource->vPLPIds[1], pSource->vPLPIds[2], 
                    pSource->vPLPIds[3]); 

                if (fPLPsChanged)
                {
                    closeCurlSocket(pSource);

                    if (pthread_join(pSource->transferThread, NULL))
                    {
                        ERRORPRINT("Failed to join transfer thread\n");
                    }

                    fHaveTransferThread = 0;
                }
                else
                {
                    LOGPRINT("No change to selected PLPs, ignoring\n");
                }

                if (pthread_mutex_unlock(&pSource->plpIdMaskChangedLock))
                {
                    ERRORPRINT("Failed to unlock\n");
                }
            }
        }
    }

out:
    pSource->fRunning = 0;

    if (fHaveTransferThread)
    {
        closeCurlSocket(pSource);

        if (pthread_join(pSource->transferThread, NULL))
        {
            ERRORPRINT("Failed to join transfer thread\n");
        }

        fHaveTransferThread = 0;
    }

    if (fHavePLPStatusThread)
    {
        if (pthread_join(pSource->plpStatusThread, NULL))
        {
            ERRORPRINT("Failed to join PLP status thread\n");
        }
    }

    return rval;
}

ENStatus
ENATSC3J288ALPHDHRSourceOpen(int32_t deviceNumber, 
    ENATSC3J288ALPHDHRSourceOperatingMode operatingMode, 
    int32_t frequencyHz, ENATSC3Source *ppSource)
{
    ENATSC3J288ALPHDHRSource *pSource = NULL;
    ENStatus rval = ENStatus_Success;

    if (deviceNumber >= atsc3J288ALPHDHRSourceGlobalsG.nSources)
    {
        ERRORPRINT("Source %d out of range %d\n", deviceNumber,
            atsc3J288ALPHDHRSourceGlobalsG.nSources);
        rval = ENStatus_Failure;
        goto out;
    }

    pSource = atsc3J288ALPHDHRSourceGlobalsG.vSources + deviceNumber;

    if (pthread_mutex_lock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (pSource->fOpened)
    {
        ERRORPRINT("Device %d already opened\n", deviceNumber);
        rval = ENStatus_Failure;
        goto out;
    }

    pSource->verbosity = ENVerbosity_Terse;

    pSource->frequencyKHz = frequencyHz/1000 + 3000;

    // Set all PLPs invalid. Note that we will actually tune the first slot
    // to PLP 0, since one PLP must be set when initially tuning and 0 seems
    // to always be present in emissions.
    pSource->vPLPIds[0] = INVALID_PLP_ID;
    pSource->vPLPIds[1] = INVALID_PLP_ID;
    pSource->vPLPIds[2] = INVALID_PLP_ID;
    pSource->vPLPIds[3] = INVALID_PLP_ID;

    if (operatingMode != ENATSC3J288ALPHDHRSourceOperatingMode_ATSC &&
        operatingMode != ENATSC3J288ALPHDHRSourceOperatingMode_ATSC3)
    {
        ERRORPRINT("Invalid operating mode %d\n", operatingMode);
        rval = ENStatus_Failure;
        goto errout;
    }

    pSource->operatingMode = operatingMode;
    // channel lower edge frequency to channel number
    if (frequencyHz >= 470000000 && frequencyHz <= 884000000)
    {
        // UHF
        pSource->rfChannelNumber = 
            14 + ((frequencyHz - 470000000) / 6000000);
    }
    else if (frequencyHz >= 174000000 && frequencyHz <= 210000000)
    {
        // VHF Hi
        pSource->rfChannelNumber = 
            7 + ((frequencyHz - 174000000) / 6000000);
    }
    else if (frequencyHz >= 54000000 && frequencyHz <= 82000000)
    {
        // VHF Lo
        pSource->rfChannelNumber = 
            2 + ((frequencyHz - 54000000) / 6000000);
    }
    else
    {
        ERRORPRINT("Invalid frequency %d\n", frequencyHz);
        rval = ENStatus_Failure;
        goto errout;
    }

    pSource->requestedPlpIdMask = 0;
    pSource->plpIdMask = 0;
    pSource->singlePLPId = INVALID_PLP_ID;
    pSource->fPLPIdMaskChanged = 0;
    pSource->LLSPLPsCallback = NULL;
    pSource->pCallbackUserData = NULL;
    pSource->curlSocket = -1;
    pSource->hATSC3J288ALP = NULL;

    pSource->fOpened = 1;

    *ppSource = (ENATSC3Source)pSource;

out:
    if (pSource)
    {
        if (pthread_mutex_unlock(&pSource->lock))
        {
            ERRORPRINT("Failed to unlock\n");
        }
    }

    return rval;

errout:
    goto out;
}

ENStatus
ATSC3J288ALPHDHRSourceClose(ENATSC3Source hSource)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pSource->fOpened = 0;

    if (pthread_mutex_unlock(&pSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    return rval;
}

ENStatus
ATSC3J288ALPHDHRSourceSetPLPIds(ENATSC3Source hSource, uint64_t plpIdMask)
{
    ENATSC3J288ALPHDHRSource *pSource = (ENATSC3J288ALPHDHRSource *)hSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pSource->plpIdMaskChangedLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pSource->requestedPlpIdMask = plpIdMask;
    pSource->fPLPIdMaskChanged = 1;

    if (pthread_mutex_unlock(&pSource->plpIdMaskChangedLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ENATSC3J288ALPHDHRSourceRegisterCallbacks(ENATSC3Source pSource,
    ENATSC3SourceLLSPLPsCallbackFunction LLSPLPsCallback,
    ENATSC3SourceBasebandPacketCallbackFunction BasebandPacketCallback,
    ENATSC3SourceALPPacketCallbackFunction ALPPacketCallback,
    void *pUserData)
{
    ENATSC3J288ALPHDHRSource *pHDHRSource = 
        (ENATSC3J288ALPHDHRSource *)pSource;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pHDHRSource->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pHDHRSource->LLSPLPsCallback = LLSPLPsCallback;
    pHDHRSource->ALPPacketCallback = ALPPacketCallback;
    pHDHRSource->pCallbackUserData = pUserData;

    if (pthread_mutex_unlock(&pHDHRSource->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ENATSC3J288ALPHDHRSourceInitializeSource(ENATSC3J288ALPHDHRSource *pSource,
    struct hdhomerun_discover_device_t *pHDHR, int32_t tunerNum)
{
    int32_t fHaveLock = 0;
    int32_t fHavePLPIdMaskChangedLock = 0;
    int32_t fHaveStatsLock = 0;
    int32_t fHavePTPInfoLock = 0;
    int32_t fHaveCurlSocketLock = 0;
    ENStatus rval = ENStatus_Success;

    pSource->hdhr = *pHDHR;
    pSource->base.RegisterCallbacks = ENATSC3J288ALPHDHRSourceRegisterCallbacks;
    pSource->base.Start = ATSC3J288ALPHDHRSourceRun;
    pSource->base.Close = ATSC3J288ALPHDHRSourceClose;
    pSource->base.SetPLPIds = ATSC3J288ALPHDHRSourceSetPLPIds;
    strncpy(pSource->base.strModel, "HDHR",
        sizeof(pSource->base.strModel) - 1);
    pSource->base.strModel[sizeof(pSource->base.strModel) - 1] = 0;
    snprintf(pSource->base.strSerial, sizeof(pSource->base.strSerial), 
        "%08X:%d", pHDHR->device_id, tunerNum);
    pSource->base.strSerial[sizeof(pSource->base.strSerial) - 1] = 0;
    pSource->tunerNum = tunerNum;
    pSource->fOpened = 0;

    if (pthread_mutex_init(&pSource->lock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveLock = 1;

    if (pthread_mutex_init(&pSource->plpIdMaskChangedLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHavePLPIdMaskChangedLock = 1;

    if (pthread_mutex_init(&pSource->statsLock, NULL))
    {
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveStatsLock = 1;

    if (pthread_mutex_init(&pSource->ptpInfoLock, NULL))
    {
        rval = ENStatus_Failure;
        goto errout;
    }
    fHavePTPInfoLock = 1;

    if (pthread_mutex_init(&pSource->curlSocketLock, NULL))
    {
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveCurlSocketLock = 1;

out:
    return rval;

errout:
    if (fHaveCurlSocketLock)
    {
        if (pthread_mutex_destroy(&pSource->curlSocketLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fHavePTPInfoLock)
    {
        if (pthread_mutex_destroy(&pSource->ptpInfoLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fHaveStatsLock)
    {
        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fHavePLPIdMaskChangedLock)
    {
        if (pthread_mutex_destroy(&pSource->plpIdMaskChangedLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fHaveLock)
    {
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    goto out;
}

static ENStatus
HDHRGetInfo(struct hdhomerun_discover_device_t *pHDHR, 
    int32_t *pFModulationATSC3, char *pModel, int32_t sModel)
{
    struct hdhomerun_device_t *pHDHomerun = NULL;
    char *pModelString;
    char *pFeatureString;
    char *pFeatureIter;
    int hdret;
    ENStatus rval = ENStatus_Success;
    
    *pFModulationATSC3 = 0;

    pHDHomerun = hdhomerun_device_create(pHDHR->device_id, pHDHR->ip_addr, 0,
        NULL);
    if (!pHDHomerun)
    {
        ERRORPRINT("Failed to create HDHR device for %08X\n", pHDHR->device_id);
        rval = ENStatus_Failure;
        goto out;
    }

    hdret = hdhomerun_device_get_var(pHDHomerun, "/sys/hwmodel",
        &pModelString, NULL);
    if (hdret != 1)
    {
        ERRORPRINT("Failed to get model for %08X (%d)\n", pHDHR->device_id, 
            hdret);
        rval = ENStatus_Failure;
        goto out;
    }
    strncpy(pModel, pModelString, sModel);
    pModel[sModel - 1] = 0;

    hdret = hdhomerun_device_get_var(pHDHomerun, "/sys/features",
        &pFeatureString, NULL);
    if (hdret != 1)
    {
        ERRORPRINT("Failed to get feature info for %08X (%d)\n", 
            pHDHR->device_id, hdret);
        rval = ENStatus_Failure;
        goto out;
    }

    pFeatureIter = pFeatureString;
    while (1)
    {
        char *pIterEnd;

        pIterEnd = strchr(pFeatureIter, '\n');
        if (!pIterEnd)
        {
            break;
        }
        if (!strncmp(pFeatureIter, "modulation: ", sizeof("modulation: ") - 1))
        {
            char *pModulation = pFeatureIter + sizeof("modulation: ") - 1;
            while (pModulation < pIterEnd)
            {
                if (!strncmp(pModulation, "atsc3", 5) && 
                    (pModulation[5] == '\n' || pModulation[5] == ' '))
                {
                    *pFModulationATSC3 = 1;
                }
                
                while (*pModulation != ' ' && *pModulation != '\n')
                {
                    pModulation++;
                }
                pModulation++;
            }
        }

        pFeatureIter = pIterEnd + 1;
    }

out:
    if (pHDHomerun)
    {
        hdhomerun_device_destroy(pHDHomerun);
    }

    return rval;
}

static ENStatus
ATSC3J288ALPHDHRSourceFindHDHomeruns(void)
{
    struct hdhomerun_discover_device_t vHDHR[HDHR_MAX_TUNERS];
    ENATSC3J288ALPHDHRSource *pSource;
    int32_t nHDHR;
    int32_t i;
    int32_t j;
    ENStatus rval = ENStatus_Success;

    nHDHR = hdhomerun_discover_find_devices_custom_v2(0,
        HDHOMERUN_DEVICE_TYPE_TUNER, HDHOMERUN_DEVICE_ID_WILDCARD,
        vHDHR, numof(vHDHR));
    if (nHDHR < 0)
    {
        ERRORPRINT("Failed to send discover\n");
        rval = ENStatus_Failure;
        goto out;
    }

    LOGPRINT_NOCONSOLE("Found %d HDHomerun devices\n", nHDHR);

    pSource = atsc3J288ALPHDHRSourceGlobalsG.vSources;
    for (i = 0; i < nHDHR; i++)
    {
        int32_t fModulationATSC3;
        char strModel[64];

        if (HDHRGetInfo(vHDHR + i, &fModulationATSC3, strModel, 
            sizeof(strModel)))
        {
            ERRORPRINT("Failed to query features for %08X\n",  
                vHDHR[i].device_id);
        }

        if (!fModulationATSC3)
        {
            continue;
        }

        // There doesn't seem to be a way to differentiate between ATSC-3 
        // capable tuners vs ATSC-1 only. We know on HDHR5-4K-DEV and 
        // HDFX-4K-DEV that tuners 0 and 1 are ATSC 1 & ATSC3 while tuners 2 
        // and 3 are ATSC-1 only.
        if (!strcmp(strModel, "HDHR5-4K-DEV") || 
            !strcmp(strModel, "HDFX-4K-DEV"))
        {
            vHDHR[i].tuner_count = 2;
        }
        else
        {
            vHDHR[i].tuner_count = 0;
        }

        for (j = 0; j < vHDHR[i].tuner_count; j++)
        {
            if (atsc3J288ALPHDHRSourceGlobalsG.nSources == 
                numof(atsc3J288ALPHDHRSourceGlobalsG.vSources))
            {
                ERRORPRINT("Too many sources, not adding %08X:%d\n", 
                    vHDHR[i].device_id, j);
            }
            else
            {
                LOGPRINT_NOFILE("ADDING %08X:%d\n",
                    vHDHR[i].device_id, j);
                if (ENATSC3J288ALPHDHRSourceInitializeSource(pSource, 
                    vHDHR + i, j))
                {
                    ERRORPRINT("Failed to add %08X:%d\n", 
                        vHDHR[i].device_id, j);
                }
                else
                {
                    pSource++;
                    atsc3J288ALPHDHRSourceGlobalsG.nSources++;
                }
            }
        }
    }

out:
    return rval;
}

ENStatus
ENATSC3J288ALPHDHRSourceInitialize(int32_t *pNDevices)
{
    ENATSC3J288ALPHDHRSource *pSource;
    int32_t fCurlInitialized = 0;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    if (curl_global_init(CURL_GLOBAL_ALL) != CURLE_OK)
    {
        ERRORPRINT("Failed to init curl\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fCurlInitialized = 1;

    atsc3J288ALPHDHRSourceGlobalsG.nSources = 0;

    ATSC3J288ALPHDHRSourceFindHDHomeruns();

    *pNDevices = atsc3J288ALPHDHRSourceGlobalsG.nSources;

out:
    return rval;

errout:
    for (i = 0, pSource = atsc3J288ALPHDHRSourceGlobalsG.vSources;
         i < atsc3J288ALPHDHRSourceGlobalsG.nSources;
         i++, pSource++)
    {
        if (pthread_mutex_destroy(&pSource->ptpInfoLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->plpIdMaskChangedLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    if (fCurlInitialized)
    {
        curl_global_cleanup();
    }

    goto out;
}

ENStatus
ENATSC3J288ALPHDHRSourceShutdown(void)
{
    ENATSC3J288ALPHDHRSource *pSource;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0, pSource = atsc3J288ALPHDHRSourceGlobalsG.vSources; 
         i < atsc3J288ALPHDHRSourceGlobalsG.nSources; 
         i++, pSource++)
    {
        if (pthread_mutex_destroy(&pSource->curlSocketLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->statsLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->plpIdMaskChangedLock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
        if (pthread_mutex_destroy(&pSource->lock))
        {
            ERRORPRINT("Failed to destroy lock\n");
        }
    }

    curl_global_cleanup();

    return rval;
}

ENStatus
ENATSC3J288ALPHDHRSourceGetDeviceInfo(int32_t deviceNumber,
    char *pModel, int32_t nModelBytes, char *pSerial, int32_t nSerialBytes)
{
    ENATSC3J288ALPHDHRSource *pSource;
    ENStatus rval = ENStatus_Success;

    if (deviceNumber >= atsc3J288ALPHDHRSourceGlobalsG.nSources)
    {
        ERRORPRINT("Source %d out of range %d\n", deviceNumber,
            numof(atsc3J288ALPHDHRSourceGlobalsG.vSources));
        rval = ENStatus_Failure;
        goto out;
    }

    pSource = atsc3J288ALPHDHRSourceGlobalsG.vSources + deviceNumber;
    strncpy(pModel, pSource->base.strModel, nModelBytes - 1);
    pModel[nModelBytes - 1] = 0;
    strncpy(pSerial, pSource->base.strSerial, nSerialBytes - 1);
    pSerial[nSerialBytes - 1] = 0;

out:
    return rval;
}
