/*
 * Copyright © 2018-2021 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <assert.h>

#include <zlib.h>
#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include <libenatsc3sls/atsc3sls_usbd.h>
#include <libenatsc3route/atsc3routesls_stsid.h>
#include <libenatsc3route/atsc3routesls_envelope.h>

#define MAX_PAYLOAD_SIZE 65536
#define NANOSECONDS_PER_SECOND 1000000000

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, char **ppOutPath,
    char **ppDocType)
{
    int32_t err = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutPath = NULL;
    *ppDocType = NULL;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-b"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutPath = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-t"))
        {
            if (i + 1 < argc)
            {
                *ppDocType = argv[i+1];
                i++;
            }
            else err = 1;
        }
        else err = 1;
    }

    if (!*ppInFilename || !*ppOutPath || !*ppDocType)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -b udpfilename -o outpath -t stsid|usbd|envelope|mpd\n", argv[0]);
    }

    return err;
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    int32_t fXML2Initialized = 0;
    char *pInFilename;
    char *pOutPath;
    char *pDocType;
    int fd = -1;
    uint8_t *pPayload = NULL;
    ENATSC3SLSUSBDBundleDescriptionNode usbdNode;
    ENATSC3ROUTESLSSTSIDNode stsidNode;
    ENATSC3ROUTESLSEnvelopeNode envelopeNode;
    int nb;
    int rval = 0;

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    xmlInitParser();
    fXML2Initialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &pOutPath, &pDocType))
    {
        rval = 1;
        goto out;
    }

    fd =  open(pInFilename, O_RDONLY);
    if (fd < 0)
    {
        fprintf(stderr, "Failed to open %s for reading: %s\n", pInFilename, 
            strerror(errno));
        rval = 1;
        goto out;
    }

    pPayload = (uint8_t *)malloc(MAX_PAYLOAD_SIZE);
    if (!pPayload)
    {
        fprintf(stderr, "Out of memory\n");
        rval = 1;
        goto out;
    }

    nb = read(fd, pPayload, MAX_PAYLOAD_SIZE);
    if (nb < 0)
    {
        fprintf(stderr, "Error reading payload: %s\n", strerror(errno));
        rval = 1;
        goto out;
    }
    else if (nb == MAX_PAYLOAD_SIZE)
    {
        fprintf(stderr, "xml file too large, exiting\n");
        rval = 1;
        goto out;
    }

    if (!strcmp(pDocType, "stsid"))
    {
        if (ENATSC3ROUTESLSSTSIDParse((char *)pPayload, nb, &stsidNode))
        {
            fprintf(stderr, "Failed to parse STSID\n");
        }
    }
    else if (!strcmp(pDocType, "usbd"))
    {
        if (ENATSC3SLSUSBDParse((char *)pPayload, nb, &usbdNode))
        {
            fprintf(stderr, "Failed to parse usbd\n");
        }
    }
    else if (!strcmp(pDocType, "envelope"))
    {
        if (ENATSC3ROUTESLSEnvelopeParse((char *)pPayload, nb, &envelopeNode))
        {
            fprintf(stderr, "Failed to parse envelope\n");
        }
    }
    else if (!strcmp(pDocType, "mpd"))
    {
        // The player level parses the MPD.
    }

out:
    if (fd >= 0)
    {
        close(fd);
    }

    if (pPayload)
    {
        free(pPayload);
    }

    if (fXML2Initialized)
    {
        xmlCleanupCharEncodingHandlers();
        xmlDictCleanup();
        xmlCleanupParser();
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}
