/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <inttypes.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <arpa/inet.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libendebug/libendebug.h>
#include <enatsc3-common/enatsc3_specdefs.h>

#include <libenatsc3utils/enatsc3bitratestatistics.h>

#include <enatsc3-common/enatsc3.h>
#include "libenatsc3alp.h"

static char *ensys = "alp";
static char *ensubsys = "main";

#define MAX_NETWORK_LAYER_PACKET_SIZE 65536

// Structure describing a segment of a network layer packet extracted from an
// ALP packet. To simplify the interface a single network layer packet wrapped 
// in a single ALP packet is treated as a network layer packet wrapped in a 
// segment (segmentSequenceNumber = 0, lastSegmentIndicator = 1). Network 
// layer packets concatenated within a single ALP packet are treated similarly.
typedef struct 
{
    // The type of the network layer packet that this segment originated from
    // (e.g. IPv4)
    ENATSC3ALPPacketType packetType;

    // The payload for the segment.
    const uint8_t *pPayload;
    int32_t payloadLength;

    // Whether the segment originated from a baseband packet containing
    // errors
    int32_t fError;

    // Whethere there is data missing between the previous segment and the
    // current one.
    int32_t fDiscontinuity;

    // The sequence number of the segment in the case of segmented packets. 
    // The sequence number is zero for the first segment of a network layer 
    // packet and increases monotonically.  The sequence number wraps to
    // zero at 32 (2^5). A discontinuity in the sequence numbering indicates
    // loss of one or more segments.
    int32_t segmentSequenceNumber;

    // The last segment indicator is 1 for the last segment in a network layer
    // packet and zero otherwise. 
    int32_t lastSegmentIndicator;

    // If sub-stream information is present, then SID contains the SID. If
    // sub-stream information is not present then SID will be -1.
    int32_t SID;

    // If a header extension is present, then extensionType, extensionLength,
    // and pExtension contain the values of the header extension. If no header
    // extensioni s presnet then extensionType will be -1.
    int32_t extensionType;
    int32_t extensionLength;
    const uint8_t *pExtension;
} ENATSC3ALPSegment;

typedef struct _ENATSC3ALPHandle_t
{
    ENVerbosity verbosity;

    // Array of PLP structures. An ATSC 3.0 broadcast can have up to 64 PLPs
    // per channel.
    struct _ENATSC3PLP
    {
        //  1 if this PLP has been added, 0 if not.
        int32_t fEnabled;

        uint32_t plpId;

        uint8_t *pNetworkLayerPacket;
        int32_t networkLayerPacketBytesUsed;
        int32_t networkLayerPacketBytesAllocated;

        // 1 if we are waiting for the first segment in a network layer packet,
        // or 0 if we are processing a multi-segment network layer packet. 
        int32_t fWaitForNetworkLayerPacketStart;

        // The type and SID of network layer packet currently being processed. 
        ENATSC3ALPPacketType packetType;
        int32_t SID;

        // Whether there was an error in the packet being processed
        int32_t fError;

        // Whether there was packet loss between the packet being processed
        // and the previously processed packet
        int32_t fDiscontinuity;

        // The time associated with the first segment of the packet
        struct timespec packetTimeUTC;

        // The packet number of the first segment of the packet
        uint64_t packetNumber;

        // When processing a multi-segment network layer packet, the expected
        // segmentSequenceNumber of the next segment.
        int32_t expectedSegmentSequenceNumber;

        // Statistics
        ENATSC3BitrateStatistics stats;
    } vPLPs[64];

    // Callbacks
    ENATSC3ALPIPV4CallbackFunction IPV4Callback;
    ENATSC3ALPLinkLayerSignalingCallbackFunction LinkLayerSignalingCallback;
    void *pCallbackUserData;

    // Statistics
    int32_t fBitrateStatisticsEnabled;

    // Master lock for this structure
    pthread_mutex_t lock;
} ENATSC3ALP, *ENATSC3ALPHandle_t;

ENStatus
ENATSC3ALPSetVerbosity(ENATSC3ALPHandle_t hATSC3ALP, 
    ENVerbosity verbosity)
{
    ENATSC3ALP *pATSC3ALP = (ENATSC3ALP *)hATSC3ALP;
    ENStatus rval = ENStatus_Success;

    pATSC3ALP->verbosity = verbosity;

    return rval;
}

ENStatus
ENATSC3ALPOpen(ENATSC3ALPHandle_t *phATSC3ALP)
{
    ENATSC3ALP *pATSC3ALP = NULL;
    int32_t fHaveLock = 0;
    int32_t i;
    struct _ENATSC3PLP *pPLP;
    ENStatus rval = ENStatus_Success;

    pATSC3ALP = (ENATSC3ALP *)calloc(sizeof(ENATSC3ALP), 1);
    if (!pATSC3ALP)
    {
        rval = ENStatus_Failure;
        goto out;
    }

    pATSC3ALP->verbosity = ENVerbosity_Terse;

    if (pthread_mutex_init(&pATSC3ALP->lock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveLock = 1;

    for (i = 0, pPLP = pATSC3ALP->vPLPs; 
         i < numof(pATSC3ALP->vPLPs); 
         i++, pPLP++)
    {
        pPLP->plpId = i;
        ENATSC3BitrateStatisticsInitialize(&pPLP->stats, 2000);
        snprintf(pPLP->stats.strName, sizeof(pPLP->stats.strName), 
            "ALP PLP %d", i);
        pPLP->stats.strName[sizeof(pPLP->stats.strName) - 1] = 0;
    }

    *phATSC3ALP = pATSC3ALP;

out:
    return rval;

errout:
    if (pATSC3ALP)
    {
        if (fHaveLock)
        {
            if (pthread_mutex_destroy(&pATSC3ALP->lock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }
        free(pATSC3ALP);
    }

    goto out;
}

ENStatus
ENATSC3ALPClose(ENATSC3ALPHandle_t hATSC3ALP)
{
    ENATSC3ALP *pATSC3ALP = (ENATSC3ALP *)hATSC3ALP;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    for (i = 0; i < numof(pATSC3ALP->vPLPs); i++)
    {
        if (pATSC3ALP->vPLPs[i].fEnabled)
        {
            free(pATSC3ALP->vPLPs[i].pNetworkLayerPacket);
            pATSC3ALP->vPLPs[i].pNetworkLayerPacket = NULL;
        }
    }

    if (pthread_mutex_destroy(&pATSC3ALP->lock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    free(pATSC3ALP);

    return rval;
}

void
ENATSC3ALPSetBitrateStatisticsEnabled(ENATSC3ALPHandle_t hATSC3ALP, 
    int32_t fEnabled)
{
    ENATSC3ALP *pATSC3ALP = (ENATSC3ALP *)hATSC3ALP;

    pATSC3ALP->fBitrateStatisticsEnabled = fEnabled;
}

ENStatus
ENATSC3ALPSetPLPs(ENATSC3ALPHandle_t hATSC3ALP, uint64_t plpIdMask)
{
    ENATSC3ALP *pATSC3ALP = (ENATSC3ALP *)hATSC3ALP;
    struct _ENATSC3PLP *pPLP;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pATSC3ALP->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    for (i = 0, pPLP = pATSC3ALP->vPLPs; 
         i < 64; 
         i++, pPLP++)
    {
        if (plpIdMask & 1)
        {
            if (!pPLP->fEnabled)
            {
                pPLP->networkLayerPacketBytesAllocated = 
                    MAX_NETWORK_LAYER_PACKET_SIZE;
                pPLP->pNetworkLayerPacket = 
                    (uint8_t *)malloc(pPLP->networkLayerPacketBytesAllocated);
                if (!pPLP->pNetworkLayerPacket)
                {
                    ERRORPRINT("Out of memory\n");
                    rval = ENStatus_Failure;
                }
                else
                {
                    pPLP->fWaitForNetworkLayerPacketStart = 1;
                    pPLP->fDiscontinuity = 0;
                    pPLP->fEnabled = 1;
                }
            }
        }
        else
        {
            if (pPLP->fEnabled)
            {
                free(pPLP->pNetworkLayerPacket);
                pPLP->pNetworkLayerPacket = NULL;

                pPLP->fEnabled = 0;
            }
        }

        plpIdMask >>= 1;
    }

    if (pthread_mutex_unlock(&pATSC3ALP->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ENATSC3ALPRegisterCallbacks(ENATSC3ALPHandle_t hATSC3ALP,
    ENATSC3ALPIPV4CallbackFunction IPV4Callback, 
    ENATSC3ALPLinkLayerSignalingCallbackFunction LinkLayerSignalingCallback,
    void *pUserData)
{
    ENATSC3ALP *pATSC3ALP = (ENATSC3ALP *)hATSC3ALP;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pATSC3ALP->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pATSC3ALP->IPV4Callback = IPV4Callback;
    pATSC3ALP->LinkLayerSignalingCallback = LinkLayerSignalingCallback;
    pATSC3ALP->pCallbackUserData = pUserData;

    if (pthread_mutex_unlock(&pATSC3ALP->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

static void
DispatchNetworkLayerPacket(ENATSC3ALP *pATSC3ALP, uint64_t packetNumber,
    struct _ENATSC3PLP *pPLP)
{
    ENATSC3ALPIPV4CallbackFunction IPV4CallbackLocal;
    ENATSC3ALPLinkLayerSignalingCallbackFunction LinkLayerSignalingCallbackLocal;
    void *pCallbackUserDataLocal;

    if (pthread_mutex_lock(&pATSC3ALP->lock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    IPV4CallbackLocal = pATSC3ALP->IPV4Callback;
    LinkLayerSignalingCallbackLocal = pATSC3ALP->LinkLayerSignalingCallback;
    pCallbackUserDataLocal = pATSC3ALP->pCallbackUserData;

    if (pthread_mutex_unlock(&pATSC3ALP->lock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    switch (pPLP->packetType)
    {
        case ENATSC3ALPPacketType_IPV4:
            if (IPV4CallbackLocal)
            {
                IPV4CallbackLocal(pPLP->plpId, &pPLP->packetTimeUTC, 
                    pPLP->packetNumber, pPLP->fError, pPLP->fDiscontinuity,
                    pPLP->pNetworkLayerPacket,
                    pPLP->networkLayerPacketBytesUsed, pPLP->SID,
                    pCallbackUserDataLocal);
            }
            break;
        case ENATSC3ALPPacketType_LinkLayerSignaling:
            if (LinkLayerSignalingCallbackLocal)
            {
                LinkLayerSignalingCallbackLocal(pPLP->plpId, 
                    &pPLP->packetTimeUTC, pPLP->packetNumber, 
                    pPLP->fError, pPLP->fDiscontinuity,
                    pPLP->pNetworkLayerPacket,
                    pPLP->networkLayerPacketBytesUsed, 
                    pCallbackUserDataLocal);
            }
            break;
        default:
            if (pATSC3ALP->verbosity > ENVerbosity_Terse)
            {
                DEBUGPRINT_NOCONSOLE(
                    "Unhandled packet type %d (bb pkt %"PRIu64")\n", 
                    pPLP->packetType, packetNumber);
            }
            break;
    }
}

static ENStatus
ProcessALPSegment(ENATSC3ALP *pATSC3ALP, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, struct _ENATSC3PLP *pPLP, 
    ENATSC3ALPSegment *pSegment)
{
    ENStatus rval = ENStatus_Success;

    if (pPLP->fWaitForNetworkLayerPacketStart)
    {
        if (pSegment->segmentSequenceNumber != 0)
        {
            if (pATSC3ALP->verbosity > ENVerbosity_Terse)
            {
                DEBUGPRINT_NOCONSOLE(
                    "Dropping segment %d waiting for start\n", 
                    pSegment->segmentSequenceNumber);
            }
            goto out;
        }

        pPLP->fWaitForNetworkLayerPacketStart = 0;
        pPLP->networkLayerPacketBytesUsed = 0;
        pPLP->packetType = pSegment->packetType;
        pPLP->packetTimeUTC = *pPacketTimeUTC;
        pPLP->packetNumber = packetNumber;
        pPLP->SID = pSegment->SID; 
        pPLP->expectedSegmentSequenceNumber = 0;
        pPLP->fError = 0;
    }

    // XXX For now we are dropping network layer packets in their entirety in 
    // the case of a missing segment. It may be desirable to present what data 
    // is available to the client and signal data loss. The client can then 
    // determine how to proceed.
    if (pSegment->segmentSequenceNumber != 
        pPLP->expectedSegmentSequenceNumber ||
        pSegment->packetType != pPLP->packetType ||
        pSegment->SID != pPLP->SID)
    {
        if (pATSC3ALP->verbosity > ENVerbosity_Terse)
        {
            DEBUGPRINT_NOCONSOLE(
                "Segment sequence error (%d, %d, %d) != (%d, %d, %d)\n", 
                pSegment->segmentSequenceNumber, pSegment->packetType, 
                pSegment->SID,
                pPLP->expectedSegmentSequenceNumber, pPLP->packetType, 
                pPLP->SID);
        }

        pPLP->fWaitForNetworkLayerPacketStart = 1;
        pPLP->fDiscontinuity = 1;
        goto out;
    }

    pPLP->expectedSegmentSequenceNumber++;

    if (pPLP->networkLayerPacketBytesUsed + pSegment->payloadLength > 
        pPLP->networkLayerPacketBytesAllocated)
    {
        if (pATSC3ALP->verbosity >= ENVerbosity_Terse)
        {
            DEBUGPRINT_NOCONSOLE("Network layer packet too large\n");
        }
        pPLP->fWaitForNetworkLayerPacketStart = 1;
        pPLP->fDiscontinuity = 1;
        goto out;
    }

    memcpy(pPLP->pNetworkLayerPacket + 
        pPLP->networkLayerPacketBytesUsed, pSegment->pPayload, 
        pSegment->payloadLength);
    pPLP->networkLayerPacketBytesUsed += pSegment->payloadLength;
    pPLP->fError |= pSegment->fError;
    if (pSegment->fDiscontinuity)
    {
        if (!pSegment->segmentSequenceNumber)
        {
            pPLP->fDiscontinuity = 1;
        }
        else
        {
            pPLP->fError = 1;
        }
    }

    if (pSegment->lastSegmentIndicator)
    {
        DispatchNetworkLayerPacket(pATSC3ALP, packetNumber, pPLP); 

        pPLP->fWaitForNetworkLayerPacketStart = 1;
        pPLP->fDiscontinuity = 0;
    }

out:
    return rval;
}

ENStatus
ENATSC3ALPProcessALPPacket(ENATSC3ALPHandle_t hATSC3ALP, uint32_t plpId, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t fError, int32_t fDiscontinuity,
    const uint8_t *pPacket, int32_t packetLength)
{
    ENATSC3ALP *pATSC3ALP = (ENATSC3ALP *)hATSC3ALP;
    uint32_t payloadConfiguration; 
    ENATSC3ALPSegment segment;
    struct _ENATSC3PLP *pPLP;
    ENStatus rval = ENStatus_Success;

    if (plpId > numof(pATSC3ALP->vPLPs))
    {
        if (pATSC3ALP->verbosity > ENVerbosity_Terse)
        {
            DEBUGPRINT_NOCONSOLE("Invalid PLP %d\n", plpId);
        }
        rval = ENStatus_Failure;
        goto out;
    }

    pPLP = pATSC3ALP->vPLPs + plpId;
    if (pATSC3ALP->fBitrateStatisticsEnabled)
    {
        ENATSC3BitrateStatisticsUpdate(&pPLP->stats, pPacketTimeUTC, 
            packetLength);
    }

    if (!pPLP->fEnabled)
    {
        goto out;
    }

    if (packetLength < 1)
    {
        if (pATSC3ALP->verbosity > ENVerbosity_Terse)
        {
            DEBUGPRINT_NOCONSOLE("Short packet\n");
        }
        rval = ENStatus_Failure;
        goto out;
    }

    segment.packetType = pPacket[0] >> 5;
    payloadConfiguration = pPacket[0] & 0x10;

    if (payloadConfiguration == 0)
    {
        uint32_t payloadOffset;

        // Payload configuration = 0: single packet (with optional header)
        // We treat the entire packet as one segment.
        segment.payloadLength = ((pPacket[0] & 0x7) << 8) | pPacket[1];
        segment.segmentSequenceNumber = 0;
        segment.lastSegmentIndicator = 1;
        segment.SID = -1;
        segment.extensionType = 1;
        segment.fError = fError;
        segment.fDiscontinuity = fDiscontinuity;

        if (segment.packetType == ENATSC3ALPPacketType_LinkLayerSignaling)
        {
            // A330 2018 treats the Additional Header for Signaling 
            // Information as part of the header so it is not included in the
            // payload length. It is easier for us to treat it as part of the
            // payload and pass this header on, so we are compensating for that
            // heere. 
            //
            // NOTE that A330 2016 (aka "Korean Mode") treats the Additional
            // Header for Signaling as part of the payload.
            segment.payloadLength += 5;
        }

        if (pPacket[0] & 0x8) // header mode
        {
            if (packetLength < 3)
            {
                if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                {
                    DEBUGPRINT_NOCONSOLE("Short packet\n");
                }
                rval = ENStatus_Failure;
                goto out;
            }

            // Additional Header for Single Packet
            segment.payloadLength |= ((pPacket[2] >> 3)) << 11;
            payloadOffset = 3;

            if (pPacket[2] & 0x2) // Substream Identifer Flag
            {
                if (packetLength <= payloadOffset)
                {
                    if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                    {
                        DEBUGPRINT_NOCONSOLE("Short packet\n");
                    }
                    rval = ENStatus_Failure;
                    goto out;
                }

                segment.SID = pPacket[payloadOffset];
                payloadOffset++;
            }

            if (pPacket[2] & 0x1) // Header extension flag
            {
                uint32_t extensionLengthMinus1;

                if (packetLength <= payloadOffset + 1)
                {
                    if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                    {
                        DEBUGPRINT_NOCONSOLE("Short packet\n");
                    }
                    rval = ENStatus_Failure;
                    goto out;
                }

                // HEF present
                extensionLengthMinus1 = pPacket[payloadOffset + 1];
                segment.extensionType = pPacket[payloadOffset];
                segment.extensionLength = extensionLengthMinus1 + 1;

                if (payloadOffset + 2 + segment.extensionLength > packetLength)
                {
                    if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                    {
                        DEBUGPRINT_NOCONSOLE("HEF payload beyond packet\n");
                    }
                    rval = ENStatus_Failure;
                    goto out;
                }

                segment.pExtension = pPacket + payloadOffset + 2;

                payloadOffset += 2 + extensionLengthMinus1 + 1;
            }
        }
        else
        {
            payloadOffset = 2;
        }

        if (payloadOffset + segment.payloadLength > packetLength)
        {
            if (pATSC3ALP->verbosity > ENVerbosity_Terse)
            {
                DEBUGPRINT_NOCONSOLE(
                    "segment length %d beyond packet end %d\n",
                    segment.payloadLength, packetLength);
            }
            rval = ENStatus_Failure;
            goto out;
        }

        segment.pPayload = pPacket + payloadOffset;
    }
    else
    {
        if (pPacket[0] & 0x8) // Concatenation/segmentation mode
        {
            int32_t payloadStart;
            int32_t payloadLength;
            const uint8_t *pSegmentLength;
            int32_t nSegments;
            int32_t i;

            if (packetLength <= 3)
            {
                if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                {
                    DEBUGPRINT_NOCONSOLE("Short packet\n");
                }
                rval = ENStatus_Failure;
                goto out;
            }

            // Concatenation mode. We treat the payload as a number of
            // segments each belonging to a different network layer packet.
            payloadLength = ((pPacket[0] & 0x7) << 8) | 
                pPacket[1] | ((pPacket[2] >> 4) << 11);
            nSegments = ((pPacket[2] >> 1) & 0x7) + 2; 

            // Advance payloadStart over the segment lengths. There are 
            // nSegments - 1 lengths. The length of the last segment is 
            // implictly signalled as the end of the next-to-last segment until
            // the end of the ALP packet.
            payloadStart = 3 + ((nSegments - 2 + 1) * 12 + 4) / 8;

            // Advance over the substream identifiers, if present. 
            // If present there is one substream identiifer for the ALP packet
            // (i.e. in concatenation mode the SID applies to all input
            // packets.)
            if (pPacket[2] & 1) // Substream Identifier Flag
            {
                segment.SID = pPacket[payloadStart];
                payloadStart++;
            }
            else
            {
                segment.SID = -1;
            }

            segment.extensionType = -1;

            // Ensure that the segment length (and sid if present) array is in 
            // the packet bounds
            if (payloadLength + payloadStart > packetLength)
            {
                if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                {
                    DEBUGPRINT_NOCONSOLE("Concatentation mode: payload len %d > packet payload %d\n",
                        payloadLength, packetLength - payloadStart);
                }
                rval = ENStatus_Failure;
                goto out;
            }

            // Set up the structures for each segment.  Note we're handling two 
            // segments in each iteration of this loop since the segment 
            // lengths are 12 bits each, with 2 lengths stuffed into 3 bytes.
            pSegmentLength = pPacket + 3;
            for (i = 0; i < nSegments - 1; i += 2)
            {
                int32_t segmentLength;

                segmentLength = (pSegmentLength[0] << 4) | 
                    (pSegmentLength[1] >> 4);

                if (payloadStart + segmentLength > packetLength)
                {
                    if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                    {
                        DEBUGPRINT_NOCONSOLE(
                            "Concatenation mode: %d/%d overflows packet\n", 
                            i, nSegments);
                    }
                    rval = ENStatus_Failure;
                    goto out;
                }

                if (payloadStart + segmentLength > payloadLength)
                {
                    if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                    {
                        DEBUGPRINT_NOCONSOLE(
                            "Concatenation mode: %d/%d overflows payload\n", 
                            i, nSegments);
                    }
                    rval = ENStatus_Failure;
                    goto out;
                }

                segment.pPayload = pPacket + payloadStart;
                segment.payloadLength = segmentLength;
                segment.segmentSequenceNumber = 0;
                segment.lastSegmentIndicator = 1;
                segment.fError = fError;
                if (!i)
                {
                    segment.fDiscontinuity = fDiscontinuity;
                }
                else
                {
                    segment.fDiscontinuity = 0;
                }

                ProcessALPSegment(hATSC3ALP, pPacketTimeUTC, packetNumber, 
                    pPLP, &segment);

                payloadStart += segmentLength;
                payloadLength -= segmentLength;

                if (i + 1 < nSegments - 1)
                {
                    segmentLength = ((pSegmentLength[1] & 0xf) << 8) | 
                        pSegmentLength[2];

                    if (payloadStart + segmentLength > packetLength)
                    {
                        if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                        {
                            DEBUGPRINT_NOCONSOLE(
                                "Concatenation mode: %d/%d overflows packet\n", 
                                i, nSegments);
                        }
                        rval = ENStatus_Failure;
                        goto out;
                    }

                    if (payloadStart + segmentLength > payloadLength)
                    {
                        if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                        {
                            DEBUGPRINT_NOCONSOLE(
                                "Concatenation mode: %d/%d overflows payload\n",
                                i, nSegments);
                        }
                        rval = ENStatus_Failure;
                        goto out;
                    }

                    segment.pPayload = pPacket + payloadStart;
                    segment.payloadLength = segmentLength;
                    segment.segmentSequenceNumber = 0;
                    segment.lastSegmentIndicator = 1;
                    segment.fError = fError;
                    segment.fDiscontinuity = 0;

                    ProcessALPSegment(hATSC3ALP, pPacketTimeUTC, packetNumber, 
                        pPLP, &segment);

                    payloadStart += segmentLength;
                    payloadLength -= segmentLength;
                }

                pSegmentLength += 3;
            }

            // There are only nSegments - 1 segment lengths. The length of the 
            // last segment is implied by the total packet payload length.
            segment.pPayload = pPacket + payloadStart;
            segment.payloadLength = payloadLength;
            segment.segmentSequenceNumber = 0;
            segment.lastSegmentIndicator = 1;
            segment.fDiscontinuity = 0;
        }
        else 
        {
            uint32_t payloadOffset;

            if (packetLength < 3)
            {
                if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                {
                    DEBUGPRINT_NOCONSOLE("Short ALP packet\n");
                }
                rval = ENStatus_Failure;
                goto out;
            }

            // Segmentation mode. The payload consists of a single segment of 
            // a multi-segment network layer packet.
            segment.payloadLength = ((pPacket[0] & 0x7) << 8) | pPacket[1];
            segment.segmentSequenceNumber = pPacket[2] >> 3;
            segment.lastSegmentIndicator = pPacket[2] & 0x4;
            segment.extensionType = -1;
            segment.fError = fError;
            if (!segment.segmentSequenceNumber)
            {
                segment.fDiscontinuity = fDiscontinuity;
            }
            else
            {
                segment.fDiscontinuity = 0;
                segment.fError |= fDiscontinuity;
            }

            payloadOffset = 3;

            if (pPacket[2] & 0x2) // Substream Identifer Flag
            {
                if (packetLength < payloadOffset + 1)
                {
                    if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                    {
                        DEBUGPRINT_NOCONSOLE("Short packet\n");
                    }
                    rval = ENStatus_Failure;
                    goto out;
                }

                segment.SID = pPacket[payloadOffset];
                payloadOffset++;
            }
            else
            {
                segment.SID = -1;
            }

            if (pPacket[2] & 0x1) // Header extension flag
            {
                uint32_t extensionLengthMinus1;

                if (packetLength < payloadOffset + 2)
                {
                    if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                    {
                        DEBUGPRINT_NOCONSOLE("Short packet\n");
                    }
                    rval = ENStatus_Failure;
                    goto out;
                }

                // extensionLengthMinus1 is the number of bytes following
                // extensionLengthMinus1 less one byte.
                segment.extensionType = pPacket[payloadOffset];
                extensionLengthMinus1 = pPacket[payloadOffset + 1];
                segment.extensionLength = extensionLengthMinus1 + 1;
        
                if (payloadOffset + 2 + segment.extensionLength > packetLength)
                {
                    if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                    {
                        DEBUGPRINT_NOCONSOLE(
                            "HEF payload beyond end of packet\n");
                    }
                    rval = ENStatus_Failure;
                    goto out;
                }

                segment.pExtension = pPacket + payloadOffset + 2;
                payloadOffset += 2 + extensionLengthMinus1 + 1;
            }

            if (payloadOffset + segment.payloadLength > packetLength)
            {
                if (pATSC3ALP->verbosity > ENVerbosity_Terse)
                {
                    DEBUGPRINT_NOCONSOLE(
                        "Segmentation mode: payload overfows packet\n");
                }
                rval = ENStatus_Failure;
                goto out;
            }

            segment.pPayload = pPacket + payloadOffset;
        }
    }

    ProcessALPSegment(hATSC3ALP, pPacketTimeUTC, packetNumber, pPLP, &segment);

out:
    return rval;
}
