/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include "atsc3lls_aeat.h"

static char *ensys = "atsc30lls";
static char *ensubsys = "aeat";

static int32_t fVerbose = 0;

#define MIN(x, y) ( ( ( x) < (y) ) ? (x) : (y) )
#define MAX(x, y) ( ( ( x) > (y) ) ? (x) : (y) )
#define numof(x) (sizeof(x) / sizeof(x[0]))

static ENStatus
LocationNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_LocationNode *pLocationNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_Location);

    pLocationNode = (ATSC3LLSAEAT_LocationNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSAEATLocationNodeAttribute_type:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue,
                &pLocationNode->typeDataOffset))
            {
                ERRORPRINT("Failed to store type\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*stype: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

static ENStatus
LocationNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_LocationNode *pLocationNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_Location);

    depth = xmlTextReaderDepth(reader);

    pLocationNode = (ATSC3LLSAEAT_LocationNode *)pParentNode;

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(&pAEATNode->dataBuffer, 
        (char *)pValue, sValue, &pLocationNode->locationDataOffset))
    {
        ERRORPRINT("Failed to store location\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*slocation: %.*s\n", depth * 2, "", sValue, pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vLocationAttributes[] = 
{
    {
        .id = ATSC3LLSAEATLocationNodeAttribute_type,
        .sName = sizeof("type") - 1,
        .pName = "type",
        .handler = LocationNodeAttributeHandler,
    },
};
EnXMLParserChildNode vLocationChildNodes[] = 
{
};
EnXMLParserNode locationNodeData = 
{
    .pName = "Location",
    .pAttributes = vLocationAttributes,
    .nAttributes = numof(vLocationAttributes),
    .pChildNodes = vLocationChildNodes,
    .nChildNodes = numof(vLocationChildNodes),
    .textHandler = LocationNodeTextHandler,
};

static ENStatus
EventDescNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_EventDescNode *pEventDescNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_EventDesc);

    pEventDescNode = (ATSC3LLSAEAT_EventDescNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSAEATEventDescNodeAttribute_lang:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pEventDescNode->langDataOffset))
            {
                ERRORPRINT("Failed to store lang\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*slang: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

static ENStatus
EventDescNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_EventDescNode *pEventDescNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_EventDesc);

    pEventDescNode = (ATSC3LLSAEAT_EventDescNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(&pAEATNode->dataBuffer, 
        (char *)pValue, sValue, &pEventDescNode->eventDescDataOffset))
    {
        ERRORPRINT("Failed to store eventDesc\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*seventDesc: %.*s\n", depth * 2, "", sValue, pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vEventDescAttributes[] = 
{
    {
        .id = ATSC3LLSAEATEventDescNodeAttribute_lang,
        .sName = sizeof("lang") - 1,
        .pName = "lang",
        .handler = EventDescNodeAttributeHandler,
    },
};
EnXMLParserChildNode vEventDescChildNodes[] = 
{
};
EnXMLParserNode eventDescNodeData = 
{
    .pName = "EventDesc",
    .pAttributes = vEventDescAttributes,
    .nAttributes = numof(vEventDescAttributes),
    .pChildNodes = vEventDescChildNodes,
    .nChildNodes = numof(vEventDescChildNodes),
    .textHandler = EventDescNodeTextHandler,
};

static ENStatus
EventCodeNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_EventCodeNode *pEventCodeNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_EventCode);

    pEventCodeNode = (ATSC3LLSAEAT_EventCodeNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSAEATEventCodeNodeAttribute_type:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pEventCodeNode->typeDataOffset))
            {
                ERRORPRINT("Failed to store type\n");
                rval = ENStatus_Failure;
                break;
            }

            if (fVerbose)
            {
                LOGPRINT("%*stype: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

static ENStatus
EventCodeNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_EventCodeNode *pEventCodeNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_EventCode);

    pEventCodeNode = (ATSC3LLSAEAT_EventCodeNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(&pAEATNode->dataBuffer, 
        (char *)pValue, sValue, &pEventCodeNode->eventCodeDataOffset))
    {
        ERRORPRINT("Failed to store eventCode\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*seventCode: %.*s\n", depth * 2, "", sValue, pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vEventCodeAttributes[] = 
{
    {
        .id = ATSC3LLSAEATEventCodeNodeAttribute_type,
        .sName = sizeof("type") - 1,
        .pName = "type",
        .handler = EventCodeNodeAttributeHandler,
    },
};
EnXMLParserChildNode vEventCodeChildNodes[] = 
{
};
EnXMLParserNode eventCodeNodeData = 
{
    .pName = "EventCode",
    .pAttributes = vEventCodeAttributes,
    .nAttributes = numof(vEventCodeAttributes),
    .pChildNodes = vEventCodeChildNodes,
    .nChildNodes = numof(vEventCodeChildNodes),
    .textHandler = EventCodeNodeTextHandler,
};

static ENStatus
HeaderNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_HeaderNode *pHeaderNode;
    int depth;
    int64_t seconds;
    int64_t milliseconds;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_Header);

    pHeaderNode = (ATSC3LLSAEAT_HeaderNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSAEATHeaderNodeAttribute_effective:
            if (EnXMLParserDateTimeToSecondsUTC((char *)pValue, sValue, 
                &seconds, &milliseconds))
            {
                ERRORPRINT("Failed to parse DateTime %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                break;
            }
            pHeaderNode->effectiveMillisecondsUTC = seconds * 1000 + 
                milliseconds;
            if (fVerbose)
            {
                LOGPRINT("%*seffective: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSAEATHeaderNodeAttribute_expires:
            if (EnXMLParserDateTimeToSecondsUTC((char *)pValue, sValue, 
                &seconds, &milliseconds))
            {
                ERRORPRINT("Failed to parse DateTime %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                break;
            }
            pHeaderNode->expiresMillisecondsUTC = seconds * 1000 +
                milliseconds;
            if (fVerbose)
            {
                LOGPRINT("%*sexpires: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

static ENStatus
ProcessEventCodeNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_EventCodeNode *pEventCodeNode;
    ATSC3LLSAEAT_HeaderNode *pHeaderNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_Header);

    pHeaderNode = (ATSC3LLSAEAT_HeaderNode *)pParentNode;
    pEventCodeNode = &pHeaderNode->eventCode;

    depth = xmlTextReaderDepth(reader);

    if (fVerbose)
    {
        LOGPRINT("%*sEventCode\n", depth * 2, "");
    }
    pEventCodeNode->base.nodeType = ATSC3LLSAEATNodeType_EventCode;

    rval = EnXMLParserProcessNode(&eventCodeNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)pEventCodeNode);

    return rval;
}

static ENStatus
ProcessEventDescNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_HeaderNode *pHeaderNode;
    ATSC3LLSAEAT_EventDescNode eventDescNode;
    int32_t eventDescNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_Header);
    pHeaderNode = (ATSC3LLSAEAT_HeaderNode *)pParentNode;

    memset(&eventDescNode, 0, sizeof(eventDescNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the header node's eventDsc data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pAEATNode->dataBuffer, 
        sizeof(eventDescNode), &eventDescNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for eventDesc node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pHeaderNode->eventDescs,
        eventDescNodeOffset))
    {
        ERRORPRINT("Failed to store eventDesc node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sEventDesc\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // EventDescNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    eventDescNode.base.nodeType = ATSC3LLSAEATNodeType_EventDesc;
    rval = EnXMLParserProcessNode(&eventDescNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)&eventDescNode);

    memcpy(pAEATNode->dataBuffer.pBuffer + eventDescNodeOffset, &eventDescNode,
        sizeof(eventDescNode));

out:
    return rval;
}

static ENStatus
ProcessLocationNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_HeaderNode *pHeaderNode;
    ATSC3LLSAEAT_LocationNode locationNode;
    int32_t locationNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_Header);
    pHeaderNode = (ATSC3LLSAEAT_HeaderNode *)pParentNode;

    memset(&locationNode, 0, sizeof(locationNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the header node's location data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pAEATNode->dataBuffer, 
        sizeof(locationNode), &locationNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for location node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pHeaderNode->locations,
        locationNodeOffset))
    {
        ERRORPRINT("Failed to store location node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sLocation\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // LocationNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    locationNode.base.nodeType = ATSC3LLSAEATNodeType_Location;
    rval = EnXMLParserProcessNode(&locationNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)&locationNode);

    memcpy(pAEATNode->dataBuffer.pBuffer + locationNodeOffset, &locationNode,
        sizeof(locationNode));

out:
    return rval;
}

EnXMLParserAttribute vHeaderAttributes[] = 
{
    {
        .id = ATSC3LLSAEATHeaderNodeAttribute_effective,
        .sName = sizeof("effective") - 1,
        .pName = "effective",
        .handler = HeaderNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATHeaderNodeAttribute_expires,
        .sName = sizeof("expires") - 1,
        .pName = "expires",
        .handler = HeaderNodeAttributeHandler,
    },
};
EnXMLParserChildNode vHeaderChildNodes[] = 
{
    {
        .id = ATSC3LLSAEATHeaderChildNode_EventCode,
        .sName = sizeof("EventCode") - 1,
        .pName = "EventCode",
        .handler = ProcessEventCodeNode,
    },
    {
        .id = ATSC3LLSAEATHeaderChildNode_EventDesc,
        .sName = sizeof("EventDesc") - 1,
        .pName = "EventDesc",
        .handler = ProcessEventDescNode,
    },
    {
        .id = ATSC3LLSAEATHeaderChildNode_Location,
        .sName = sizeof("Location") - 1,
        .pName = "Location",
        .handler = ProcessLocationNode,
    },
};
EnXMLParserNode headerNodeData = 
{
    .pName = "Header",
    .pAttributes = vHeaderAttributes,
    .nAttributes = numof(vHeaderAttributes),
    .pChildNodes = vHeaderChildNodes,
    .nChildNodes = numof(vHeaderChildNodes),
    .textHandler = NULL,
};

static ENStatus
AEATextNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_AEATextNode *pAEATextNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_AEAText);

    pAEATextNode = (ATSC3LLSAEAT_AEATextNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSAEATAEATextNodeAttribute_lang:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pAEATextNode->langDataOffset))
            {
                ERRORPRINT("Failed to store lang\n");
                rval = ENStatus_Failure;
                break;
            }

            if (fVerbose)
            {
                LOGPRINT("%*saeaText: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

static ENStatus
AEATextNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_AEATextNode *pAEATextNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_AEAText);

    pAEATextNode = (ATSC3LLSAEAT_AEATextNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(&pAEATNode->dataBuffer, 
        (char *)pValue, sValue, &pAEATextNode->aeaTextDataOffset))
    {
        ERRORPRINT("Failed to store aeaText\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*stext: %.*s\n", depth * 2, "", sValue, pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vAEATextAttributes[] = 
{
    {
        .id = ATSC3LLSAEATAEATextNodeAttribute_lang,
        .sName = sizeof("lang") - 1,
        .pName = "lang",
        .handler = AEATextNodeAttributeHandler,
    },
};
EnXMLParserChildNode vAEATextChildNodes[] = 
{
};
EnXMLParserNode aeaTextNodeData = 
{
    .pName = "AEAText",
    .pAttributes = vAEATextAttributes,
    .nAttributes = numof(vAEATextAttributes),
    .pChildNodes = vAEATextChildNodes,
    .nChildNodes = numof(vAEATextChildNodes),
    .textHandler = AEATextNodeTextHandler,
};

static ENStatus
ServiceNameNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_ServiceNameNode *pServiceNameNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_ServiceName);

    pServiceNameNode = (ATSC3LLSAEAT_ServiceNameNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSAEATServiceNameNodeAttribute_lang:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pServiceNameNode->langDataOffset))
            {
                ERRORPRINT("Failed to store lang\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*slang: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

static ENStatus
ServiceNameNodeTextHandler(EnXMLParserNode *pNodeData, const xmlChar *pValue,
    int32_t sValue, xmlTextReaderPtr reader, void *pUserData,
    EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_ServiceNameNode *pServiceNameNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_ServiceName);

    pServiceNameNode = (ATSC3LLSAEAT_ServiceNameNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    if (EnXMLParserDataBufferAddStringAndZeroTerminate(&pAEATNode->dataBuffer, 
        (char *)pValue, sValue, &pServiceNameNode->serviceNameDataOffset))
    {
        ERRORPRINT("Failed to store serviceName\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*stext: %.*s\n", depth * 2, "", sValue, pValue);
    }

out:
    return rval;
}

EnXMLParserAttribute vServiceNameAttributes[] = 
{
    {
        .id = ATSC3LLSAEATServiceNameNodeAttribute_lang,
        .sName = sizeof("lang") - 1,
        .pName = "lang",
        .handler = ServiceNameNodeAttributeHandler,
    },
};
EnXMLParserChildNode vServiceNameChildNodes[] = 
{
};
EnXMLParserNode serviceNameNodeData = 
{
    .pName = "ServiceName",
    .pAttributes = vServiceNameAttributes,
    .nAttributes = numof(vServiceNameAttributes),
    .pChildNodes = vServiceNameChildNodes,
    .nChildNodes = numof(vServiceNameChildNodes),
    .textHandler = ServiceNameNodeTextHandler,
};

static ENStatus
LiveMediaNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_LiveMediaNode *pLiveMediaNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_LiveMedia);

    pLiveMediaNode = (ATSC3LLSAEAT_LiveMediaNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSAEATLiveMediaNodeAttribute_bsid:
            if (pLiveMediaNode->nBSIDs == numof(pLiveMediaNode->vBSIDs))
            {
                ERRORPRINT("Too many bsids\n");
                rval = ENStatus_Failure;
                break;
            }
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pLiveMediaNode->vBSIDs[pLiveMediaNode->nBSIDs]))
            {
                ERRORPRINT("Failed to parse bsid %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                break;
            }
            pLiveMediaNode->nBSIDs++;
            if (fVerbose)
            {
                LOGPRINT("%*sbsid: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSAEATLiveMediaNodeAttribute_serviceId:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pLiveMediaNode->serviceId))
            {
                ERRORPRINT("Failed to parse servieId %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sserviceId: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

static ENStatus
ProcessServiceNameNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_LiveMediaNode *pLiveMediaNode;
    ATSC3LLSAEAT_ServiceNameNode serviceNameNode;
    int32_t serviceNameNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_LiveMedia);
    pLiveMediaNode = (ATSC3LLSAEAT_LiveMediaNode *)pParentNode;

    memset(&serviceNameNode, 0, sizeof(serviceNameNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the rating node's ratingValueAbbrevs data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pAEATNode->dataBuffer, 
        sizeof(serviceNameNode), &serviceNameNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for serviceName node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pLiveMediaNode->serviceNames,
        serviceNameNodeOffset))
    {
        ERRORPRINT("Failed to store serviceName node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sServiceName\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingValueAbbrevNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    serviceNameNode.base.nodeType = ATSC3LLSAEATNodeType_ServiceName;
    rval = EnXMLParserProcessNode(&serviceNameNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)&serviceNameNode);

    memcpy(pAEATNode->dataBuffer.pBuffer + serviceNameNodeOffset, 
        &serviceNameNode, sizeof(serviceNameNode));

out:
    return rval;
}
 
EnXMLParserAttribute vLiveMediaAttributes[] = 
{
    {
        .id = ATSC3LLSAEATLiveMediaNodeAttribute_bsid,
        .sName = sizeof("bsid") - 1,
        .pName = "bsid",
        .handler = LiveMediaNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATLiveMediaNodeAttribute_serviceId,
        .sName = sizeof("serviceId") - 1,
        .pName = "serviceId",
        .handler = LiveMediaNodeAttributeHandler,
    },
};
EnXMLParserChildNode vLiveMediaChildNodes[] = 
{
    {
        .id = ATSC3LLSAEATLiveMediaChildNode_ServiceName,
        .sName = sizeof("ServiceName") - 1,
        .pName = "ServiceName",
        .handler = ProcessServiceNameNode,
    },
};
EnXMLParserNode liveMediaNodeData = 
{
    .pName = "LiveMedia",
    .pAttributes = vLiveMediaAttributes,
    .nAttributes = numof(vLiveMediaAttributes),
    .pChildNodes = vLiveMediaChildNodes,
    .nChildNodes = numof(vLiveMediaChildNodes),
    .textHandler = NULL,
};

static ENStatus
MediaNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_MediaNode *pMediaNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_Media);

    pMediaNode = (ATSC3LLSAEAT_MediaNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSAEATMediaNodeAttribute_lang:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pMediaNode->langDataOffset))
            {
                ERRORPRINT("Failed to store lang\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*slang: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSAEATMediaNodeAttribute_mediaDesc:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pMediaNode->mediaDescDataOffset))
            {
                ERRORPRINT("Failed to store mediaDesc\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*smediaDesc: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSAEATMediaNodeAttribute_mediaType:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pMediaNode->mediaTypeDataOffset))
            {
                ERRORPRINT("Failed to store mediaType\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*smediaType: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSAEATMediaNodeAttribute_url:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pMediaNode->urlDataOffset))
            {
                ERRORPRINT("Failed to store url\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*surl: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSAEATMediaNodeAttribute_alternateUrl:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pMediaNode->alternateUrlDataOffset))
            {
                ERRORPRINT("Failed to store alternateUrl\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*salternateUrl: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSAEATMediaNodeAttribute_contentType:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pMediaNode->contentTypeDataOffset))
            {
                ERRORPRINT("Failed to store contentType\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*scontentType: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSAEATMediaNodeAttribute_contentLength:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pMediaNode->contentLength))
            {
                ERRORPRINT("Failed to parse contentLength %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*scontentLength: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSAEATMediaNodeAttribute_mediaAssoc:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pMediaNode->mediaAssocDataOffset))
            {
                ERRORPRINT("Failed to store mediaAssoc\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*smediaAssoc: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

EnXMLParserAttribute vMediaAttributes[] = 
{
    {
        .id = ATSC3LLSAEATMediaNodeAttribute_lang,
        .sName = sizeof("lang") - 1,
        .pName = "lang",
        .handler = MediaNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATMediaNodeAttribute_mediaDesc,
        .sName = sizeof("mediaDesc") - 1,
        .pName = "mediaDesc",
        .handler = MediaNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATMediaNodeAttribute_mediaType,
        .sName = sizeof("mediaType") - 1,
        .pName = "mediaType",
        .handler = MediaNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATMediaNodeAttribute_url,
        .sName = sizeof("url") - 1,
        .pName = "url",
        .handler = MediaNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATMediaNodeAttribute_alternateUrl,
        .sName = sizeof("alternateUrl") - 1,
        .pName = "alternateUrl",
        .handler = MediaNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATMediaNodeAttribute_contentType,
        .sName = sizeof("contentType") - 1,
        .pName = "contentType",
        .handler = MediaNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATMediaNodeAttribute_contentLength,
        .sName = sizeof("contentLength") - 1,
        .pName = "contentLength",
        .handler = MediaNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATMediaNodeAttribute_mediaAssoc,
        .sName = sizeof("mediaAssoc") - 1,
        .pName = "mediaAssoc",
        .handler = MediaNodeAttributeHandler,
    },
};
EnXMLParserChildNode vMediaChildNodes[] = 
{
};
EnXMLParserNode mediaNodeData = 
{
    .pName = "Media",
    .pAttributes = vMediaAttributes,
    .nAttributes = numof(vMediaAttributes),
    .pChildNodes = vMediaChildNodes,
    .nChildNodes = numof(vMediaChildNodes),
    .textHandler = NULL,
};

static ENStatus
AEANodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_AEANode *pAEANode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_AEA);

    pAEANode = (ATSC3LLSAEAT_AEANode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSAEATAEANodeAttribute_aeaId:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pAEANode->aeaIdDataOffset))
            {
                ERRORPRINT("Failed to store aeaId\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*saeaId: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSAEATAEANodeAttribute_issuer:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pAEANode->issuerDataOffset))
            {
                ERRORPRINT("Failed to store issuer\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sissuer: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSAEATAEANodeAttribute_audience:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pAEANode->audienceDataOffset))
            {
                ERRORPRINT("Failed to store audience\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*saudience: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSAEATAEANodeAttribute_aeaType:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pAEANode->aeaTypeDataOffset))
            {
                ERRORPRINT("Failed to store aeaType\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*saeaType: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSAEATAEANodeAttribute_refAEAId:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pAEANode->refAEAIdDataOffset))
            {
                ERRORPRINT("Failed to store refAEAId\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*srefAEAId: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSAEATAEANodeAttribute_priority:
            if (EnXMLParser_antoul((char *)pValue, sValue, &pAEANode->priority))
            {
                ERRORPRINT("Failed to parse priority %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*spriority: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSAEATAEANodeAttribute_wakeup:
            if (EnXMLParserStringToBool((char *)pValue, sValue, 
                &pAEANode->wakeup))
            {
                ERRORPRINT("Failed to parse boolean %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*swakeup: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

static ENStatus
ProcessAEAHeaderNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_AEANode *pAEANode;
    ATSC3LLSAEAT_HeaderNode *pHeaderNode;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_AEA);
    pAEANode = (ATSC3LLSAEAT_AEANode *)pParentNode;

    pHeaderNode = &pAEANode->header;

    if (EnXMLParserDataOffsetArrayInitialize(&pHeaderNode->eventDescs,
        &pAEATNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate eventDescs array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(&pHeaderNode->locations,
        &pAEATNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate locations array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sHeader\n", xmlTextReaderDepth(reader) * 2, "");
    }
    pHeaderNode->base.nodeType = ATSC3LLSAEATNodeType_Header;

    rval = EnXMLParserProcessNode(&headerNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)pHeaderNode);

out:
    return rval;
}

static ENStatus
ProcessAEAAEATextNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_AEANode *pAEANode;
    ATSC3LLSAEAT_AEATextNode aeaTextNode;
    int32_t aeaTextNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_AEA);
    pAEANode = (ATSC3LLSAEAT_AEANode *)pParentNode;

    memset(&aeaTextNode, 0, sizeof(aeaTextNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the header node's aeaTexts data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pAEATNode->dataBuffer, 
        sizeof(aeaTextNode), &aeaTextNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for media node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pAEANode->aeaTexts,
        aeaTextNodeOffset))
    {
        ERRORPRINT("Failed to store aeaText node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*saeaText\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // AEATextNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    aeaTextNode.base.nodeType = ATSC3LLSAEATNodeType_AEAText;
    rval = EnXMLParserProcessNode(&aeaTextNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)&aeaTextNode);

    memcpy(pAEATNode->dataBuffer.pBuffer + aeaTextNodeOffset, &aeaTextNode,
        sizeof(aeaTextNode));

out:
    return rval;
}

static ENStatus
ProcessAEALiveMediaNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_AEANode *pAEANode;
    ATSC3LLSAEAT_LiveMediaNode *pLiveMediaNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_AEA);

    pAEANode = (ATSC3LLSAEAT_AEANode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    pLiveMediaNode = &pAEANode->liveMedia;

    if (EnXMLParserDataOffsetArrayInitialize(&pLiveMediaNode->serviceNames,
        &pAEATNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate serviceNames array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sLiveMedia\n", depth * 2, "");
    }
    pLiveMediaNode->base.nodeType = ATSC3LLSAEATNodeType_LiveMedia;

    rval = EnXMLParserProcessNode(&mediaNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)pLiveMediaNode);

out:
    return rval;
}

static ENStatus
ProcessAEAMediaNode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode = (ATSC3LLSAEAT_AEATNode *)pUserData;
    ATSC3LLSAEAT_AEANode *pAEANode;
    ATSC3LLSAEAT_MediaNode mediaNode;
    int32_t mediaNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_AEA);
    pAEANode = (ATSC3LLSAEAT_AEANode *)pParentNode;

    memset(&mediaNode, 0, sizeof(mediaNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the rating node's ratingValueAbbrevs data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pAEATNode->dataBuffer, 
        sizeof(mediaNode), &mediaNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for media node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pAEANode->medias,
        mediaNodeOffset))
    {
        ERRORPRINT("Failed to store media node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sMedia\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingValueAbbrevNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    mediaNode.base.nodeType = ATSC3LLSAEATNodeType_Media;
    rval = EnXMLParserProcessNode(&mediaNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)&mediaNode);

    memcpy(pAEATNode->dataBuffer.pBuffer + mediaNodeOffset, &mediaNode,
        sizeof(mediaNode));

out:
    return rval;
}

EnXMLParserAttribute vAEAAttributes[] = 
{
    {
        .id = ATSC3LLSAEATAEANodeAttribute_aeaId,
        .sName = sizeof("aeaId") - 1,
        .pName = "aeaId",
        .handler = AEANodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATAEANodeAttribute_issuer,
        .sName = sizeof("issuer") - 1,
        .pName = "issuer",
        .handler = AEANodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATAEANodeAttribute_audience,
        .sName = sizeof("audience") - 1,
        .pName = "audience",
        .handler = AEANodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATAEANodeAttribute_aeaType,
        .sName = sizeof("aeaType") - 1,
        .pName = "aeaType",
        .handler = AEANodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATAEANodeAttribute_refAEAId,
        .sName = sizeof("refAEAId") - 1,
        .pName = "refAEAId",
        .handler = AEANodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATAEANodeAttribute_priority,
        .sName = sizeof("priority") - 1,
        .pName = "priority",
        .handler = AEANodeAttributeHandler,
    },
    {
        .id = ATSC3LLSAEATAEANodeAttribute_wakeup,
        .sName = sizeof("wakeup") - 1,
        .pName = "wakeup",
        .handler = AEANodeAttributeHandler,
    },
};
EnXMLParserChildNode vAEAChildNodes[] = 
{
    {
        .id = ATSC3LLSAEATAEAChildNode_Header,
        .sName = sizeof("Header") - 1,
        .pName = "Header",
        .handler = ProcessAEAHeaderNode,
    },
    {
        .id = ATSC3LLSAEATAEAChildNode_AEAText,
        .sName = sizeof("AEAText") - 1,
        .pName = "AEAText",
        .handler = ProcessAEAAEATextNode,
    },
    {
        .id = ATSC3LLSAEATAEAChildNode_LiveMedia,
        .sName = sizeof("LiveMedia") - 1,
        .pName = "LiveMedia",
        .handler = ProcessAEALiveMediaNode,
    },
    {
        .id = ATSC3LLSAEATAEAChildNode_Media,
        .sName = sizeof("Media") - 1,
        .pName = "Media",
        .handler = ProcessAEAMediaNode,
    },
};
EnXMLParserNode aeaNodeData = 
{
    .pName = "AEA",
    .pAttributes = vAEAAttributes,
    .nAttributes = numof(vAEAAttributes),
    .pChildNodes = vAEAChildNodes,
    .nChildNodes = numof(vAEAChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessAEANode(EnXMLParserChildNode *pChildNode, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode;
    ATSC3LLSAEAT_AEANode aeaNode;
    int32_t aeaNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_AEAT);
    pAEATNode = (ATSC3LLSAEAT_AEATNode *)pParentNode;

    memset(&aeaNode, 0, sizeof(aeaNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the rating node's ratingValueAbbrevs data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pAEATNode->dataBuffer, 
        sizeof(aeaNode), &aeaNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for aea node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pAEATNode->aeas,
        aeaNodeOffset))
    {
        ERRORPRINT("Failed to store aea node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(&aeaNode.medias,
        &pAEATNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate medias array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(&aeaNode.aeaTexts,
        &pAEATNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate aeaTexts array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (fVerbose)
    {
        LOGPRINT("%*sAEA\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingValueAbbrevNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    aeaNode.base.nodeType = ATSC3LLSAEATNodeType_AEA;
    rval = EnXMLParserProcessNode(&aeaNodeData, reader, pUserData, 
        (EnXMLParserNodeBase *)&aeaNode);

    memcpy(pAEATNode->dataBuffer.pBuffer + aeaNodeOffset, &aeaNode,
        sizeof(aeaNode));

out:
    return rval;
}

static ENStatus
AEATNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSAEAT_AEATNode *pAEATNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSAEATNodeType_AEAT);

    pAEATNode = (ATSC3LLSAEAT_AEATNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSAEATNodeAttribute_aeaTableId:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pAEATNode->dataBuffer, (char *)pValue, sValue, 
                &pAEATNode->aeaTableIdDataOffset))
            {
                ERRORPRINT("Failed to store aeaTableId\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*saeaId: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

EnXMLParserAttribute vAEATAttributes[] = 
{
    {
        .id = ATSC3LLSAEATNodeAttribute_aeaTableId,
        .sName = sizeof("aeaTableId") - 1,
        .pName = "aeaTableId",
        .handler = AEATNodeAttributeHandler,
    },
};
EnXMLParserChildNode vAEATChildNodes[] = 
{
    {
        .id = ATSC3LLSAEATChildNode_AEA,
        .sName = sizeof("AEA") - 1,
        .pName = "AEA",
        .handler = ProcessAEANode,
    },
};
EnXMLParserNode aeatNodeData = 
{
    .pName = "AEAT",
    .pAttributes = vAEATAttributes,
    .nAttributes = numof(vAEATAttributes),
    .pChildNodes = vAEATChildNodes,
    .nChildNodes = numof(vAEATChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessAEATNode(xmlTextReaderPtr reader, ATSC3LLSAEAT_AEATNode *pAEATNode)
{
    ENStatus rval = ENStatus_Success;

    if (fVerbose)
    {
        LOGPRINT("AEAT\n");
    }

    if (EnXMLParserDataOffsetArrayInitialize(&pAEATNode->aeas,
        &pAEATNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate aeas array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    rval = EnXMLParserProcessNode(&aeatNodeData, reader, pAEATNode, 
        (EnXMLParserNodeBase *)pAEATNode);

out:
    return rval;
}

ENStatus
ATSC3LLSParseAEAT(const char *pXML, int32_t sXML, 
    ATSC3LLSAEAT_AEATNode *pParsedAEAT)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pParsedAEAT, 0, sizeof(ATSC3LLSAEAT_AEATNode));
    pParsedAEAT->base.nodeType = ATSC3LLSAEATNodeType_AEAT;

    if (EnXMLParserDataBufferInitialize(&pParsedAEAT->dataBuffer, 8192, 4096))
    {
        ERRORPRINT("Failed to initialize AEAT data buffer\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);
    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "AEAT"))
        {
            int32_t lret;
            lret = ProcessAEATNode(reader, pParsedAEAT);
            if (lret)
            {
                rval = ENStatus_Failure;
                goto errout;
            }
        }
        else
        {
            ERRORPRINT("Top-level object isn't AEAT (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse AEAT XML\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedAEAT->dataBuffer))
        {
            ERRORPRINT("Failed to free AEAT data buffer\n");
        }
    }

    goto out;
}

ENStatus
ATSC3LLSFreeParsedAEAT(ATSC3LLSAEAT_AEATNode *pParsedAEAT)
{
    ENStatus rval = ENStatus_Success;

    if (EnXMLParserDataBufferFree(&pParsedAEAT->dataBuffer))
    {
        ERRORPRINT("Failed to free AEAT data buffer\n");
    }

    return rval;
}

