/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 * '
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include "atsc3lls_onscreenmessage.h"

static char *ensys = "atsc30lls";
static char *ensubsys = "slt";

static int32_t fVerbose = 0;

#define MIN(x, y) ( ( ( x) < (y) ) ? (x) : (y) )
#define MAX(x, y) ( ( ( x) > (y) ) ? (x) : (y) )
#define numof(x) (sizeof(x) / sizeof(x[0]))

static ENStatus
KeepScreenClearNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSOSMN_KeepScreenClearNode *pATSC3LLSOSMNKeepScreenClearNode;
    int depth;
    int64_t seconds;
    int64_t milliseconds;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSOSMNNodeType_KeepScreenClear);

    pATSC3LLSOSMNKeepScreenClearNode = 
        (ATSC3LLSOSMN_KeepScreenClearNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSOSMNKeepScreenClearNodeAttribute_bsid:
            if (EnXMLParserListOfUnsignedShortToEnXMLArray((char *)pValue,
                sValue, &pATSC3LLSOSMNKeepScreenClearNode->bsids))
            {
                ERRORPRINT("Failed to parse bsid %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sbsid: %*.s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSOSMNKeepScreenClearNodeAttribute_serviceId:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pATSC3LLSOSMNKeepScreenClearNode->serviceId))
            {
                ERRORPRINT("Failed to parse serviceId %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sserviceId: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSOSMNKeepScreenClearNodeAttribute_serviceIdRange:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pATSC3LLSOSMNKeepScreenClearNode->serviceIdRange))
            {
                ERRORPRINT("Failed to parse serviceIdRange %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sserviceIdRange: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSOSMNKeepScreenClearNodeAttribute_notificationDuration:
            if (EnXMLParserXMLDurationToSeconds((char *)pValue, sValue, 
                &seconds, &milliseconds))
            {
                ERRORPRINT("Failed to parse duration %.*s\n", sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            pATSC3LLSOSMNKeepScreenClearNode->notificationDurationMilliseconds =
                seconds * 1000 + milliseconds;
            if (fVerbose)
            {
                LOGPRINT("%*snotificationDuration: %.*s\n", depth * 2, "", 
                    sValue, pValue);
            }
            break;
        case ATSC3LLSOSMNKeepScreenClearNodeAttribute_kscFlag:
            if (EnXMLParserStringToBool((char *)pValue, sValue, 
                &pATSC3LLSOSMNKeepScreenClearNode->kscFlag))
            {
                ERRORPRINT("Failed to parse kscFlag %.*s\n", sValue, 
                    pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*skscFlag: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSOSMNKeepScreenClearNodeAttribute_version:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pATSC3LLSOSMNKeepScreenClearNode->version))
            {
                ERRORPRINT("Failed to parse version %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sversion: %*.s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

out:
    return rval;
}

EnXMLParserAttribute vKeepScreenClearAttributes[] = 
{
    {
        .id = ATSC3LLSOSMNKeepScreenClearNodeAttribute_bsid,
        .sName = sizeof("bsid") - 1, 
        .pName = "bsid",
        .handler = KeepScreenClearNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSOSMNKeepScreenClearNodeAttribute_serviceId,
        .sName = sizeof("serviceId") - 1, 
        .pName = "serviceId",
        .handler = KeepScreenClearNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSOSMNKeepScreenClearNodeAttribute_serviceIdRange,
        .sName = sizeof("serviceIdRange") - 1, 
        .pName = "serviceIdRange",
        .handler = KeepScreenClearNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSOSMNKeepScreenClearNodeAttribute_notificationDuration,
        .sName = sizeof("notificationDuration") - 1, 
        .pName = "notificationDuration",
        .handler = KeepScreenClearNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSOSMNKeepScreenClearNodeAttribute_kscFlag,
        .sName = sizeof("kscFlag") - 1, 
        .pName = "kscFlag",
        .handler = KeepScreenClearNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSOSMNKeepScreenClearNodeAttribute_version,
        .sName = sizeof("version") - 1, 
        .pName = "version",
        .handler = KeepScreenClearNodeAttributeHandler,
    },
};
EnXMLParserChildNode vKeepScreenClearChildNodes[] = 
{
};
EnXMLParserNode keepScreenClearNodeData = 
{
    .pName = "KeepScreenClear",
    .pAttributes = vKeepScreenClearAttributes,
    .nAttributes = numof(vKeepScreenClearAttributes),
    .pChildNodes = vKeepScreenClearChildNodes,
    .nChildNodes = numof(vKeepScreenClearChildNodes),
    .textHandler = NULL,
};

static ENStatus
OnscreenMessageNotificationNodeAttributeHandler(
    EnXMLParserAttribute *pAttribute, const xmlChar *pValue, int32_t sValue, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ATSC3LLSOSMNNodeType_OnscreenMessageNotification);

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSOSMNOnscreenMessageNotificationNodeAttribute_XMLNS:
            if (fVerbose)
            {
                LOGPRINT("%*sxmlns: %*.s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

    return rval;
}

static ENStatus
ProcessKeepScreenClearNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSOSMN_OnscreenMessageNotificationNode 
        *pATSC3LLSOSMNOnscreenMessageNotificationNode;
    ATSC3LLSOSMN_KeepScreenClearNode keepScreenClearNode;
    int32_t keepScreenClearNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ATSC3LLSOSMNNodeType_OnscreenMessageNotification);
    pATSC3LLSOSMNOnscreenMessageNotificationNode = 
        (ATSC3LLSOSMN_OnscreenMessageNotificationNode *)pUserData;

    memset(&keepScreenClearNode, 0, sizeof(keepScreenClearNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the osmn node's keepScreenClears offset array.
    if (EnXMLParserDataBufferAllocateSpace(
        &pATSC3LLSOSMNOnscreenMessageNotificationNode->dataBuffer, 
        sizeof(keepScreenClearNode), &keepScreenClearNodeOffset))
    {
        ERRORPRINT(
            "Failed to allocate space for keepScreenClear node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(
        &pATSC3LLSOSMNOnscreenMessageNotificationNode->keepScreenClears, 
        keepScreenClearNodeOffset))
    {
        ERRORPRINT("Failed to store keepScreenClear offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayInitialize(&keepScreenClearNode.bsids,
        &pATSC3LLSOSMNOnscreenMessageNotificationNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate ratingRegionTables array\n");
        rval = ENStatus_Failure;
        goto out;
    }


    if (fVerbose)
    {
        LOGPRINT("%*sKeepScreenClear\n", xmlTextReaderDepth(reader) * 2, "");
    }

    // Since the data buffer could be reallocated, we are using a temporary
    // ratingValueAbbrevNode that lives on the stack. Once we are done
    // processing the node it is committed to the data buffer.
    keepScreenClearNode.base.nodeType = ATSC3LLSOSMNNodeType_KeepScreenClear;
    rval = EnXMLParserProcessNode(&keepScreenClearNodeData, reader,
        pUserData, (EnXMLParserNodeBase *)&keepScreenClearNode);

    memcpy(pATSC3LLSOSMNOnscreenMessageNotificationNode->dataBuffer.pBuffer +
        keepScreenClearNodeOffset, &keepScreenClearNode,
        sizeof(keepScreenClearNode));

out:
    return rval;
}

EnXMLParserAttribute vOnscreenMessageNotificationAttributes[] = 
{
    {
        .id = ATSC3LLSOSMNOnscreenMessageNotificationNodeAttribute_XMLNS,
        .sName = sizeof("xmlns") - 1, 
        .pName = "xmlns",
        .handler = OnscreenMessageNotificationNodeAttributeHandler,
    },
};
EnXMLParserChildNode vOnscreenMessageNotificationChildNodes[] = 
{
    {
        .id = ATSC3LLSOSMNOnscreenMessageNotificationChildNode_KeepScreenClear,
        .sName = sizeof("KeepScreenClear") - 1,
        .pName = "KeepScreenClear",
        .handler = ProcessKeepScreenClearNode,
    },
};
EnXMLParserNode onScreenMessageNotificationNodeData = 
{
    .pName = "OnscreenMessageNotification",
    .pAttributes = vOnscreenMessageNotificationAttributes,
    .nAttributes = numof(vOnscreenMessageNotificationAttributes),
    .pChildNodes = vOnscreenMessageNotificationChildNodes,
    .nChildNodes = numof(vOnscreenMessageNotificationChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessOnscreenMessageNotificationNode(xmlTextReaderPtr reader, 
    ATSC3LLSOSMN_OnscreenMessageNotificationNode 
        *pATSC3LLSOSMNOnscreenMessageNotificationNode)
{
    ENStatus rval = ENStatus_Success;

    if (fVerbose)
    {
        LOGPRINT("OnscreenMessageNotification\n");
    }

    if (EnXMLParserDataOffsetArrayInitialize(
        &pATSC3LLSOSMNOnscreenMessageNotificationNode->keepScreenClears,
        &pATSC3LLSOSMNOnscreenMessageNotificationNode->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate ratingRegionTables array\n");
        rval = ENStatus_Failure;
        goto out;
    }

    rval = EnXMLParserProcessNode(&onScreenMessageNotificationNodeData, 
        reader, pATSC3LLSOSMNOnscreenMessageNotificationNode,
        (EnXMLParserNodeBase *)pATSC3LLSOSMNOnscreenMessageNotificationNode);

out:
    return rval;
}

ENStatus
ATSC3LLSParseOnscreenMessageNotification(const char *pXML, int32_t sXML,
    ATSC3LLSOSMN_OnscreenMessageNotificationNode 
        *pATSC3LLSOSMNOnscreenMessageNotificationNode)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pATSC3LLSOSMNOnscreenMessageNotificationNode, 0, 
        sizeof(ATSC3LLSOSMN_OnscreenMessageNotificationNode));
    pATSC3LLSOSMNOnscreenMessageNotificationNode->base.nodeType = 
        ATSC3LLSOSMNNodeType_OnscreenMessageNotification;

    if (EnXMLParserDataBufferInitialize(
        &pATSC3LLSOSMNOnscreenMessageNotificationNode->dataBuffer, 4096, 4096))
    {
        ERRORPRINT("Failed to initialize OSMN data buffer\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);
    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "OnscreenMessageNotification"))
        {
            int32_t lret;
            lret = ProcessOnscreenMessageNotificationNode(reader, 
                pATSC3LLSOSMNOnscreenMessageNotificationNode);
            if (lret)
            {
                rval = ENStatus_Failure;
                goto errout;
            }
        }
        else
        {
            ERRORPRINT(
                "Top-level object isn't OnscreenMessageNotification (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse OnscreenMessageNotification XML\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(
            &pATSC3LLSOSMNOnscreenMessageNotificationNode->dataBuffer))
        {
            ERRORPRINT("Failed to free OSMN data buffer\n");
        }
    }

    goto out;
}

ENStatus
ATSC3LLSFreeParsedOnscreenMessageNotification(
    ATSC3LLSOSMN_OnscreenMessageNotificationNode
        *pATSC3LLSOSMNOnscreenMessageNotificationNode)
{
    ENStatus rval = ENStatus_Success;

    if (EnXMLParserDataBufferFree(
        &pATSC3LLSOSMNOnscreenMessageNotificationNode->dataBuffer))
    {
        ERRORPRINT("Failed to free OSMN data buffer\n");
    }

    return rval;
}

