/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libenutils/libenutils_status.h>
#include <libendebug/libendebug.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include "atsc3lls_systemtime.h"

static char *ensys = "atsc30lls";
static char *ensubsys = "time";

static int32_t fVerbose = 0;

#define MIN(x, y) ( ( ( x) < (y) ) ? (x) : (y) )
#define MAX(x, y) ( ( ( x) > (y) ) ? (x) : (y) )
#define numof(x) (sizeof(x) / sizeof(x[0]))

static ENStatus
SystemTimeNodeAttributeHandler(EnXMLParserAttribute *pAttribute, 
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader, 
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ATSC3LLSSystemTime_SystemTimeNode *pSystemTimeNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ATSC3LLSSystemTimeNodeType_SystemTime);

    pSystemTimeNode = (ATSC3LLSSystemTime_SystemTimeNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ATSC3LLSSystemTimeNodeAttribute_XMLNS:
            if (fVerbose)
            {
                LOGPRINT("%*sxmlns: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSSystemTimeNodeAttribute_CurrentUtcOffset:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pSystemTimeNode->currentUtcOffset))
            {
                ERRORPRINT("Failed to parse currentUtcOffset %.*s\n",
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*scurrentUtcOffset: %.*s\n", depth * 2, "", 
                    sValue, pValue);
            }
            break;
        case ATSC3LLSSystemTimeNodeAttribute_PtpPrepend:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pSystemTimeNode->ptpPrepend))
            {
                ERRORPRINT("Failed to parse ptpPrepend %.*s\n",
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sptpPrepend: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSSystemTimeNodeAttribute_Leap59:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pSystemTimeNode->leap59))
            {
                ERRORPRINT("Failed to parse leap59 %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sleap59: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSSystemTimeNodeAttribute_Leap61:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pSystemTimeNode->leap61))
            {
                ERRORPRINT("Failed to parse leap61 %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sleap61: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSSystemTimeNodeAttribute_UtcLocalOffset:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pSystemTimeNode->dataBuffer, (char *)pValue, sValue,
                &pSystemTimeNode->utcLocalOffsetDataOffset))
            {
                ERRORPRINT("Failed to store utcLocalOffset\n");
                rval = ENStatus_Failure;
                break;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sutcLocalOffset: %.*s\n", depth* 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSSystemTimeNodeAttribute_DsStatus:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pSystemTimeNode->dsStatus))
            {
                ERRORPRINT("Failed to parse dsStatus %.*s\n",
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sdsStatus: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        case ATSC3LLSSystemTimeNodeAttribute_DsDayOfMonth:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pSystemTimeNode->dsDayOfMonth))
            {
                ERRORPRINT("Failed to parse dsDayOfMonth %.*s\n",
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sdsDayOfMonth: %.*s\n", depth * 2, "", sValue, 
                    pValue);
            }
            break;
        case ATSC3LLSSystemTimeNodeAttribute_DsHour:
            if (EnXMLParser_antoul((char *)pValue, sValue, 
                &pSystemTimeNode->dsHour))
            {
                ERRORPRINT("Failed to parse dsHour %.*s\n",
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            if (fVerbose)
            {
                LOGPRINT("%*sdsHour: %.*s\n", depth * 2, "", sValue, pValue);
            }
            break;
        default:
            if (fVerbose)
            {
                LOGPRINT("Unhandled attribute %d (%s)\n", pAttribute->id, 
                    pAttribute->pName);
            }
            break;
    }

out:
    return rval;
}

EnXMLParserAttribute vSystemTimeAttributes[] = 
{
    {
        .id = ATSC3LLSSystemTimeNodeAttribute_XMLNS, 
        .sName = sizeof("xmlns") - 1, 
        .pName = "xmlns",
        .handler = SystemTimeNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSystemTimeNodeAttribute_CurrentUtcOffset, 
        .sName = sizeof("currentUtcOffset") - 1, 
        .pName = "currentUtcOffset",
        .handler = SystemTimeNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSystemTimeNodeAttribute_PtpPrepend, 
        .sName = sizeof("ptpPrepend") - 1, 
        .pName = "ptpPrepend",
        .handler = SystemTimeNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSystemTimeNodeAttribute_Leap59, 
        .sName = sizeof("leap59") - 1, 
        .pName = "leap59",
        .handler = SystemTimeNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSystemTimeNodeAttribute_Leap61, 
        .sName = sizeof("leap61") - 1, 
        .pName = "leap61",
        .handler = SystemTimeNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSystemTimeNodeAttribute_UtcLocalOffset, 
        .sName = sizeof("utcLocalOffset") - 1, 
        .pName = "utcLocalOffset",
        .handler = SystemTimeNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSystemTimeNodeAttribute_DsStatus, 
        .sName = sizeof("dsStatus") - 1, 
        .pName = "dsStatus",
        .handler = SystemTimeNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSystemTimeNodeAttribute_DsDayOfMonth, 
        .sName = sizeof("dsDayOfMonth") - 1, 
        .pName = "dsDayOfMonth",
        .handler = SystemTimeNodeAttributeHandler,
    },
    {
        .id = ATSC3LLSSystemTimeNodeAttribute_DsHour, 
        .sName = sizeof("dsHour") - 1, 
        .pName = "dsHour",
        .handler = SystemTimeNodeAttributeHandler,
    },
};
EnXMLParserChildNode vSystemTimeChildNodes[] = 
{
};
EnXMLParserNode systemTimeNodeData = 
{
    .pName = "SystemTime",
    .pAttributes = vSystemTimeAttributes,
    .nAttributes = numof(vSystemTimeAttributes),
    .pChildNodes = vSystemTimeChildNodes,
    .nChildNodes = numof(vSystemTimeChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessSystemTimeNode(xmlTextReaderPtr reader, void *pUserData,
    ATSC3LLSSystemTime_SystemTimeNode *pSystemTimeNode)
{
    ENStatus rval = ENStatus_Success;

    if (fVerbose)
    {
        LOGPRINT("SystemTime\n");
    }

    rval = EnXMLParserProcessNode(&systemTimeNodeData, reader, pUserData,
        (EnXMLParserNodeBase *)pSystemTimeNode);

    return rval;
}

ENStatus
ATSC3LLSParseSystemTime(const char *pSystemTime, int32_t sSystemTime, 
    ATSC3LLSSystemTime_SystemTimeNode *pParsedSystemTime)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer = 0;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pParsedSystemTime, 0, sizeof(ATSC3LLSSystemTime_SystemTimeNode));
    pParsedSystemTime->base.nodeType = ATSC3LLSSystemTimeNodeType_SystemTime;

    if (EnXMLParserDataBufferInitialize(&pParsedSystemTime->dataBuffer,
        4096, 4096))
    {
        ERRORPRINT("Failed to initialize SystemTime data buffer\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    reader = xmlReaderForMemory(pSystemTime, sSystemTime, NULL, NULL, 0);
    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "SystemTime"))
        {
            int32_t lret;

            lret = ProcessSystemTimeNode(reader, NULL, pParsedSystemTime);
            if (lret)
            {
                rval = ENStatus_Failure;
                goto errout;
            }
        }
        else
        {
            ERRORPRINT("Top-level object isn't SystemTime (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Failed to parse SystemTime XML\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedSystemTime->dataBuffer))
        {
            ERRORPRINT("Failed to free SystemTime data buffer\n");
        }
    }

    goto out;
}

ENStatus
ATSC3LLSFreeParsedSystemTime(
    ATSC3LLSSystemTime_SystemTimeNode *pParsedSystemTime)
{
    ENStatus rval = ENStatus_Success;

    if (EnXMLParserDataBufferFree(&pParsedSystemTime->dataBuffer))
    {
        ERRORPRINT("Failed to free SystemTime data buffer\n");
    }

    return rval;
}

