/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <errno.h>
#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include "atsc3routesls_envelope.h"

static char *ensys = "routesls";
static char *ensubsys = "envelope";

static int32_t fVerbose = 0;

#define VERBOSEPRINT(x, ...) \
    if (fVerbose) \
    { \
        LOGPRINT(x, ##__VA_ARGS__); \
    }

static ENStatus
ItemNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSEnvelopeNode *pEnvelopeNode = 
        (ENATSC3ROUTESLSEnvelopeNode *)pUserData;
    ENATSC3ROUTESLSEnvelopeItemNode *pItemNode;
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == ENATSC3ROUTESLSEnvelopeNodeType_item);

    pItemNode = (ENATSC3ROUTESLSEnvelopeItemNode *)pParentNode;

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSEnvelopeItemNodeAttribute_metadataURI:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pEnvelopeNode->dataBuffer, (char *)pValue, sValue,
                &pItemNode->metadataURIDataOffset))
            {
                ERRORPRINT("Failed to store metadataURI\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*smetadataURI: %.*s\n", depth * 2, "", sValue, 
                pValue);
            break;
        case ENATSC3ROUTESLSEnvelopeItemNodeAttribute_version:
            if (EnXMLParser_antoul((char *)pValue, sValue, &pItemNode->version))
            {
                ERRORPRINT("Failed to parse version %.*s\n", 
                    sValue, pValue);
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*sversion: %.*s\n", depth * 2, "", sValue, pValue);
            break;
        case ENATSC3ROUTESLSEnvelopeItemNodeAttribute_contentType:
            if (EnXMLParserDataBufferAddStringAndZeroTerminate(
                &pEnvelopeNode->dataBuffer, (char *)pValue, sValue,
                &pItemNode->contentTypeDataOffset))
            {
                ERRORPRINT("Failed to store contentType\n");
                rval = ENStatus_Failure;
                goto out;
            }
            VERBOSEPRINT("%*scontentType: %.*s\n", depth * 2, "", sValue, 
                pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, 
                pAttribute->pName);
            break;
    }

out:
    return rval;
}

static EnXMLParserAttribute vItemAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSEnvelopeItemNodeAttribute_metadataURI,
        .sName = sizeof("metadataURI") - 1,
        .pName = "metadataURI",
        .handler = ItemNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSEnvelopeItemNodeAttribute_version,
        .sName = sizeof("version") - 1,
        .pName = "version",
        .handler = ItemNodeAttributeHandler,
    },
    {
        .id = ENATSC3ROUTESLSEnvelopeItemNodeAttribute_contentType,
        .sName = sizeof("contentType") - 1,
        .pName = "contentType",
        .handler = ItemNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vItemChildNodes[] =
{
};
static EnXMLParserNode itemNodeData = 
{
    .pName = "item",
    .pAttributes = vItemAttributes,
    .nAttributes = numof(vItemAttributes),
    .pChildNodes = vItemChildNodes,
    .nChildNodes = numof(vItemChildNodes),
    .textHandler = NULL,
};

static ENStatus
MetadataEnvelopeNodeAttributeHandler(EnXMLParserAttribute *pAttribute,
    const xmlChar *pValue, int32_t sValue, xmlTextReaderPtr reader,
    void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    int depth;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ENATSC3ROUTESLSEnvelopeNodeType_metadataEnvelope);

    depth = xmlTextReaderDepth(reader);

    switch (pAttribute->id)
    {
        case ENATSC3ROUTESLSEnvelopeNodeAttribute_XMLNS:
            VERBOSEPRINT("%*sxmlns: %.*s\n", depth * 2, "", sValue, pValue);
            break;
        default:
            VERBOSEPRINT("%*sUnhandled attribute %d (%s)\n", depth * 2, "", 
                pAttribute->id, pAttribute->pName);
            break;
    }

    return rval;
}

static ENStatus
ProcessItemNode(EnXMLParserChildNode *pChildNode, 
    xmlTextReaderPtr reader, void *pUserData, EnXMLParserNodeBase *pParentNode)
{
    ENATSC3ROUTESLSEnvelopeNode *pEnvelopeNode;
    ENATSC3ROUTESLSEnvelopeItemNode itemNode;
    int32_t itemNodeOffset;
    ENStatus rval = ENStatus_Success;

    assert(pParentNode->nodeType == 
        ENATSC3ROUTESLSEnvelopeNodeType_metadataEnvelope);
    pEnvelopeNode = (ENATSC3ROUTESLSEnvelopeNode *)pParentNode;

    memset(&itemNode, 0, sizeof(itemNode));

    // Allocate space for the node from the data buffer and store the
    // offset in the data offset array.
    if (EnXMLParserDataBufferAllocateSpace(&pEnvelopeNode->dataBuffer, 
        sizeof(ENATSC3ROUTESLSEnvelopeItemNode), &itemNodeOffset))
    {
        ERRORPRINT("Failed to allocate space for item node\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (EnXMLParserDataOffsetArrayAppendOffset(&pEnvelopeNode->items,
        itemNodeOffset))
    {
        ERRORPRINT("Failed to store item node offset\n");
        rval = ENStatus_Failure;
        goto out;
    }

    VERBOSEPRINT("%*sItem\n", xmlTextReaderDepth(reader) * 2, "");

    // Since the data buffer could be reallocated, we are using a temporary
    // itemNode that lives on the stack. Once we are done processing the node 
    // it is committed to the data buffer.
    itemNode.base.nodeType = ENATSC3ROUTESLSEnvelopeNodeType_item;
    rval = EnXMLParserProcessNode(&itemNodeData, reader, 
        pUserData, (EnXMLParserNodeBase *)&itemNode);

    memcpy(pEnvelopeNode->dataBuffer.pBuffer + itemNodeOffset,
        &itemNode, sizeof(itemNode));

out:
    return rval;
}

static EnXMLParserAttribute vMetadataEnvelopeAttributes[] = 
{
    {
        .id = ENATSC3ROUTESLSEnvelopeNodeAttribute_XMLNS,
        .sName = sizeof("xmlns") - 1,
        .pName = "xmlns",
        .handler = MetadataEnvelopeNodeAttributeHandler,
    },
};
static EnXMLParserChildNode vMetadataEnvelopeChildNodes[] =
{
    {
        .id = ENATSC3ROUTESLSEnvelopeChildNode_item,
        .sName = sizeof("item") - 1,
        .pName = "item",
        .handler = ProcessItemNode,
    },
};
static EnXMLParserNode metadataEnvelopeNodeData = 
{
    .pName = "metadataEnvelope",
    .pAttributes = vMetadataEnvelopeAttributes,
    .nAttributes = numof(vMetadataEnvelopeAttributes),
    .pChildNodes = vMetadataEnvelopeChildNodes,
    .nChildNodes = numof(vMetadataEnvelopeChildNodes),
    .textHandler = NULL,
};

static ENStatus
ProcessMetadataEnvelopeNode(xmlTextReaderPtr reader, 
    ENATSC3ROUTESLSEnvelopeNode *pParsedEnvelope)
{
    ENStatus rval = ENStatus_Success;

    VERBOSEPRINT("%*smetadataEnvelope\n", xmlTextReaderDepth(reader) * 2, "");

    rval = EnXMLParserProcessNode(&metadataEnvelopeNodeData, reader, 
        pParsedEnvelope, (EnXMLParserNodeBase *)pParsedEnvelope);

    return rval;
}

ENStatus
ENATSC3ROUTESLSEnvelopeParse(const char *pXML, int32_t sXML, 
    ENATSC3ROUTESLSEnvelopeNode *pParsedEnvelope)
{
    xmlTextReaderPtr reader = NULL;
    int32_t fHaveDataBuffer;
    int xmlret;
    ENStatus rval = ENStatus_Success;

    memset(pParsedEnvelope, 0, sizeof(ENATSC3ROUTESLSEnvelopeNode));
    reader = xmlReaderForMemory(pXML, sXML, NULL, NULL, 0);

    if (!reader)
    {
        ERRORPRINT("Failed to create xml reader\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (EnXMLParserDataBufferInitialize(&pParsedEnvelope->dataBuffer, 4096, 
        4096))
    {
        ERRORPRINT("Failed to initialize Envelope data buffer\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveDataBuffer = 1;

    if (EnXMLParserDataOffsetArrayInitialize(&pParsedEnvelope->items,
        &pParsedEnvelope->dataBuffer, 16, 16))
    {
        ERRORPRINT("Failed to allocate Items array\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    xmlret = xmlTextReaderRead(reader);
    if (xmlret == 1)
    {
        int nodeType;
        const xmlChar *pLocalName;

        nodeType = xmlTextReaderNodeType(reader);
        pLocalName = xmlTextReaderConstLocalName(reader);

        if (nodeType == XML_READER_TYPE_ELEMENT && 
            !strcmp((char *)pLocalName, "metadataEnvelope"))
        {
            pParsedEnvelope->base.nodeType = 
                ENATSC3ROUTESLSEnvelopeNodeType_metadataEnvelope;
            ProcessMetadataEnvelopeNode(reader, pParsedEnvelope);
        }
        else
        {
            ERRORPRINT("Top-level object isn't metadataEnvelope (%d/%s)\n",
                nodeType, pLocalName);
            rval = ENStatus_Failure;
            goto errout;
        }
    }
    else
    {
        ERRORPRINT("Couldn't find metadataEnvelope\n");
        rval = ENStatus_Failure;
        goto errout;
    }

out:
    if (reader)
    {
        xmlFreeTextReader(reader);
    }

    return rval;

errout:
    if (fHaveDataBuffer)
    {
        if (EnXMLParserDataBufferFree(&pParsedEnvelope->dataBuffer))
        {
            ERRORPRINT("Failed to free Envelope data buffer\n");
        }
    }

    goto out;
}

ENStatus
ENATSC3ROUTESLSEnvelopeFree(ENATSC3ROUTESLSEnvelopeNode *pParsedEnvelope)
{
    ENStatus rval = ENStatus_Success;

    if (EnXMLParserDataBufferFree(&pParsedEnvelope->dataBuffer))
    {
        ERRORPRINT("Failed to free Envelope data buffer\n");
    }

    return rval;
}
