/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com). 
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

#include <libmultipartparser/multipart_parser.h>

#include <libxml2/libxml/xmlreader.h>

#include <libendebug/libendebug.h>
#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>

#include <enatsc3-common/enatsc3.h>
#include <libenatsc3utils/enxml.h>

#include <enatsc3-common/enatsc3.h>
#include "atsc3routesls_stsid.h"
#include "atsc3routesls_envelope.h"

#include "atsc3routesls_mime.h"

static char *ensys = "routesls";
static char *ensubsys = "mime";

// NAB_2019_SINCLAIR_MIME_HEADER_ERROR_WAR enables code that will attempt to 
// locate the boundary string even if it cannot be extracted from the 
// Content-Type header line.
// #define NAB_2019_SINCLAIR_MIME_HEADER_ERROR_WAR 1

typedef struct
{
    enum
    {
        Header_Invalid = 0,
        Header_ContentType = 1,
        Header_ContentLocation = 2,
    } currentHeader;

    char strContentType[128];
    char strContentLocation[128];
    const char *pPartStart;
    const char *pPartData;

    ENATSC3ROUTESLS_ROUTEMIME_XMLDocumentCallback XMLDocumentCallback;
    void *pCbUserData;

    // rval is set to ENStatus_Failure if any callback for any docuemnt
    // returns a status other than Success;
    ENStatus rval;
} ROUTEMIMEParser;

static inline void
ResetMIMEParser(ROUTEMIMEParser *pMIMEParser)
{
    pMIMEParser->currentHeader = Header_Invalid;
    pMIMEParser->strContentType[0] = 0;
    pMIMEParser->strContentLocation[0] = 0;
    pMIMEParser->pPartData = NULL;
}

int 
HeaderNameCallback(multipart_parser* p, const char *at, size_t length, 
    void *userdata)
{
    ROUTEMIMEParser *pMIMEParser = (ROUTEMIMEParser *)userdata;

    if (length == sizeof("content-location") - 1 && 
        !strncasecmp(at, "content-location", sizeof("content-location") - 1))
    {
        pMIMEParser->currentHeader = Header_ContentLocation;
    }
    else if (length == sizeof("content-type") - 1 && 
        !strncasecmp(at, "content-type", sizeof("content-type") - 1))
    {
        pMIMEParser->currentHeader = Header_ContentType;
    }
    else
    {
        pMIMEParser->currentHeader = Header_Invalid;
    }
  
    return 0;
}

int 
HeaderValueCallback(multipart_parser* p, const char *at, size_t length, 
    void *userdata)
{
    ROUTEMIMEParser *pMIMEParser = (ROUTEMIMEParser *)userdata;
    int rval = 0;

    if (pMIMEParser->currentHeader == Header_ContentType)
    {
        if (length >= sizeof(pMIMEParser->strContentType))
        {
            ERRORPRINT("Content-Type too long: %zd\n", length);
            rval = 1;
            goto out;
        }
        strncpy(pMIMEParser->strContentType, at, length);
        pMIMEParser->strContentType[length] = 0;
    }
    else if (pMIMEParser->currentHeader == Header_ContentLocation)
    {
        if (length >= sizeof(pMIMEParser->strContentLocation))
        {
            ERRORPRINT("Content-Length too long: %zd\n", length);
            rval = 1;
            goto out;
        }
        strncpy(pMIMEParser->strContentLocation, at, length);
        pMIMEParser->strContentLocation[length] = 0;
    }

out:
   return rval;
}

int 
MIMEHeadersCompleteCallback(multipart_parser* p, const char *at, void *userdata)
{
    ROUTEMIMEParser *pMIMEParser = (ROUTEMIMEParser *)userdata;
    int rval = 0;

    pMIMEParser->pPartData = at;

    return rval;
}

int 
MIMEPartStartCallback(multipart_parser* p, const char *at, void *userdata)
{
    ROUTEMIMEParser *pMIMEParser = (ROUTEMIMEParser *)userdata;
    int rval = 0;

    // The parser will still be pointing at the the LF terminating the boundary
    if (*at == '\n')
    {
        at++;
    }

    pMIMEParser->pPartStart = at;

    return rval;
}

int 
MIMEPartEndCallback(multipart_parser* p, const char *at, void *userdata)
{
    ROUTEMIMEParser *pMIMEParser = (ROUTEMIMEParser *)userdata;
    int rval = 0;

    if (pMIMEParser->XMLDocumentCallback)
    {
        rval = pMIMEParser->XMLDocumentCallback(pMIMEParser->pPartData,
            at - pMIMEParser->pPartData, pMIMEParser->pPartStart, 
            at - pMIMEParser->pPartStart, pMIMEParser->strContentType,
            pMIMEParser->strContentLocation, pMIMEParser->pCbUserData);
        if (rval)
        {
            LOGPRINT("MIME part %s rejected\n", 
                pMIMEParser->strContentLocation);
            pMIMEParser->rval = 1;
        }
    }

    ResetMIMEParser(pMIMEParser);

    return rval;
}

static inline int32_t
SkipFoldedWhitespace(const char *pDocument, int32_t length)
{
    int32_t pos = 0;

    // Allow folding
    if (pos + 2 < length && pDocument[0] == '\r' && pDocument[1] == '\n')
    {
        pDocument += 2;
        pos += 2;
    }

    // Skip over optional whitespace
    while (pos < length && isspace(*pDocument))
    {
        pDocument++;
        pos++;
    }

    return pos;
}

// get_next_header processes one line in a document pointed to by pDocument
// and having length of the length argument. Headers are expected to take
// the form:
//
//     name:value;argname=argvalue;argname=argvalue;...
//
// whitespace is permitted after the colon following name and after semicolons. 
//
// Note that for folded lines RFCXXXX allows crlf to be inserted prior
// to whitespace. However in order to accomodate documents observed in actual
// broadcasts this implementation allows crlf to occur where whitespace is 
// allowed but does not actually require whitespace to follow the crlf.
static int32_t 
GetNextHeader(const char *pDocument, int32_t length,  int32_t *pNameLen, 
    const char **ppValueStart, int *pValueLen)
{
    int32_t pos = 0;
    int32_t nameLen = 0;
    int32_t valueLen = 0;
    int32_t nb;
    char c;

    *ppValueStart = NULL;
    *pNameLen = 0;
    *pValueLen = 0;

    while (pos < length)
    {
        c = *(pDocument++);
        if (c == ':')
        {
            pos++;
            break;
        }
        else if (c == '\r')
        {
            if (pos + 1 < length)
            {
                if (*pDocument == '\n')
                {
                    // This is a malformed line - no value. Skip over it
                    pos += 2;
                    goto out;
                }
            }
        }
        pos++;
        nameLen++;
    }

    nb = SkipFoldedWhitespace(pDocument, length);
    pDocument += nb;
    length -= nb;
    pos += nb;

    *ppValueStart = pDocument;

    // Look for CRLF terminating value
    while (pos < length)
    {
        c = *(pDocument++);

        // Check for folded lines. There *should* be whitespace following 
        // the crlf, but at least in some current bitstreams it is missing. 
        // So assume any ";\r\n" sequence is a folded line.
        if (c == ';')
        {
            nb = SkipFoldedWhitespace(pDocument, length);
            pDocument += nb;
            pos += nb;
            valueLen += nb;
        }

        // Check for CRLF terminating the header line
        if (c == '\r' && pos + 1 < length  && *pDocument == '\n')
        {
            pos += 2;
            break;
        }

        pos++;
        valueLen++;
    }

out:
    *pNameLen = nameLen;
    *pValueLen = valueLen;

    return pos;
}

#ifdef NAB_2019_SINCLAIR_MIME_HEADER_ERROR_WAR
int32_t
ScanForBoundary(const char *pXML, int32_t sXML, const char **ppBoundary, 
    int32_t *pBoundaryLen)
{
    int32_t boundaryLen = 0;
    int32_t rval = 1;

    if (sXML >= 2 && pXML[0] == '-' && pXML[1] == '-')
    {
        sXML -= 2;
        pXML += 2;
        *ppBoundary = pXML;

        while (sXML >= 2 && pXML[0] != '\r' && pXML[1] != '\n')
        {
            pXML++;
            sXML--;
            boundaryLen++;
        }

        if (pXML[0] == '\r' && pXML[1] == '\n')
        {
            *pBoundaryLen = boundaryLen;
            rval = 0;
        }
    }

    return rval;
}
#endif

ENStatus
ENATSC3ROUTESLSMultipartMIMEHeaderParse(const char *pXML, int32_t sXML, 
    ENATSC3ROUTESLSMultipartMIMEHeaderInfo_t *pInfo, 
    const char **ppFirstPartStart, int32_t *pNBytesRemaining)
{
    const char *pBoundary = NULL;

#ifdef NAB_2019_SINCLAIR_MIME_HEADER_ERROR_WAR
    int32_t boundaryLen;
#endif
    ENStatus rval = ENStatus_Success;

    pInfo->pContentType = NULL;
    pInfo->nParameters = 0;

    // Process MIME headers. The primary item we need to extract is the 
    // boundary string, but also verify that this is in fact a multipart
    // document.
    while (sXML > 2)
    {
        int32_t consumed;
        int32_t nameLen;
        const char *pValue;
        int32_t valueLen;

        // End of headers
        if (pXML[0] == '\r' && pXML[1] == '\n')
        {
            pXML += 2;
            sXML -= 2;
            break;
        }

        consumed = GetNextHeader(pXML, sXML, &nameLen, &pValue,
            &valueLen);
        if (pValue)
        {
            if (nameLen == sizeof("content-type") - 1 &&
                !strncasecmp(pXML, "content-type", 
                    sizeof("content-type") - 1))
            {
                pInfo->pContentType = pValue;
                pInfo->contentTypeLen = 0;

                while (valueLen && *pValue != ';' && !isspace(*pValue))
                {
                    pInfo->contentTypeLen++;
                    valueLen--;
                    pValue++;
                }

                // Parse content-type arguments
                while (valueLen)
                {
                    const char *pArgNameStart;
                    int32_t argNameLen = 0;
                    const char *pArgValueStart;
                    int32_t argValueLen = 0;
                    int32_t fQuoted = 0;

                    while (isspace(*pValue) && valueLen)
                    {
                        pValue++;
                        valueLen--;
                    }

                    if (*pValue == ';')
                    {
                        pValue++;
                        valueLen--;
                        while (isspace(*pValue) && valueLen)
                        {
                            pValue++;
                            valueLen--;
                        }
                    }

                    if (!valueLen)
                    {
                        break;
                    }

                    pArgNameStart = pValue;
                    while (*pValue != '=' && valueLen)
                    {
                        pValue++;
                        valueLen--;
                        argNameLen++;
                    }

                    if (!valueLen)
                    {
                        ERRORPRINT("Unexpected end of argument name\n");
                        rval = ENStatus_Failure;
                        goto out;
                    }

                    pValue++;
                    valueLen--;

                    if (*pValue == '"')
                    {
                        if (valueLen)
                        {
                            fQuoted = 1;
                            pValue++;
                            valueLen--;
                        }
                        else
                        {
                            ERRORPRINT("Unexpected termination in boundary\n");
                            rval = ENStatus_Failure;
                            goto out;
                        }
                    }

                    pArgValueStart = pValue;

                    if (fQuoted)
                    {
                        while (valueLen && *pValue != '"')
                        {
                            pValue++;
                            valueLen--;
                            argValueLen++;
                        }

                        if (!valueLen)
                        {
                            ERRORPRINT(
                                "Failed to find closing quote in boundary\n");
                            rval = ENStatus_Failure;
                            goto out;
                        }

                        pValue++;
                        valueLen--;
                    }
                    else
                    {
                        while (valueLen && *pValue != ';')
                        {
                            pValue++;
                            valueLen--;
                            argValueLen++;
                        }
                    }

                    if (pInfo->nParameters < numof(pInfo->vParameters))
                    {
                        pInfo->vParameters[pInfo->nParameters].pName = 
                            pArgNameStart;
                        pInfo->vParameters[pInfo->nParameters].nameLen =
                            argNameLen;
                        pInfo->vParameters[pInfo->nParameters].pValue = 
                            pArgValueStart;
                        pInfo->vParameters[pInfo->nParameters].valueLen =
                            argValueLen;
                        pInfo->nParameters++;
                    }

                    if (argNameLen == sizeof("boundary") - 1 &&
                        !strncasecmp(pArgNameStart, "boundary", 
                        sizeof("boundary") - 1))
                    {
                        pBoundary = pArgValueStart;
#ifdef NAB_2019_SINCLAIR_MIME_HEADER_ERROR_WAR
                        boundaryLen = argValueLen;
#endif
                    }
                }
            }
        }

        pXML += consumed;
        sXML -= consumed;
    }

    // Skip over any extra CRLFs which may be separating the message headers
    // from the body. It's not clear whether this is allowed in the various
    // MIME specs, bit it's currently the case in Santa Barbara's ROUTE
    // and there's no harm in being a little permissive.
    while (sXML >= 2 && pXML[0] == '\r' && pXML[1] == '\n')
    {
        sXML -= 2;
        pXML += 2;
    }

    if (!pBoundary)
    {
        ERRORPRINT("Failed to find boundary in MIME headers\n");

#ifdef NAB_2019_SINCLAIR_MIME_HEADER_ERROR_WAR
        // If we couldn't find the boundary string in the MIME headers then see 
        // if we can find it by scanning the document. In pXML should now point 
        // to a boundary, so see if we can use it.
        if (ScanForBoundary(pXML, sXML, &pBoundary, &boundaryLen))
        {
            ERRORPRINT("Candidate boundary string not found via scan\n");
            rval = ENStatus_Failure;
            goto out;
        }
        ERRORPRINT("Using scanned boundary %.*s\n", boundaryLen, pBoundary);
#else
        rval = ENStatus_Failure;
        goto out;
#endif
    }

    *pNBytesRemaining = sXML;
    *ppFirstPartStart = pXML;

out:
    return rval;
}

ENStatus
ENATSC3ROUTESLSMultipartMIMEPartsParse(const char *pXML, int32_t sXML, 
    const char *pBoundary, int32_t boundaryLen,  
    ENATSC3ROUTESLS_ROUTEMIME_XMLDocumentCallback XMLDocumentCallback, 
    void *pUserData)
{
    ROUTEMIMEParser mimeParser;
    multipart_parser_settings multipartParserSettings;
    multipart_parser* pMultipartParser = NULL;
    int32_t nb;
    ENStatus rval = ENStatus_Success;

    // Now process the MIME parts
    mimeParser.XMLDocumentCallback = XMLDocumentCallback;
    mimeParser.pCbUserData = pUserData;
    mimeParser.rval = ENStatus_Success;
    ResetMIMEParser(&mimeParser);

    memset(&multipartParserSettings, 0, sizeof(multipart_parser_settings));
    multipartParserSettings.on_header_field = HeaderNameCallback;
    multipartParserSettings.on_header_value = HeaderValueCallback;
    multipartParserSettings.on_headers_complete = MIMEHeadersCompleteCallback;
    multipartParserSettings.on_part_data_begin = MIMEPartStartCallback;
    multipartParserSettings.on_part_data_end = MIMEPartEndCallback;
    multipartParserSettings.userdata = &mimeParser;

    char strBoundary[128] = "--";
    if (boundaryLen + 2 >= sizeof(strBoundary))
    {
        ERRORPRINT("Boundary string too long\n");
        rval = ENStatus_Failure;
        goto out;
    }
    strncpy(strBoundary + 2, pBoundary, boundaryLen);
    strBoundary[boundaryLen + 2] = 0;

    pMultipartParser = multipart_parser_init(strBoundary,
        &multipartParserSettings);

    nb = multipart_parser_execute(pMultipartParser, pXML, sXML);
    if (nb != sXML)
    {
        ERRORPRINT("Failed to fully parse multipart mime (%d < %d)\n", nb, 
            sXML);
        rval = ENStatus_Failure;
    }

    if (mimeParser.rval)
    {
        ERRORPRINT("One or more documents rejected\n");
        rval = ENStatus_Failure;
    }

    if (pMultipartParser)
    {
        multipart_parser_free(pMultipartParser);
    }

out:
    return rval;
}
