/*
 * hdhomerun_video.c
 *
 * Copyright © 2006-2010 Silicondust USA Inc. <www.silicondust.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "hdhomerun.h"

#ifdef ENTANGLE
#include <sys/mman.h>
#include <inttypes.h>
#include <sys/syscall.h>
#endif

struct hdhomerun_video_sock_t {
	pthread_mutex_t lock;
	struct hdhomerun_debug_t *dbg;
	hdhomerun_sock_t sock;

	volatile size_t head;
	volatile size_t tail;
	uint8_t *buffer;
	size_t buffer_size;
	size_t advance;

	pthread_t thread;
	volatile bool_t terminate;

#ifdef ENTANGLE
	volatile uint64_t packet_count;
	volatile uint64_t transport_error_count;
	volatile uint64_t network_error_count;
	volatile uint64_t sequence_error_count;
	volatile uint64_t overflow_error_count;
	volatile uint64_t pid_packet_count[0x2000];
#else
	volatile uint32_t packet_count;
	volatile uint32_t transport_error_count;
	volatile uint32_t network_error_count;
	volatile uint32_t sequence_error_count;
	volatile uint32_t overflow_error_count;
#endif

	volatile uint32_t rtp_sequence;
	volatile uint8_t sequence[0x2000];

#ifdef ENTANGLE
	uint8_t *stats_buffer;
	int32_t statsPidCC;
	int32_t video_started;
	intptr_t mlock_start;
	intptr_t mlock_end;
#endif
};

static THREAD_FUNC_PREFIX hdhomerun_video_thread_execute(void *arg);

struct hdhomerun_video_sock_t *hdhomerun_video_create(uint16_t listen_port, bool_t allow_port_reuse, size_t buffer_size, struct hdhomerun_debug_t *dbg)
{
	/* Create object. */
	struct hdhomerun_video_sock_t *vs = (struct hdhomerun_video_sock_t *)calloc(1, sizeof(struct hdhomerun_video_sock_t));
	if (!vs) {
		hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to allocate video object\n");
		return NULL;
	}

	vs->dbg = dbg;
	vs->sock = HDHOMERUN_SOCK_INVALID;
	pthread_mutex_init(&vs->lock, NULL);

#ifdef ENTANGLE
	vs->statsPidCC = 0;
#endif

	/* Reset sequence tracking. */
	hdhomerun_video_flush(vs);

	/* Buffer size. */
	vs->buffer_size = (buffer_size / VIDEO_DATA_PACKET_SIZE) * VIDEO_DATA_PACKET_SIZE;
	if (vs->buffer_size == 0) {
		hdhomerun_debug_printf(dbg, "hdhomerun_video_create: invalid buffer size (%lu bytes)\n", (unsigned long)buffer_size);
		goto error;
	}
	vs->buffer_size += VIDEO_DATA_PACKET_SIZE;

	/* Create buffer. */
	vs->buffer = (uint8_t *)malloc(vs->buffer_size);
	if (!vs->buffer) {
		hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to allocate buffer (%lu bytes)\n", (unsigned long)vs->buffer_size);
		goto error;
	}

#ifdef ENTANGLE
        vs->stats_buffer = (uint8_t *)malloc(vs->buffer_size / 188 * 128);
        if (!vs->stats_buffer) {
                hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to allocate stats buffer (%lu bytes)\n", (unsigned long)(buffer_size / VIDEO_DATA_PACKET_SIZE * 188));
                goto error;
        }
#endif

#ifdef ENTANGLE
	{
		int pagesize;
		
		pagesize = getpagesize();
		vs->mlock_start = (intptr_t)vs->buffer & ~(pagesize - 1);
		vs->mlock_end = (intptr_t)vs->buffer + vs->buffer_size;
		vs->mlock_end = (vs->mlock_end + pagesize - 1) & ~(pagesize - 1);

		if (mlock((void *)vs->mlock_start, vs->mlock_end - vs->mlock_start)) {
			hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to lock capture buffer %p (0x%"PRIx64"- 0x%"PRIx64")\n", vs->buffer, vs->mlock_start, vs->mlock_end);
		}

                vs->mlock_start = (intptr_t)vs->stats_buffer & ~(pagesize - 1);
                vs->mlock_end = (intptr_t)vs->stats_buffer + (buffer_size / VIDEO_DATA_PACKET_SIZE * 188);
                vs->mlock_end = (vs->mlock_end + pagesize - 1) & ~(pagesize - 1);

                if (mlock((void *)vs->mlock_start, vs->mlock_end - vs->mlock_start)) {
                        hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to lock stats buffer %p (0x%"PRIx64"- 0x%"PRIx64")\n", vs->stats_buffer, vs->mlock_start, vs->mlock_end);
                }
	}

	vs->video_started = 0;
#endif
	
	/* Create socket. */
	vs->sock = hdhomerun_sock_create_udp();
	if (vs->sock == HDHOMERUN_SOCK_INVALID) {
		hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to allocate socket\n");
		goto error;
	}

	/* Expand socket buffer size. */
	int rx_size = 1024 * 1024;
	setsockopt(vs->sock, SOL_SOCKET, SO_RCVBUF, (char *)&rx_size, sizeof(rx_size));

#ifdef ENTANGLE
        {
            socklen_t len = sizeof(rx_size);
            if (getsockopt(vs->sock, SOL_SOCKET, SO_RCVBUF,
                (void *)&rx_size, &len) < 0)
            {
                fprintf(stderr, "WARN: Failed to get SO_RCVBUF: %s\n", 
                    strerror(errno));
            }

            // Note that the kernel doubles the requested SO_RCVBUF
            if (rx_size != 2 * 1024 * 1024)
            {
                fprintf(stderr, "WARN SO_RCVBUF too small (%d)\n", rx_size);
            }
        }
#endif

	/* Bind socket. */
	if (!hdhomerun_sock_bind(vs->sock, INADDR_ANY, listen_port, allow_port_reuse)) {
		hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to bind socket (port %u)\n", listen_port);
		goto error;
	}

#ifdef ENTANGLE
	{
		struct sched_param sched_param;
		pthread_attr_t thread_attr;

		if (pthread_attr_init(&thread_attr)) {
			hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to create thread attr\n");
		}
		else {
			if (pthread_attr_setinheritsched(&thread_attr, PTHREAD_EXPLICIT_SCHED)) {
				hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to disable inherited scheduling policy\n");
			}

			if (pthread_attr_setschedpolicy(&thread_attr, SCHED_FIFO)) {
				hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to set RT FIFO scheduling policy\n");
			}

			/* Realtime FIFO prioririties range 0 (low) to 32 (high) */
			sched_param.sched_priority = 30;
			if (pthread_attr_setschedparam(&thread_attr, &sched_param)) {
				hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to set RT FIFO scheduling prority\n");
			}

			if (pthread_create(&vs->thread, &thread_attr, &hdhomerun_video_thread_execute, vs) != 0) {
				hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to start thread with realtime attributes, trying non-realtime\n");
				if (pthread_create(&vs->thread, NULL, &hdhomerun_video_thread_execute, vs) != 0) {
					hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to start thread with non-realtime attributes\n");
					if (pthread_attr_destroy(&thread_attr)) {
						hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to destroy thread attr\n");
					}
					goto error;
				}
			}
			if (pthread_attr_destroy(&thread_attr)) {
				hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to destroy thread attr\n");
			}
			pthread_setname_np(vs->thread, "hdhrvideo");
		}
	}
#else
	/* Start thread. */
	if (pthread_create(&vs->thread, NULL, &hdhomerun_video_thread_execute, vs) != 0) {
		hdhomerun_debug_printf(dbg, "hdhomerun_video_create: failed to start thread\n");
		goto error;
	}
#endif

	/* Success. */
	return vs;

error:
	if (vs->sock != HDHOMERUN_SOCK_INVALID) {
		hdhomerun_sock_destroy(vs->sock);
	}
	if (vs->buffer) {
		free(vs->buffer);
	}
#ifdef ENTANGLE
	if (vs->stats_buffer)
	{
		free(vs->stats_buffer);
	}
#endif

	free(vs);
	return NULL;
}

#ifdef ENTANGLE
void hdhomerun_entangle_video_start(struct hdhomerun_video_sock_t *vs)
{
	vs->video_started = 1;
}

void hdhomerun_entangle_video_stop(struct hdhomerun_video_sock_t *vs)
{
	vs->video_started = 0;
}
#endif

void hdhomerun_video_destroy(struct hdhomerun_video_sock_t *vs)
{
	vs->terminate = TRUE;
	pthread_join(vs->thread, NULL);

	hdhomerun_sock_destroy(vs->sock);
	free(vs->buffer);
#ifdef ENTANGLE
	if (vs->stats_buffer)
	{
		free(vs->stats_buffer);
	}
#endif

	free(vs);
}

hdhomerun_sock_t hdhomerun_video_get_sock(struct hdhomerun_video_sock_t *vs)
{
	return vs->sock;
}

uint16_t hdhomerun_video_get_local_port(struct hdhomerun_video_sock_t *vs)
{
	uint16_t port = hdhomerun_sock_getsockname_port(vs->sock);
	if (port == 0) {
		hdhomerun_debug_printf(vs->dbg, "hdhomerun_video_get_local_port: getsockname failed (%d)\n", hdhomerun_sock_getlasterror());
		return 0;
	}

	return port;
}

int hdhomerun_video_join_multicast_group(struct hdhomerun_video_sock_t *vs, uint32_t multicast_ip, uint32_t local_ip)
{
	if (!hdhomerun_sock_join_multicast_group(vs->sock, multicast_ip, local_ip)) {
		hdhomerun_debug_printf(vs->dbg, "hdhomerun_video_join_multicast_group: setsockopt failed (%d)\n", hdhomerun_sock_getlasterror());
		return -1;
	}

	return 1;
}

void hdhomerun_video_leave_multicast_group(struct hdhomerun_video_sock_t *vs, uint32_t multicast_ip, uint32_t local_ip)
{
	if (!hdhomerun_sock_leave_multicast_group(vs->sock, multicast_ip, local_ip)) {
		hdhomerun_debug_printf(vs->dbg, "hdhomerun_video_leave_multicast_group: setsockopt failed (%d)\n", hdhomerun_sock_getlasterror());
	}
}

static void hdhomerun_video_stats_ts_pkt(struct hdhomerun_video_sock_t *vs, uint8_t *ptr)
{
	uint16_t packet_identifier = ((uint16_t)(ptr[1] & 0x1F) << 8) | (uint16_t)ptr[2];

#ifdef ENTANGLE
	vs->pid_packet_count[packet_identifier]++;
#endif
	if (packet_identifier == 0x1FFF) {
		return;
	}

	bool_t transport_error = ptr[1] >> 7;
	if (transport_error) {
		vs->transport_error_count++;
		vs->sequence[packet_identifier] = 0xFF;
		return;
	}

	uint8_t sequence = ptr[3] & 0x0F;

	uint8_t previous_sequence = vs->sequence[packet_identifier];
	vs->sequence[packet_identifier] = sequence;

	if (previous_sequence == 0xFF) {
		return;
	}
	if (sequence == ((previous_sequence + 1) & 0x0F)) {
		return;
	}
	if (sequence == previous_sequence) {
		return;
	}

	vs->sequence_error_count++;
}

static void hdhomerun_video_parse_rtp(struct hdhomerun_video_sock_t *vs, struct hdhomerun_pkt_t *pkt)
{
	pkt->pos += 2;
	uint32_t rtp_sequence = hdhomerun_pkt_read_u16(pkt);
	pkt->pos += 8;

	uint32_t previous_rtp_sequence = vs->rtp_sequence;
	vs->rtp_sequence = rtp_sequence;

	/* Initial case - first packet received. */
	if (previous_rtp_sequence == 0xFFFFFFFF) {
		return;
	}

	/* Normal case - next sequence number. */
	if (rtp_sequence == ((previous_rtp_sequence + 1) & 0xFFFF)) {
		return;
	}

	/* Error case - sequence missed. */
	vs->network_error_count++;

	/* Restart pid sequence check after packet loss. */
	int i;
	for (i = 0; i < 0x2000; i++) {
		vs->sequence[i] = 0xFF;
	}
}

static THREAD_FUNC_PREFIX hdhomerun_video_thread_execute(void *arg)
{
	struct hdhomerun_video_sock_t *vs = (struct hdhomerun_video_sock_t *)arg;
	struct hdhomerun_pkt_t pkt_inst;
#ifdef ENTANGLE
	struct timespec mononow;
	struct timespec last_overflow_message_time;
	struct timespec last_data_received_time;
#endif

#ifdef ENTANGLE
	clock_gettime(CLOCK_MONOTONIC_COARSE, &mononow);
	last_overflow_message_time = mononow;
	last_data_received_time = mononow;
#endif

	while (!vs->terminate) {
		struct hdhomerun_pkt_t *pkt = &pkt_inst;
		hdhomerun_pkt_reset(pkt);

		/* Receive. */
		size_t length = VIDEO_RTP_DATA_PACKET_SIZE;
		if (!hdhomerun_sock_recv(vs->sock, pkt->end, &length, 25)) {
#ifdef ENTANGLE
			if (vs->video_started) {
				int64_t tDelta;
				clock_gettime(CLOCK_MONOTONIC_COARSE, &mononow);

				tDelta = mononow.tv_sec - last_data_received_time.tv_sec;
				tDelta *= 1000000000;
				tDelta += mononow.tv_nsec - last_data_received_time.tv_nsec;
				if (tDelta > 1000000000) {
					hdhomerun_debug_printf(vs->dbg, 
						"No data received on hdhr for more than 1s\n");
					last_data_received_time = mononow;
				}
			}
			else {
				clock_gettime(CLOCK_MONOTONIC_COARSE, &last_data_received_time);
			}
#endif
			continue;
		}

		pkt->end += length;

		if (length == VIDEO_RTP_DATA_PACKET_SIZE) {
			hdhomerun_video_parse_rtp(vs, pkt);
			length = (int)(pkt->end - pkt->pos);
		}

		if (length != VIDEO_DATA_PACKET_SIZE) {
			/* Data received but not valid - ignore. */
			continue;
		}

		pthread_mutex_lock(&vs->lock);

#ifdef ENTANGLE
		clock_gettime(CLOCK_MONOTONIC_COARSE, &mononow);
		last_data_received_time = mononow;
#endif

		/* Store in ring buffer. */
		size_t head = vs->head;
		uint8_t *ptr = vs->buffer + head;
		memcpy(ptr, pkt->pos, length);

		/* Stats. */
		vs->packet_count++;
		hdhomerun_video_stats_ts_pkt(vs, ptr + TS_PACKET_SIZE * 0);
		hdhomerun_video_stats_ts_pkt(vs, ptr + TS_PACKET_SIZE * 1);
		hdhomerun_video_stats_ts_pkt(vs, ptr + TS_PACKET_SIZE * 2);
		hdhomerun_video_stats_ts_pkt(vs, ptr + TS_PACKET_SIZE * 3);
		hdhomerun_video_stats_ts_pkt(vs, ptr + TS_PACKET_SIZE * 4);
		hdhomerun_video_stats_ts_pkt(vs, ptr + TS_PACKET_SIZE * 5);
		hdhomerun_video_stats_ts_pkt(vs, ptr + TS_PACKET_SIZE * 6);

#ifdef ENTANGLE
		uint8_t *pStatsPacket = vs->stats_buffer + (head / 188 * 128);
		struct hdhomerun_stats_packet_payload_t *pStatsPayload;

		pStatsPayload = (struct hdhomerun_stats_packet_payload_t *)(pStatsPacket);
		
		// We genrate one stats packet per network packet. The remaining
		// 6 are marked as invalid.
		pStatsPayload->valid = 1;
		clock_gettime(CLOCK_REALTIME_COARSE, &pStatsPayload->tRealtime);
		pStatsPayload->tMonotonic = mononow;
		pStatsPayload->stats.packet_count = vs->packet_count;
		pStatsPayload->stats.transport_error_count = vs->transport_error_count;
		pStatsPayload->stats.network_error_count = vs->network_error_count;
		pStatsPayload->stats.sequence_error_count = vs->sequence_error_count;
		pStatsPayload->stats.overflow_error_count = vs->overflow_error_count;
		pStatsPayload->stats.null_packet_count = vs->pid_packet_count[0x1fff];
		
		((struct hdhomerun_stats_packet_payload_t *)(pStatsPacket + (128 * 1)))->valid = 0;
		((struct hdhomerun_stats_packet_payload_t *)(pStatsPacket + (128 * 2)))->valid = 0;
		((struct hdhomerun_stats_packet_payload_t *)(pStatsPacket + (128 * 3)))->valid = 0;
		((struct hdhomerun_stats_packet_payload_t *)(pStatsPacket + (128 * 4)))->valid = 0;
		((struct hdhomerun_stats_packet_payload_t *)(pStatsPacket + (128 * 5)))->valid = 0;
		((struct hdhomerun_stats_packet_payload_t *)(pStatsPacket + (128 * 6)))->valid = 0;
#endif

		/* Calculate new head. */
		head += length;

		if (head >= vs->buffer_size) {
			head -= vs->buffer_size;
		}

		/* Check for buffer overflow. */
		if (head == vs->tail) {
			vs->overflow_error_count++;
			pthread_mutex_unlock(&vs->lock);
#ifdef ENTANGLE
			{
				int64_t tDelta;

				tDelta = mononow.tv_sec - last_overflow_message_time.tv_sec;
				tDelta *= 1000000000;
				tDelta += mononow.tv_nsec - last_overflow_message_time.tv_nsec;
				if (tDelta > 1000000000) {
					hdhomerun_debug_printf(vs->dbg, "hdhr overflow\n");
					last_overflow_message_time = mononow;
				}
			}
#endif
			continue;
		}

		vs->head = head;

		pthread_mutex_unlock(&vs->lock);
	}

	return NULL;
}

#ifdef ENTANGLE
uint8_t *hdhomerun_video_get_stats_head(struct hdhomerun_video_sock_t *vs)
{
	return vs->stats_buffer + (vs->tail / 188 * 128);
}

int hdhomerun_video_get_pid_packet_count_snapshot(struct hdhomerun_video_sock_t *vs, uint64_t *psnapshot)
{
	pthread_mutex_lock(&vs->lock);
	memcpy(psnapshot, (void *)vs->pid_packet_count, sizeof(vs->pid_packet_count));
	pthread_mutex_unlock(&vs->lock);

        return 1;
}

#endif

uint8_t *hdhomerun_video_recv(struct hdhomerun_video_sock_t *vs, size_t max_size, size_t *pactual_size)
{
	pthread_mutex_lock(&vs->lock);
    
	size_t head = vs->head;
	size_t tail = vs->tail;

	if (vs->advance > 0) {
		tail += vs->advance;
		if (tail >= vs->buffer_size) {
			tail -= vs->buffer_size;
		}
	
		vs->tail = tail;
	}

	if (head == tail) {
		vs->advance = 0;
		*pactual_size = 0;
		pthread_mutex_unlock(&vs->lock);
		return NULL;
	}

	size_t size = (max_size / VIDEO_DATA_PACKET_SIZE) * VIDEO_DATA_PACKET_SIZE;
	if (size == 0) {
		vs->advance = 0;
		*pactual_size = 0;
		pthread_mutex_unlock(&vs->lock);
		return NULL;
	}

	size_t avail;
	if (head > tail) {
		avail = head - tail;
	} else {
		avail = vs->buffer_size - tail;
	}
	if (size > avail) {
		size = avail;
	}
	vs->advance = size;
	*pactual_size = size;
	uint8_t *result = vs->buffer + tail;

	pthread_mutex_unlock(&vs->lock);
	return result;
}

void hdhomerun_video_flush(struct hdhomerun_video_sock_t *vs)
{
	pthread_mutex_lock(&vs->lock);

	vs->tail = vs->head;
	vs->advance = 0;

	vs->rtp_sequence = 0xFFFFFFFF;

	int i;
	for (i = 0; i < 0x2000; i++) {
		vs->sequence[i] = 0xFF;
	}

	vs->packet_count = 0;
	vs->transport_error_count = 0;
	vs->network_error_count = 0;
	vs->sequence_error_count = 0;
	vs->overflow_error_count = 0;
#ifdef ENTANGLE
	memset((void*)vs->pid_packet_count, 0, sizeof(vs->pid_packet_count));
#endif

	pthread_mutex_unlock(&vs->lock);
}

void hdhomerun_video_debug_print_stats(struct hdhomerun_video_sock_t *vs)
{
	struct hdhomerun_video_stats_t stats;
	hdhomerun_video_get_stats(vs, &stats);

	hdhomerun_debug_printf(vs->dbg, "video sock: pkt=%u net=%u te=%u miss=%u drop=%u\n",
		(unsigned int)stats.packet_count, (unsigned int)stats.network_error_count,
		(unsigned int)stats.transport_error_count, (unsigned int)stats.sequence_error_count,
		(unsigned int)stats.overflow_error_count
	);
}

void hdhomerun_video_get_stats(struct hdhomerun_video_sock_t *vs, struct hdhomerun_video_stats_t *stats)
{
	memset(stats, 0, sizeof(struct hdhomerun_video_stats_t));

	pthread_mutex_lock(&vs->lock);

	stats->packet_count = vs->packet_count;
	stats->network_error_count = vs->network_error_count;
	stats->transport_error_count = vs->transport_error_count;
	stats->sequence_error_count = vs->sequence_error_count;
	stats->overflow_error_count = vs->overflow_error_count;
#ifdef ENTANGLE
	stats->null_packet_count = vs->pid_packet_count[0x1fff];
#endif

	pthread_mutex_unlock(&vs->lock);
}
