/*
 * Copyright © 2018-2022 Koherence, LLC (www.koherence.com).
 *
 * This file is part of the Entangle ATSC 3.0 Base Stack.
 *
 * The Entangle ATSC 3.0 Base Stack is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Affero General Public License
 * version 3 as published by the Free Software Foundation with the addition of 
 * the following permission added to Section 15 as permitted in Section 7(a): 
 * FOR ANY PART OF THE COVERED WORK IN WHICH THE COPYRIGHT IS OWNED BY 
 * KOHERENCE, LLC, KOHERENCE, LLC DISCLAIMS THE WARRANTY OF NON INFRINGEMENT 
 * OF THIRD PARTY RIGHTS.
 *
 * The Entangle ATSC 3.0 Base Stack is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License 
 * along with this program; if not, see http://www.gnu.org/licenses or write 
 * to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA.
 *
 * You can contact Koherence, LLC at email address info@koherence.com
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <assert.h>

#include <libxml2/libxml/xmlreader.h>
#include <zlib.h>

#include <openssl/x509.h>

#include <libenutils/libenutils_status.h>
#include <libenutils/libenutils_defs.h>
#include <libenutils/libenutils.h>
#include <libendebug/libendebug.h>

#include <libenatsc3utils/enxml.h>
#include <libenatsc3utils/enipfilterset.h>

#include <enatsc3-common/enatsc3.h>
#include <enatsc3-common/enatsc3_specdefs.h>
#include <libenatsc3utils/engunzip.h>
#include <libenatsc3baseband/libenatsc3baseband.h>
#include <libenatsc3alp/libenatsc3alp.h>
#include <libenatsc3ipv4/libenatsc3ipv4.h>
#include <libenatsc3ipv4/internetchecksum.h>
#include <libenatsc3lls/libenatsc3lls.h>
#include <libenatsc3lls/atsc3lls_slt.h>
#include <libenatsc3lls/atsc3lls_aeat.h>
#include <libenatsc3lls/atsc3lls_cdt.h>
#include <libenatsc3sls/atsc3sls_held.h>
#include <libenatsc3ssp/libenatsc3ssp.h>
#include <libenatsc3route/atsc3routesls_mime.h>
#include <libenatsc3route/atsc3routesls_stsid.h>
#include <libenatsc3route/libenatsc3route.h>
#include <libenatsc3dashfeeder/libenatsc3dashfeeder.h>
#include <libenatsc3mmtp/libenatsc3mmtp.h>
#include <libenatsc3mmtp/atsc3mmtpsls_usbd.h>
#include <libenatsc3esg/libenatsc3esg.h>

#include "enatsc3source.h"
#include "basebandfilesource.h"
#include "airwavzrzrsdk2source.h"
#include "hdhrj288alpsource.h"

static char *ensys = "bb2fmp4";
static char *ensubsys = "main";

#define MAX_PAYLOAD_SIZE 65536

#define TOI_INVALID -1

#define DASHCONSUMER_URL_MAX_LEN 256

#define REP_ID_MAX_LENGTH 256

#define IsLinearService(serviceCategory)  \
    (serviceCategory == ATSC3LLSSLTServiceCategory_LinearAV || \
     serviceCategory == ATSC3LLSSLTServiceCategory_LinearAudioOnly)
#define IsAppBasedService(serviceCategory) \
     (serviceCategory == ATSC3LLSSLTServiceCategory_AppBased)
#define IsESGService(serviceCategory) \
     (serviceCategory == ATSC3LLSSLTServiceCategory_ESG)
#define IsEASService(serviceCategory) \
     (serviceCategory == ATSC3LLSSLTServiceCategory_EAS)

typedef enum
{
    ENATSC3SourceType_Invalid = 0,
    ENATSC3SourceType_BasebandFile = 1,
    ENATSC3SourceType_BasebandRZRSDK2 = 2,
    ENATSC3SourceType_J288ALPHDHR = 4,
} ENATSC3SourceType;

typedef struct
{
    // The PLP on which the LMT containing the multicast entry was received.
    // This is NOT necessarily the same as the PLP on which the multicast
    // described by the entry is carried.
    uint8_t lmtPLPId;

    // The PLP carrying the multicast
    uint8_t plpId;

    // Multicast description
    uint32_t srcAddr;
    uint32_t destAddr;
    uint16_t srcPort;
    uint16_t destPort;
    uint8_t sidFlag;
    uint8_t compressedFlag;
    uint8_t contextId;
    uint8_t SID;

    // The packet time at which the multicast entry was added
    struct timespec tAdded;
} LMTMulticastEntry;

#define SLTEntryPresentFlag_Protected                   0x0001
#define SLTEntryPresentFlag_Hidden                      0x0002
#define SLTEntryPresentFlag_Essential                   0x0004
#define SLTEntryPresentFlag_BroadbandAccessRequired     0x0008
#define SLTEntryPresentFlag_SLSProtocol                 0x0010
#define SLTEntryPresentFlag_MajorChannelNo              0x0020
#define SLTEntryPresentFlag_MinorChannelNo              0x0040
#define SLTEntryPresentFlag_ServiceId                   0x0080
#define SLTEntryPresentFlag_ServiceCategory             0x0100
#define SLTEntryPresentFlag_ShortServiceName            0x0200
#define SLTEntryPresentFlag_DestinationIPAddress        0x0400
#define SLTEntryPresentFlag_DestinationUdpPort          0x0800

#define SLTEntryFlag_Protected                  1
#define SLTEntryFlag_Hidden                     2
#define SLTEntryFlag_Essential                  4
#define SLTEntryFlag_BroadbandAccessRequired    8

typedef struct
{
    // The group id of the SLT from which the entry originated
    uint8_t groupId;

    // Bitmask of SLTEntryFlag_*
    uint8_t flags;

    uint8_t slsProtocol;

    uint16_t majorChannelNo;
    uint16_t minorChannelNo;

    uint32_t slsDestinationIpAddress;
    uint16_t slsDestinationPort;

    uint16_t serviceId;
    uint8_t serviceCategory;

    char strShortServiceName[11];

    // Bitmask of SLTEntryPresentFlag_* indicating which of the parameters
    // above are valid in the SLT entry.
    uint32_t presentMask;

    // The time at which the SLT entry was added
    struct timespec tAdded;
    // the time at which the SLT entry was last processed (due to an
    // SLT change, which is not the last time it was observed)
    struct timespec tProcessed;
} SLTEntry;

typedef struct
{
    uint64_t llsPLPIdMask;
    uint64_t servicePLPIdMask;

    SLTEntry *pSLTEntries;
    int32_t nSLTEntries;
    LMTMulticastEntry *pLMTMulticastEntries;
    int32_t nLMTMulticastEntries;
} ENATSC3State;

typedef struct
{
    // Whether this entry is valid
    int32_t fValid;

    LMTMulticastEntry entry;
} LMTMulticastEntryInternal;

typedef struct
{
    // Whether this entry is valid
    uint8_t fValid;

    SLTEntry entry;
} SLTEntryInternal;

typedef struct _ENATSC3Session ENATSC3Session;

typedef struct
{
    uint64_t nextObjectId;
    pthread_mutex_t idsLock;
} IdAllocator;

// A Service Session handles a single service. An ATSC3 Session can have
// multiple Service Sessions.
typedef struct _ServiceSession 
{
    int32_t fRunning;
    int32_t verbosity;

    // The main session on which this service is running
    ENATSC3Session *pSession;

    // We are using the serviceId to uniquely identify a service. There
    // should only be one service with id serviceId in the emission.
    //
    // A331 6.3.2: @serviceId: 16-bit integer that shall uniquely identify 
    // this Service within the scope of this Broadcast area.
    //
    // If any the properties of the service, such as channel major/minor
    // or ROUTE/MMTP (sls) protocol changes then it is assumed that it
    // is the same service, but configured differently.
    uint16_t serviceId;

    // The lls_group_id that this service belongs to.
    uint8_t lls_group_id;

    // The table version of the SLT from which the service parameters were 
    // extracted.
    uint8_t sltTableVersion;

    // The service category for this service per A/331
    uint16_t serviceCategory;

    // The major and minor channel number. -1 if not present.
    uint16_t channelMajor;
    uint16_t channelMinor;
    char strGlobalServiceID[128];
    char strShortServiceName[32];
    uint8_t sltSvcSeqNum;
    uint8_t slsProtocol;
    uint8_t slsMajorProtocolVersion;
    uint8_t slsMinorProtocolVersion;
    uint16_t serviceGroupId;

    // The PLPs currently in use for receiving the service
    uint64_t servicePLPIdMask;

    // The program number assigned to this service.
    uint16_t basePid;

    // The IP address/port filter set for this service
    EnIPFilterSet *pIPFilterSet;

    // SLS objects that are injected periodically to the stream. These are
    // protected by the slsLock.

    // The curent STSID (if any) for the service
    int32_t fHaveSTSID;
    uint32_t stsidVersion;
    uint32_t stsidCRC32;

    // The current MPD (if any) for the service
    int32_t fHaveMPD;
    char *pCurrentMPD;
    int32_t mpdLength;
    uint32_t mpdVersion;
    uint32_t mpdCRC32;
    struct timespec tMPDUTC;

    // The current HELD (if any) for the service
    ENATSC3SLSHELDNode currentHELD;
    struct timespec tHELDUTC;
    int32_t fHaveHELD;
    uint32_t heldVersion;
    uint32_t routeHELDCRC32;    // Only used for ROUTE/DASH sessions, not MMTP
    struct _HTMLEntryPackageInfo
    {
        char *pAppContextId;
        char *pRequiredCapabilities;
        int32_t fAppRendering;
        char *pBroadcastEntryPackageUrl;
        char *pBroadcastEntryPageUrl;
        char *pBroadbandEntryPageUrl;
        int64_t clearAppContextCacheDateSecondsUTC;
        int64_t validFromSecondsUTC;
        int64_t validUntilSecondsUTC;
        uint32_t vTSIs[8];
        int32_t nTSIs;
    } vHTMLEntryPackageInfos[16];
    int32_t nHTMLEntryPackageInfos;
    char *pActiveAppContextId;

    pthread_mutex_t slsLock;

    // NRT file reception
    struct _LCTChannelInfo
    {
        int32_t fInUse;

        uint32_t tsi;
        uint32_t addr;
        uint16_t port;
        char strId[REP_ID_MAX_LENGTH];
        ENATSC3ROUTESessionIdType idType;

        int64_t currentTOI;
        uint8_t codepoint;
        char strCurrentURL[256];
        int64_t objectSize;
        int64_t pos;
        int32_t payloadOffset;

        ENUserData userdata;
    } vLCTChannelInfos[32];

    IdAllocator idAllocator;

    struct ServiceSessionStream
    {
        // for MMTP services
        struct 
        {
            int32_t fActive;
            int32_t fHaveStreamSignalingInfo;
            char strAssetId[64];
            int32_t packet_id;
        } mmtp;

        // for DASH services
        struct
        {
            uint32_t tsi;
            char strURL[256];
        } dash;

        // File we are writing to.
        int fd;
    } vStreams[15];

    // ROUTE variables
    ENATSC3ROUTESessionHandle_t hROUTESession;
    ENATSC3ESGSessionHandle_t hESGSession;
    ENATSC3ROUTESessionType routeSessionType;
    ENATSC3DASHFeederHandle_t hDASHFeeder;

    // MMTP variables
    MMTPSessionHandle_t hMMTPSession;
    int32_t fHaveMMTPUSBD;
    int32_t mmtpMPTTableVersion;
    uint32_t mmtpMPTTableCRC32;

    // file counter for this session
    uint32_t mediaFileCounter;

    struct _ServiceSession *pPrev;
    struct _ServiceSession *pNext;
} ServiceSession;

typedef struct _ENATSC3Session
{
    int32_t fRunning;
    char *pOutPath;

    int32_t fXMLErrorHandlerRegistered;

    ENATSC3SourceType sourceType;
    int32_t deviceNumber;
    char *pInFilename;
    int ifd;

    int32_t requestedChannelMajor;
    int32_t requestedChannelMinor;

    int32_t verbosity;
    int32_t fIgnoreUDPChecksumErrors;

    // The ATSC 3 source object
    ENATSC3Source pSource;

    // tuning parameters
    int32_t frequency;

    // The PLPs carrying LLS
    uint64_t llsPLPIdMask;

    // Lock for llsPLPIdMask and servicePLPIdMask
    pthread_mutex_t plpsLock;

    // Link-Level Signaling vars
    struct
    {
        // The last version of the LMT that was successfully processed
        int32_t lmtVersion;
        int32_t lmtLength;
        uint8_t *pLMT;
        int32_t lmtBytesAllocated;

        struct timespec tFirstObserved;
        struct timespec tMostRecentlyObserved;
    } vLMTs[64];
    LMTMulticastEntryInternal *pLMTMulticastEntries;
    int32_t nLMTMulticastEntriesAllocated;

    // Lock protecting the link level signaling state above
    pthread_mutex_t linkLayerSignalingLock;

    // Low-Level Signalng (LLS) vars
    // Array of LLS, one for each LLS group
    struct LLSInfo
    {
        struct
        {
            // The last version of the SLT that was successfully processed 
            int32_t version;
            
            // Used for determining if the gzip payload has changed
            uint32_t crc32;
            uint32_t length;

            // The packet time when the current version of the SLT was seen.
            // The first time an SLT of the current versionw as seen as well
            // as its most recent occurance. Times are recorded for both
            // signed and unsigned versions of the SLT (in the event that
            // both are present in the broadcast).
            struct timespec tSignedFirstObserved;
            struct timespec tUnsignedFirstObserved;
            struct timespec tSignedMostRecentlyObserved;
            struct timespec tUnsignedMostRecentlyObserved;
            struct timespec tSignedMostRecentlyAccepted;
            struct timespec tUnsignedMostRecentlyAccepted;
        } slt;

        struct
        {
            // The last version of the AEAT that was successfully processed 
            int32_t version;
            
            // Used for determining if the gzip payload has changed
            uint32_t crc32;
            uint32_t length;

            // The packet time when the current version of the SLT was seen.
            // The first time an SLT of the current versionw as seen as well
            // as its most recent occurance. Times are recorded for both
            // signed and unsigned versions of the SLT (in the event that
            // both are present in the broadcast).
            struct timespec tSignedFirstObserved;
            struct timespec tUnsignedFirstObserved;
            struct timespec tSignedMostRecentlyObserved;
            struct timespec tUnsignedMostRecentlyObserved;
            struct timespec tSignedMostRecentlyAccepted;
            struct timespec tUnsignedMostRecentlyAccepted;
        } aeat;

        struct
        {
            // The last version of the SMT that was successfully processed
            int32_t version;

            // Used for determining if table has changed
            uint32_t crc32;
            int32_t length;

            struct timespec tFirstObserved;
            struct timespec tMostRecentlyObserved;
            struct timespec tMostRecentlyAccepted;
        } smt;

        struct
        {
            // The last version of the CDT that was successfully processed
            int32_t version;

            // Used for determining if table has changed
            uint32_t crc32;
            int32_t length;

            struct timespec tFirstObserved;
            struct timespec tMostRecentlyObserved;
            struct timespec tMostRecentlyAccepted;
        } cdt;
    } vLLS[256];

    // An array of services in the broadcast stream being transmuxed. Note 
    // that if an emission has mutiple SLT tables, then all services are merged 
    // in pSLTEntries. In the event of multiple SLT tables it  is assumed that 
    // each SLT table has a unique group_id, as indicated by A/331 6.2: "The 
    // value of LLS_group_id shall be unique within this broadcast stream. 
    SLTEntryInternal *pSLTEntries;
    int32_t nSLTEntriesAllocated;

    // Lock protecting the LLS data above
    pthread_mutex_t llsLock;

    // Baseband parsing
    ENATSC3BasebandHandle_t hATSC3Baseband;

    // ALP parsing
    ENATSC3ALPHandle_t hATSC3ALP;

    // IPv4 handling
    ENATSC3IPV4Handle_t hATSC3IPV4;

    // An array of sessions, one per service in the emission 
    ServiceSession *pServiceSessions;
    int32_t nServiceSessions;
    pthread_mutex_t serviceSessionsListLock;

    pthread_cond_t muxStatsCond;

    IdAllocator idAllocator;

    // ATSC 3.0 Security and Service Protection
    ENATSC3SSPHandle_t hSSP;
    // The following are intended for development purposes.
    // fSSPDontVerifySigningCertificate should be set to 1 if the A3SA root
    // certificate is not available (meaning certifiate chain verification
    // cannot be performed).
    int32_t fSSPDontVerifySigningCertificate;
    // fSSPIgnoreOCSPResponseExpired should be set to 1 if replaying a baseband
    // capture so that dates will be ignored. It should be set to 0 for live 
    // sources or if a baseband capture should still have valid unexpired 
    // objects.
    int32_t fSSPIgnoreOCSPResponseExpired;
} ENATSC3Session, *ENATSC3SessionHandle_t;

// Forward declarations
int32_t
ProcessLLSTable(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    ENATSC3LLSTableHeader *pLLSTableHeader, const uint8_t *pPayload, 
    int32_t payloadLength, int32_t fSigned, ENATSC3Session *pSession);

struct
{
    int32_t fInitialized;
    ENATSC3SourceType sourceTypeMask;
    int32_t nBasebandFileDevices;
#if ENABLE_RZR
    int32_t nBasebandRZRSDK2Devices;
#endif
#if ENABLE_HDHR
    int32_t nJ288ALPHDHRDevices;
#endif
} atsc3GlobalsG =
{
    .fInitialized = 0,
    .sourceTypeMask = ENATSC3SourceType_Invalid,
    .nBasebandFileDevices = 0,
#if ENABLE_RZR
    .nBasebandRZRSDK2Devices = 0,
#endif
#if ENABLE_HDHR
    .nJ288ALPHDHRDevices = 0,
#endif
};

void
xmlGenericErrorLogHandler(void *ctx ATTRIBUTE_UNUSED, const char *msg, ...) 
{
    ENATSC3Session *pSession = (ENATSC3Session *)ctx;

    if (pSession->verbosity)
    {
        va_list args;

        va_start(args, msg);
        vfprintf(stderr, msg, args);
        va_end(args);
    }
}

static inline uint64_t 
AllocateObjectId(IdAllocator *pAllocator)
{
    uint64_t objectId;

    if (pthread_mutex_lock(&pAllocator->idsLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    objectId = ++pAllocator->nextObjectId;
    if (objectId == 0)
    {
        objectId = ++pAllocator->nextObjectId;
    }

    if (pthread_mutex_unlock(&pAllocator->idsLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return objectId;
}

//////////////////////////////////////////////////////////////////////////////
// Service Session List Management.
//
// A Service Session handles a single service. An ATSC3 Session can have
// multiple Service Sessions.
//////////////////////////////////////////////////////////////////////////////

static void 
InsertServiceSession_ServiceSessionListLocked(
    ENATSC3Session *pSession, 
    ServiceSession *pServiceSession)
{
    if (pSession->pServiceSessions)
    {
        pSession->pServiceSessions->pPrev = pServiceSession;
    }
    pServiceSession->pNext = pSession->pServiceSessions;
    pServiceSession->pPrev = NULL;

    pSession->pServiceSessions = pServiceSession;

    pSession->nServiceSessions++;
}

static void 
RemoveServiceSession_ServiceSessionListLocked(
    ENATSC3Session *pSession, 
    ServiceSession *pServiceSession)
{
    if (pServiceSession->pPrev)
    {
        pServiceSession->pPrev->pNext = pServiceSession->pNext;
    }
    else
    {
        pSession->pServiceSessions = pServiceSession->pNext;
    }

    if (pServiceSession->pNext)
    {
        pServiceSession->pNext->pPrev = pServiceSession->pPrev;
    }

    pServiceSession->pNext = NULL;
    pServiceSession->pPrev = NULL;

    pSession->nServiceSessions--;
}

//////////////////////////////////////////////////////////////////////////////
// ROUTE LCT channel management
//////////////////////////////////////////////////////////////////////////////

static ENStatus
FindLCTChannelForTSI(ServiceSession *pServiceSession, 
    uint32_t tsi, struct _LCTChannelInfo **ppLCTChannelInfo)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pServiceSession->vLCTChannelInfos); i++)
    {
        if (pServiceSession->vLCTChannelInfos[i].fInUse &&
            pServiceSession->vLCTChannelInfos[i].tsi == tsi)
        {
            *ppLCTChannelInfo = pServiceSession->vLCTChannelInfos + i;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

static ENStatus
AddLCTChannelForTSI(ServiceSession *pServiceSession, 
    uint32_t tsi, uint32_t addr, uint16_t port, const char *pId, 
    ENATSC3ROUTESessionIdType idType, struct _LCTChannelInfo **ppLCTChannelInfo)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    if (strlen(pId) >= sizeof(pServiceSession->vLCTChannelInfos[0].strId))
    {
        ERRORPRINT("id %s too long\n", pId);
        goto out;
    }

    for (i = 0; i < numof(pServiceSession->vLCTChannelInfos); i++)
    {
        if (!pServiceSession->vLCTChannelInfos[i].fInUse)
        {
            pServiceSession->vLCTChannelInfos[i].fInUse = 1;
            pServiceSession->vLCTChannelInfos[i].tsi = tsi;
            pServiceSession->vLCTChannelInfos[i].addr = addr;
            pServiceSession->vLCTChannelInfos[i].port = port;
            pServiceSession->vLCTChannelInfos[i].currentTOI = TOI_INVALID;
            strcpy(pServiceSession->vLCTChannelInfos[i].strId, pId);
            pServiceSession->vLCTChannelInfos[i].idType = idType;
            *ppLCTChannelInfo = pServiceSession->vLCTChannelInfos + i;
            LOGPRINT_NOCONSOLE("%s: Added LCT channel tsi %u\n", 
                pServiceSession->strShortServiceName, tsi); 
            rval = ENStatus_Success;
            break;
        }
    }

out:
    return rval;
}
static ENStatus
RemoveLCTChannelForTSI(ServiceSession *pServiceSession, 
    uint32_t tsi)
{
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0; i < numof(pServiceSession->vLCTChannelInfos); i++)
    {
        if (pServiceSession->vLCTChannelInfos[i].fInUse &&
            pServiceSession->vLCTChannelInfos[i].tsi == tsi)
        {
            pServiceSession->vLCTChannelInfos[i].fInUse = 0;
            LOGPRINT_NOCONSOLE("%s: Removed LCT channel tsi %u\n", 
                pServiceSession->strShortServiceName, tsi);
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

void
LCTChannelAddedCallback(struct timespec *pPacketTime, uint64_t packetNumber, 
    uint32_t addr, uint16_t port, uint32_t tsi, const char *pId, 
    ENATSC3ROUTESessionIdType idType, void *pCbUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;
    struct _LCTChannelInfo *pLCTChannelInfo;

    if (AddLCTChannelForTSI(pServiceSession, tsi, addr, port, pId, idType, 
        &pLCTChannelInfo))
    {
        ERRORPRINT("Failed to allocate LCT Channel info for tsi %u\n", tsi);
    }
}

void
LCTChannelRemovedCallback(struct timespec *pPacketTime, uint64_t packetNumber, 
    uint32_t addr, uint16_t port, uint32_t tsi, void *pCbUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;

    if (RemoveLCTChannelForTSI(pServiceSession, tsi))
    {
        ERRORPRINT("Failed to free LCT Channel info for tsi %u for %p\n", tsi,
            pServiceSession);

        int32_t i;
        for (i = 0; i < numof(pServiceSession->vLCTChannelInfos); i++)
        {
            ERRORPRINT("  fInUse %d tsi %u\n", 
                pServiceSession->vLCTChannelInfos[i].fInUse,
                pServiceSession->vLCTChannelInfos[i].tsi);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
// HELD management
//////////////////////////////////////////////////////////////////////////////
static inline struct _HTMLEntryPackageInfo *
GetActiveHTMLEntryPackage(ServiceSession *pServiceSession, 
    struct timespec *pPacketTimeUTC) 
{
    struct _HTMLEntryPackageInfo *pInfo;
    int32_t i;

    for (i = 0, pInfo = pServiceSession->vHTMLEntryPackageInfos; 
         i < pServiceSession->nHTMLEntryPackageInfos;
         i++, pInfo++)
    {
        if (pPacketTimeUTC->tv_sec >= pInfo->validFromSecondsUTC && 
            pPacketTimeUTC->tv_sec <= pInfo->validUntilSecondsUTC)
        {
            break;
        }
    }

    if (i == pServiceSession->nHTMLEntryPackageInfos)
    {
        return NULL;
    }

    return pInfo;
}

static void
HandleHELDAppContextIdChanged(ServiceSession *pServiceSession, 
    const char *pAppContextId)
{
    // The app should do something here to recognize the new active app.

    return;
}

static void
CheckActiveAppContextId(ServiceSession *pServiceSession,
    struct timespec *pPacketTimeUTC)
{
    struct _HTMLEntryPackageInfo *pActive;
    char *pAppContextId = NULL;
    int32_t fChanged = 0;

    pActive = GetActiveHTMLEntryPackage(pServiceSession, pPacketTimeUTC);
    if (pActive)
    {
        pAppContextId = pActive->pAppContextId;
    }

    if (pAppContextId)
    {
        if (pServiceSession->pActiveAppContextId)
        {
            if (strcmp(pServiceSession->pActiveAppContextId, pAppContextId))
            {
                fChanged = 1;
            }
        }
        else
        {
            fChanged = 1;
        }
    }
    else
    {
        if (pServiceSession->pActiveAppContextId)
        {
            fChanged = 1;
        }
    }

    if (fChanged)
    {
        if (pServiceSession->pActiveAppContextId)
        {
            free(pServiceSession->pActiveAppContextId);
            pServiceSession->pActiveAppContextId = NULL;
        }

        if (pAppContextId)
        {
            pServiceSession->pActiveAppContextId = strdup(pAppContextId);
            if (!pServiceSession->pActiveAppContextId)
            {
                ERRORPRINT("Out of memory\n");
            }
        }
        else
        {
            pServiceSession->pActiveAppContextId = NULL;
        }

        HandleHELDAppContextIdChanged(pServiceSession, 
            pServiceSession->pActiveAppContextId);
    }
}

ENStatus
ROUTEHELDAvailableCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, const char *pHELD, int32_t sHELD, int32_t version, 
    uint32_t crc32, void *pUserData)
{
    ServiceSession *pServiceSession = (ServiceSession *)pUserData;
    ENATSC3SLSHELDNode previousHELD;
    int32_t fHavePreviousHELD = 0;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pServiceSession->slsLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (pServiceSession->fHaveHELD)
    {
        if (pServiceSession->heldVersion == version)
        {
            if (pServiceSession->routeHELDCRC32 == crc32)
            {
                goto out;
            }
            else
            {
                ERRORPRINT("HELD crc32 changed unexpectedly\n");
            }
        }

        LOGPRINT("NEW HELD Version %d -> %d\n", pServiceSession->heldVersion,
            version);
        previousHELD = pServiceSession->currentHELD;
        fHavePreviousHELD = 1;
    }

    if (ENATSC3SLSHELDParse(pHELD, sHELD, &pServiceSession->currentHELD))
    {
        ERRORPRINT("Failed to parse HELD\n");
        if (fHavePreviousHELD)
        {
            pServiceSession->currentHELD = previousHELD;
            fHavePreviousHELD = 0;
        }
        rval = ENStatus_Failure;
        goto out;
    }

    ENATSC3SLSHELDHTMLEntryPackageNode *pEntryPackageNode;
    struct _HTMLEntryPackageInfo *pInfo;
    int32_t i;

    for (i = 0, pInfo = pServiceSession->vHTMLEntryPackageInfos; 
         i < pServiceSession->currentHELD.htmlEntryPackages.nOffsetsUsed; 
         i++, pInfo++)
    {
        if (i == numof(pServiceSession->vHTMLEntryPackageInfos))
        {
            ERRORPRINT("Too many HTMLEntryPakages (%d > %d)\n",
                pServiceSession->currentHELD.htmlEntryPackages.nOffsetsUsed,
                numof(pServiceSession->vHTMLEntryPackageInfos));
            break;
        }

        pEntryPackageNode = EnXMLParserDataOffsetArrayGetPointer(
            &pServiceSession->currentHELD.htmlEntryPackages, i);

        if (pEntryPackageNode->base.presentFlags & 
            ENATSC3SLSHELDHTMLEntryPackage_appContextId)
        {
            pInfo->pAppContextId = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer, 
                pEntryPackageNode->appContextIdDataOffset);
        }
        else
        {
            pInfo->pAppContextId = NULL;
        }

        if (pEntryPackageNode->base.presentFlags & 
            ENATSC3SLSHELDHTMLEntryPackage_requiredCapabilities)
        {
            pInfo->pRequiredCapabilities = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer, 
                pEntryPackageNode->requiredCapabilitiesDataOffset);
        }
        else
        {
            pInfo->pRequiredCapabilities = NULL;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_appRendering)
        {
            pInfo->fAppRendering = pEntryPackageNode->appRendering;
        }
        else
        {
            pInfo->fAppRendering = 0;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_clearAppContextCacheDate)
        {
            pInfo->clearAppContextCacheDateSecondsUTC = 
                pEntryPackageNode->clearAppContextCacheDateSecondsUTC;
        }
        else
        {
            pInfo->clearAppContextCacheDateSecondsUTC = 0x7fffffffffffffff;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_bcastEntryPackageUrl)
        {
            pInfo->pBroadcastEntryPackageUrl = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer,
                pEntryPackageNode->bcastEntryPackageUrlDataOffset);
        }
        else
        {
            pInfo->pBroadcastEntryPackageUrl = NULL;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_bcastEntryPageUrl)
        {
            pInfo->pBroadcastEntryPageUrl = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer,
                pEntryPackageNode->bcastEntryPageUrlDataOffset);
        }
        else
        {
            pInfo->pBroadcastEntryPageUrl = NULL;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_bbandEntryPageUrl)
        {
            pInfo->pBroadbandEntryPageUrl = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer,
                pEntryPackageNode->bbandEntryPageUrlDataOffset);
        }
        else
        {
            pInfo->pBroadbandEntryPageUrl = NULL;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_validFrom)
        {
            pInfo->validFromSecondsUTC = pEntryPackageNode->validFromSecondsUTC;
        }
        else
        {
            pInfo->validFromSecondsUTC = 0;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_validUntil)
        {
            pInfo->validUntilSecondsUTC = 
                pEntryPackageNode->validUntilSecondsUTC;
        }
        else
        {
            pInfo->validUntilSecondsUTC = 0x7fffffffffffffff;
        }

        pInfo->nTSIs = 0;
        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_lctTSIRef)
        {
            char *pTSIs;
            
            pTSIs = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer,
                pEntryPackageNode->lctTSIRefDataOffset);
           
            while (pInfo->nTSIs < numof(pInfo->vTSIs))
            {
                char *pEndPtr;

                pInfo->vTSIs[pInfo->nTSIs] = strtol(pTSIs, &pEndPtr, 10);
                if (pEndPtr == pTSIs)
                {
                    // No number found
                    break;
                }

                pInfo->nTSIs++;
                pTSIs = pEndPtr;
            }
        }
    }

    pServiceSession->nHTMLEntryPackageInfos = i;
    pServiceSession->fHaveHELD = 1;
    pServiceSession->heldVersion = version;
    pServiceSession->routeHELDCRC32 = crc32;
    pServiceSession->tHELDUTC = *pPacketTimeUTC;

out:
    if (pthread_mutex_unlock(&pServiceSession->slsLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    if (fHavePreviousHELD)
    {
        if (ENATSC3SLSHELDFree(&previousHELD))
        {
            ERRORPRINT("Failed to free HELD\n");
        }
    }

    return rval;
}

ENStatus
ROUTEMPDAvailableCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, const char *pMPD, int32_t sMPD, int32_t version, 
    uint32_t crc32, void *pUserData)
{
    ServiceSession *pServiceSession = (ServiceSession *)pUserData;
    ENStatus rval = ENStatus_Success;

    // The app can do something with the MPD
    (void)pServiceSession;

    return rval;
}

void
HandleContentRating(ServiceSession *pServiceSession, 
    const char *pScheme, const char *pValue)
{
    // The app can do something here with the content rating.

    return;
}

ENStatus
ROUTESTSIDAvailableCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, const char *pSTSID, int32_t sSTSID, 
    ENATSC3ROUTESLSSTSIDNodePtr pParsedSTSID, int32_t version, uint32_t crc32, 
    void *pUserData)
{
    ServiceSession *pServiceSession =
        (ServiceSession *)pUserData;
    ENATSC3ROUTESLSSTSIDNode *pSTSIDNode = pParsedSTSID;
    int32_t i;
    int32_t j;
    ENStatus rval = ENStatus_Success;

    if (pthread_mutex_lock(&pServiceSession->slsLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (pServiceSession->fHaveSTSID)
    {
        if (pServiceSession->stsidVersion == version)
        {
            if (pServiceSession->stsidCRC32 == crc32)
            {
                goto out;
            }
            else
            {
                ERRORPRINT("STSID crc32 changed unexpectedly\n");
            }
        }

        LOGPRINT("NEW STSID Version %d -> %d\n", pServiceSession->stsidVersion,
            version);
    }

    for (i = 0; i < pSTSIDNode->RSs.nOffsetsUsed; i++)
    {
        ENATSC3ROUTESLSSTSIDRSNode *pRS;

        pRS = (ENATSC3ROUTESLSSTSIDRSNode *)EnXMLParserDataOffsetArrayGetPointer(
            &pSTSIDNode->RSs, i);
        for (j = 0; j < pRS->LSs.nOffsetsUsed; j++)
        {
            ENATSC3ROUTESLSSTSIDLSNode *pLS;

            pLS = (ENATSC3ROUTESLSSTSIDLSNode *)EnXMLParserDataOffsetArrayGetPointer(
                &pRS->LSs, j);
            if (pLS->base.presentFlags & ENATSC3ROUTESLSSTSIDLSChildNode_SrcFlow)
            {
                ENATSC3ROUTESLSSTSIDSrcFlowNode *pSrcFlowNode;

                pSrcFlowNode = &pLS->srcFlow;
                if (pSrcFlowNode->base.presentFlags & 
                    ENATSC3ROUTESLSSTSIDSrcFlowChildNode_ContentInfo)
                {
                    ENATSC3ROUTESLSSTSIDContentInfoNode *pContentInfoNode;

                    pContentInfoNode = &pSrcFlowNode->contentInfo;
                    if (pContentInfoNode->base.presentFlags & 
                        ENATSC3ROUTESLSSTSIDContentInfoChildNode_MediaInfo)
                    {
                        ENATSC3ROUTESLSSTSIDMediaInfoNode *pMediaInfoNode;

                        pMediaInfoNode = &pContentInfoNode->mediaInfo;
                        if (pMediaInfoNode->base.presentFlags & 
                            ENATSC3ROUTESLSSTSIDMediaInfoChildNode_ContentRating)
                        {
                            ENATSC3ROUTESLSSTSIDContentRatingNode *pContentRatingNode;
                            char *pSchemeIdUri = NULL;
                            char *pValue = NULL;

                            pContentRatingNode = &pMediaInfoNode->contentRating;

                            if (pContentRatingNode->base.presentFlags &
                                ENATSC3ROUTESLSSTSIDContentRatingNodeAttribute_schemeIdUri)
                            {
                                pSchemeIdUri = EnXMLParserDataBufferGetPointer(
                                    &pSTSIDNode->dataBuffer, 
                                    pContentRatingNode->schemeIdUriDataOffset);
                            }

                            if (pContentRatingNode->base.presentFlags &
                                ENATSC3ROUTESLSSTSIDContentRatingNodeAttribute_value)
                            {   
                                pValue = EnXMLParserDataBufferGetPointer(
                                    &pSTSIDNode->dataBuffer, 
                                    pContentRatingNode->valueDataOffset);
                            }
                            HandleContentRating(pServiceSession, pSchemeIdUri, 
                                pValue);
                        }
                    }
                }
            }
        }
    }

    pServiceSession->fHaveSTSID = 1;
    pServiceSession->stsidVersion = version;
    pServiceSession->stsidCRC32 = crc32;

out:
    if (pthread_mutex_unlock(&pServiceSession->slsLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ROUTEFileBeginCallback(struct timespec *pPacketTimeUTC, const char *pURL, 
    uint32_t toi, uint32_t tsi, char *pContentType, ENUserData *pUserData, 
    ServiceSession *pServiceSession)
{
    struct _HTMLEntryPackageInfo *pInfo;
    char *pAppContextId = "";
    int32_t i;
    int32_t j;
    ENStatus rval = ENStatus_Success;

    if (!pServiceSession->fHaveHELD)
    {
        rval = ENStatus_Failure;
        goto out;
    }

    // Check if the active HELD context has changed.
    CheckActiveAppContextId(pServiceSession, pPacketTimeUTC);

    // We are looking for HELD-related packages. For our purproses the package
    // can be either the URL explicitly given as the Broadcast Entry Package
    // URL, or any file which is present on the TSIs indicated in the HELD.
    for (i = 0, pInfo = pServiceSession->vHTMLEntryPackageInfos; 
         i < pServiceSession->nHTMLEntryPackageInfos; 
         i++, pInfo++)
    {
        if (pInfo->pBroadcastEntryPackageUrl &&
            !strcmp(pURL, pInfo->pBroadcastEntryPackageUrl))
        {
            pAppContextId = pInfo->pAppContextId;
            break;
        }

        for (j = 0; j < pInfo->nTSIs; j++)
        {
            if (tsi == pInfo->vTSIs[j])
            {
                pAppContextId = pInfo->pAppContextId;
                break;
            }
        }

        if (j != pInfo->nTSIs)
        {
            break;
        }
    }

    if (i == pServiceSession->nHTMLEntryPackageInfos)
    {
        rval = ENStatus_Failure;
        goto out;
    }

    // We're starting receipt of a HELD-related object. If the app isn't
    // interested then return ENStatus_Failure. Otherwise object receipt
    // contines with the Data and End callbacks (and the Abort callback may
    // be called if object receipt is aborted at the ROUTE level.)

    // Allocate an object ID and store it in the user data so that the 
    // object can be identified in the Data, End and Abort callbacks. If the
    // app requires more context then it can store a pointer to a structure
    // or other information in pUserData.
    pUserData->uint64Value = AllocateObjectId(&pServiceSession->idAllocator);
    (void)pAppContextId;

out:
    return rval;
}

void
ROUTEFileEndCallback(struct timespec *pPacketTimeUTC, 
    int64_t objectSize, ENUserData *pUserData, 
    ServiceSession *pServiceSession)
{
    // An object that we expressed interest in has completed. The object can
    // be identified by information stored in pUserData when 
    // the ROUTEFileBeginCallback was called for this object.
}

void
ROUTEFileAbortCallback(struct timespec *pPacketTimeUTC, 
    ENUserData *pUserData, ServiceSession *pServiceSession)
{
    // An object that we expressed interest in was aborted at the ROUTE level
    // The object can be identified by information stored in pUserData when 
    // the ROUTEFileBeginCallback was called for this object.
}

void
ROUTEFileDataCallback(struct timespec *pPacketTimeUTC, 
    const uint8_t *pPayload, int32_t sPayload, char *pURL, uint32_t toi, 
    int64_t objectPosition, int64_t objectSize, ENUserData *pUserData, 
    ServiceSession *pServiceSession)
{
    // Data for an object that we expressed interest in has arrived.
    // The object can be identified by information stored in pUserData when 
    // the ROUTEFileBeginCallback was called for this object.
}

// ROUTEObjectDataCallback is called by libenats3route when data has 
// arrived on an LCT channel.
void
ROUTEObjectDataCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    uint32_t tsi, uint32_t toi, uint8_t codepoint, int32_t formatId,
    int32_t frag, int32_t order, const char *pId,
    ENATSC3ROUTESessionIdType idType, const uint8_t *pPayload, 
    int32_t payloadLength, int64_t objectPosition, int64_t tolTransferLength,
    int64_t ftiTransferLength, int32_t fError, void *pCbUserData)
{
    ServiceSession *pServiceSession = (ServiceSession *)pCbUserData;
    struct _LCTChannelInfo *pLCTChannelInfo;
    int32_t payloadOffset = 0;

    if (FindLCTChannelForTSI(pServiceSession, tsi, &pLCTChannelInfo))
    {
        ERRORPRINT("Failed to find LCT channel info for tsi %u\n", tsi);
        goto out;
    }

    if (pLCTChannelInfo->currentTOI == TOI_INVALID)
    {
        // We only start processing a new object at offset 0.
        if (objectPosition != 0)
        {
            goto out;
        }
    }

    if (toi != pLCTChannelInfo->currentTOI &&
        pLCTChannelInfo->currentTOI != TOI_INVALID)
    {
        // If we get here then the object Id has changed prematurely. This
        // would suggest that we didn't receive the end of the previous
        // object.
        ERRORPRINT(
            "%"PRIu64" object changed (%"PRId64":%s) %"PRId64" != %"PRId64"\n",
            packetNumber, pLCTChannelInfo->currentTOI,
            pLCTChannelInfo->strCurrentURL,
            pLCTChannelInfo->objectSize,
            pLCTChannelInfo->pos);

        ROUTEFileAbortCallback(pPacketTimeUTC, 
            &pLCTChannelInfo->userdata, pServiceSession);
        pLCTChannelInfo->currentTOI = TOI_INVALID;

        if (objectPosition != 0)
        {
            // We need to wait til we're at the start of a new object.
            goto out;
        }
    }

    // If we've gotten this far then we are processing an object.
    if (pLCTChannelInfo->currentTOI == TOI_INVALID)
    {
        char strContentType[64];

        if (formatId == ENATSC3LCTChannelObjectFormatId_Entity)
        {
            if (ENATSC3ROUTESessionGetEntityModeObjectURLAndContentType(
                pServiceSession->hROUTESession, tsi, pPayload, payloadLength,
                &payloadOffset, pLCTChannelInfo->strCurrentURL,
                sizeof(pLCTChannelInfo->strCurrentURL), strContentType,
                sizeof(strContentType)))
            {
                fprintf(stderr, "Failed to get url for tsi %u toi %u\n",
                    tsi, toi);
                goto out;
            }

            pPayload += payloadOffset;
            payloadLength -= payloadOffset;
        }
        else if (formatId == ENATSC3LCTChannelObjectFormatId_File)
        {
            if (ENATSC3ROUTESessionGetObjectURLAndContentType(
                pServiceSession->hROUTESession, tsi, toi, 
                pLCTChannelInfo->strCurrentURL, 
                sizeof(pLCTChannelInfo->strCurrentURL), strContentType, 
                sizeof(strContentType)))
            {
                ERRORPRINT("Failed to get url for tsi %u toi %u\n",
                    tsi, toi);
                goto out;
            }
        }
        else
        {
            fprintf(stderr, "Unhandled ROUTE formatId %d\n", formatId);
        }

        if (ROUTEFileBeginCallback(pPacketTimeUTC, 
            pLCTChannelInfo->strCurrentURL, toi, tsi, strContentType, 
            &pLCTChannelInfo->userdata, pServiceSession))
        {
            goto out;
        }

        pLCTChannelInfo->currentTOI = toi;
        pLCTChannelInfo->objectSize = -1;
        pLCTChannelInfo->pos = 0;
        pLCTChannelInfo->payloadOffset = payloadOffset;
    }

    if (pLCTChannelInfo->objectSize == -1)
    {
        if (tolTransferLength != -1)
        {
            pLCTChannelInfo->objectSize = tolTransferLength;
        }
        else if (ftiTransferLength != -1)
        {
            pLCTChannelInfo->objectSize = ftiTransferLength;
        }
    }
    else
    {
        if (tolTransferLength != -1)
        {
            if (pLCTChannelInfo->objectSize != tolTransferLength)
            {
                ERRORPRINT("%"PRIu64" TOL object size changed %"PRId64" != %"PRId64"\n",
                    packetNumber, pLCTChannelInfo->objectSize, tolTransferLength);
                ROUTEFileAbortCallback(pPacketTimeUTC, 
                    &pLCTChannelInfo->userdata, pServiceSession);
                pLCTChannelInfo->currentTOI = -1;
                goto out;
            }
        }

        if (ftiTransferLength != -1)
        {
            if (pLCTChannelInfo->objectSize != ftiTransferLength)
            {
                ERRORPRINT("%"PRIu64" FTI object size changed %"PRId64" != %"PRId64"\n",
                    packetNumber, pLCTChannelInfo->objectSize,
                    ftiTransferLength);
                ROUTEFileAbortCallback(pPacketTimeUTC, 
                    &pLCTChannelInfo->userdata, pServiceSession);
                pLCTChannelInfo->currentTOI = TOI_INVALID;
                goto out;
            }
        }
    }

    // We currently expect data to be delivered in order.
    //
    // XXXX Strictly speaking data may arrive out of order if @order=0 in the
    // SrcFlow's Payload element.
    if (pLCTChannelInfo->pos != objectPosition)
    {
        ERRORPRINT("%"PRIu64" %"PRIu64".%.06lu av offset error %"PRId64" != %"PRId64" (%"PRId64" missing) (size %"PRId64")\n",
            packetNumber,
            pPacketTimeUTC->tv_sec, pPacketTimeUTC->tv_nsec / 1000000,
            pLCTChannelInfo->pos, objectPosition,
            objectPosition - pLCTChannelInfo->pos,
            pLCTChannelInfo->objectSize);
        ERRORPRINT("Aborting object\n");

        ROUTEFileAbortCallback(pPacketTimeUTC, 
            &pLCTChannelInfo->userdata, pServiceSession);
        pLCTChannelInfo->currentTOI = TOI_INVALID;
        goto out;
    }

    ROUTEFileDataCallback(pPacketTimeUTC, pPayload, payloadLength, 
        pLCTChannelInfo->strCurrentURL, pLCTChannelInfo->currentTOI, 
        pLCTChannelInfo->pos ? 
            pLCTChannelInfo->pos - pLCTChannelInfo->payloadOffset : 
            0, 
        pLCTChannelInfo->objectSize == -1 ? 
            pLCTChannelInfo->objectSize :
            pLCTChannelInfo->objectSize - pLCTChannelInfo->payloadOffset, 
        &pLCTChannelInfo->userdata, pServiceSession);

    pLCTChannelInfo->pos += payloadLength + payloadOffset;

    // See if the object is now complete
    if (pLCTChannelInfo->pos == pLCTChannelInfo->objectSize)
    {
        LOGPRINT_NOFILE("MediaFileEnd %s toi %d\n", 
            pLCTChannelInfo->strCurrentURL, pLCTChannelInfo->currentTOI);
        ROUTEFileEndCallback(pPacketTimeUTC,
            pLCTChannelInfo->objectSize, &pLCTChannelInfo->userdata, 
            pServiceSession);
        pLCTChannelInfo->currentTOI = TOI_INVALID;
    }

out:
    return;
}

void
ROUTESessionResetCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, void *pCbUserData)
{
    ServiceSession *pServiceSession =
        (ServiceSession *)pCbUserData;
    int32_t i;

    for (i = 0; i < numof(pServiceSession->vLCTChannelInfos); i++)
    {
        if (pServiceSession->vLCTChannelInfos[i].fInUse)
        {
            struct _LCTChannelInfo *pLCTChannelInfo;

            pLCTChannelInfo = pServiceSession->vLCTChannelInfos + i;

            LOGPRINT("ROUTE Session reset tsi %u for %p\n",
                pServiceSession->vLCTChannelInfos[i].tsi, pServiceSession);

            if (pLCTChannelInfo->currentTOI != TOI_INVALID)
            {
                ROUTEFileAbortCallback(pPacketTimeUTC, 
                    &pLCTChannelInfo->userdata, pServiceSession);
                pLCTChannelInfo->currentTOI = TOI_INVALID;
            }
        }
    }

    // Force a reacquisition of all service layer signaling
    pServiceSession->fHaveSTSID = 0;
    if (pServiceSession->pCurrentMPD)
    {
        free(pServiceSession->pCurrentMPD);
        pServiceSession->pCurrentMPD = NULL;
        pServiceSession->fHaveMPD = 0;
    }
    if (pServiceSession->fHaveHELD)
    {
        if (ENATSC3SLSHELDFree(&pServiceSession->currentHELD))
        {
            ERRORPRINT("Failed to free parsed HELD\n");
        }
        pServiceSession->fHaveHELD = 0;
    }
}

void
ESGServiceFragmentCallback(struct timespec *pPacketTimeUTC, int32_t esgSet,
    uint32_t idXMLOffset, int32_t idLength, uint32_t version, 
    uint32_t globalServiceIDXMLOffset, int32_t globalServiceIDLen, 
    const char *pXML, int32_t xmlLength, void *pCbUserData)
{
    ServiceSession *pServiceSession;

    pServiceSession = (ServiceSession *)pCbUserData;

    // The app can do something with the service fragment.
    (void)pServiceSession;

    return;
}

void
ESGScheduleFragmentCallback(struct timespec *pPacketTimeUTC, int32_t esgSet,
    uint32_t idXMLOffset, int32_t idLength, uint32_t version, 
    uint32_t serviceIdReferenceXMLOffset, int32_t serviceIdReferenceLen,
    const char *pXML, int32_t xmlLength, void *pCbUserData)
{
    ServiceSession *pServiceSession;

    pServiceSession = (ServiceSession *)pCbUserData;

    // The app can do something with the schedule fragment
    (void)pServiceSession;

    return;
}

void
ESGContentFragmentCallback(struct timespec *pPacketTimeUTC, int32_t esgSet,
    uint32_t idXMLOffset, int32_t idLength, uint32_t version, 
    const char *pXML, int32_t xmlLength, void *pCbUserData)
{
    ServiceSession *pServiceSession;

    pServiceSession = (ServiceSession *)pCbUserData;

    // The app can do something with the content fragment
    (void)pServiceSession;

    return;
}

ENStatus
ENATSC3SessionServiceSessionCreate(
    ServiceSession **ppServiceSession)
{
    int32_t fHaveSLSLock = 0;
    int32_t fHaveIDsLock = 0;
    ServiceSession *pServiceSession;
    ENStatus rval = ENStatus_Success;

    *ppServiceSession = NULL;

    pServiceSession = (ServiceSession *)calloc(1, sizeof(ServiceSession));
    if (!pServiceSession)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (pthread_mutex_init(&pServiceSession->slsLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveSLSLock = 1;

    if (pthread_mutex_init(&pServiceSession->idAllocator.idsLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveIDsLock = 1;

    pServiceSession->servicePLPIdMask = 0;
    pServiceSession->serviceGroupId = 0xffff;

    if (EnIPFilterSetAllocate(64, &pServiceSession->pIPFilterSet))
    {
        ERRORPRINT("Failed to allocate IP filter set\n");
        rval = 1;
        goto out;
    }

    *ppServiceSession = pServiceSession;

out:
    return rval;

errout:
    if (pServiceSession)
    {
        if (fHaveIDsLock)
        {
            if (pthread_mutex_destroy(&pServiceSession->idAllocator.idsLock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        if (fHaveSLSLock)
        {
            if (pthread_mutex_destroy(&pServiceSession->slsLock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        if (pServiceSession->pIPFilterSet)
        {
            if (EnIPFilterSetFree(pServiceSession->pIPFilterSet))
            {
                ERRORPRINT("Failed to free IP filter set\n");
            }
        }

        free(pServiceSession);
    }

    goto out;
}

ENStatus
ENATSC3SessionServiceSessionFree(
    ServiceSession *pServiceSession)
{
    int32_t i;
    ENStatus rval = ENStatus_Success;

    free(pServiceSession->pActiveAppContextId);

    if (pServiceSession->fHaveHELD)
    {
        if (ENATSC3SLSHELDFree(&pServiceSession->currentHELD))
        {
            ERRORPRINT("Failed to free parsed HELD\n");
        }
    }

    if (pServiceSession->pCurrentMPD)
    {
        free(pServiceSession->pCurrentMPD);
        pServiceSession->pCurrentMPD = NULL;
    }

    if (EnIPFilterSetFree(pServiceSession->pIPFilterSet))
    {
        ERRORPRINT("Failed to free IP filter set\n");
    }
    pServiceSession->pIPFilterSet = NULL;

    if (pthread_mutex_destroy(
        &pServiceSession->idAllocator.idsLock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    if (pthread_mutex_destroy(&pServiceSession->slsLock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }


    for (i = 0; i < numof(pServiceSession->vStreams); i++)
    {
        if (pServiceSession->vStreams[i].fd >= 0)
        {
            close(pServiceSession->vStreams[i].fd);
            pServiceSession->vStreams[i].fd = -1;
        }
    }

    free(pServiceSession);

    return rval;
}

ENStatus
ENATSC3SessionServiceSessionQuiesce(
    ServiceSession *pServiceSession)
{
    ENStatus rval = ENStatus_Success;

    pServiceSession->fRunning = 0;

    return rval;
}

ENStatus
ENATSC3SessionServiceSessionStop(ServiceSession *pServiceSession)
{
    int32_t i;
    ENStatus rval = ENStatus_Success;

    pServiceSession->fRunning = 0;

    if (pServiceSession->hDASHFeeder)
    {
        if (ENATSC3DASHFeederFree(pServiceSession->hDASHFeeder))
        {
            ERRORPRINT("Failed to free DASH consumer\n");
        }
        pServiceSession->hDASHFeeder = NULL;
    }

    if (pServiceSession->hESGSession)
    {
        if (ENATSC3ESGSessionUnregisterCallbacks(
            pServiceSession->hESGSession,
            ESGServiceFragmentCallback, ESGScheduleFragmentCallback, 
            ESGContentFragmentCallback, pServiceSession))
        {
            ERRORPRINT("WARN: Failed to unregister ESG callbacks\n");
        }

        if (ENATSC3ESGSessionFree(pServiceSession->hESGSession))
        {
            ERRORPRINT("Failed to free ESG session\n");
        }
        pServiceSession->hESGSession = NULL;
    }

    if (pServiceSession->hROUTESession)
    {
        if (ENATSC3ROUTESessionUnregisterLCTChannelCallbacks(
            pServiceSession->hROUTESession, 
            LCTChannelAddedCallback, LCTChannelRemovedCallback, 
            pServiceSession))
        {
            ERRORPRINT("WARN: Failed to unregister LCT Channel callbacks\n");
        }

        if (pServiceSession->routeSessionType == 
            ENATSC3ROUTESessionType_ATSC3_DASH)
        {
            if (ENATSC3ROUTESessionUnregisterObjectCallbacks(
                pServiceSession->hROUTESession,
                ROUTEObjectDataCallback, ROUTESessionResetCallback, 
                pServiceSession))
            {
                ERRORPRINT("WARN: Failed to unregister ROUTE object callbacks\n");
            }

            if (ENATSC3ROUTESessionUnregisterSTSIDAvailableCallback(
                pServiceSession->hROUTESession, 
                ROUTESTSIDAvailableCallback, pServiceSession))
            {
                ERRORPRINT("WARN: Failed to unregister ROUTE HELD available callback\n");
            }

            if (ENATSC3ROUTESessionUnregisterHELDAvailableCallback(
                pServiceSession->hROUTESession, 
                ROUTEHELDAvailableCallback, pServiceSession))
            {
                ERRORPRINT("WARN: Failed to unregister ROUTE HELD available callback\n");
            }

            if (ENATSC3ROUTESessionUnregisterMPDAvailableCallback(
                pServiceSession->hROUTESession, 
                ROUTEMPDAvailableCallback, pServiceSession))
            {
                ERRORPRINT("WARN: Failed to unregister ROUTE HELD available callback\n");
            }
        }

        if (ENATSC3ROUTESessionFree(pServiceSession->hROUTESession))
        {
            ERRORPRINT("Failed to free ROUTE session\n");
        }
        pServiceSession->hROUTESession = NULL;
    }

    if (pServiceSession->hMMTPSession)
    {
        if (ENATSC3MMTPSessionFree(pServiceSession->hMMTPSession))
        {
            ERRORPRINT("Failed to free MMTP session\n");
        }
        pServiceSession->hMMTPSession = NULL;
    }

    if (pServiceSession->pIPFilterSet)
    {
        EnIPFilterSetClear(pServiceSession->pIPFilterSet);
    }

    for (i = 0; i < numof(pServiceSession->vStreams); i++)
    {
        if (pServiceSession->vStreams[i].fd >= 0)
        {
            close(pServiceSession->vStreams[i].fd);
            pServiceSession->vStreams[i].fd = -1;
        }
    }

    return rval;
}

// The Link-Layer Signaling lock must be held
static inline void 
RemoveMulticastEntriesForLMTPLP_Locked(ENATSC3Session *pSession, 
    int32_t plpId)
{
    LMTMulticastEntryInternal *pEntry;
    int32_t i;
    
    for (i = 0, pEntry = pSession->pLMTMulticastEntries; 
         i < pSession->nLMTMulticastEntriesAllocated; 
         i++, pEntry++)
    {
        if (pEntry->fValid && pEntry->entry.lmtPLPId == plpId)
        {
            pEntry->fValid = 0;
        }
    }
}

// The Link-Layer Signaling lock must be held
static int32_t
AddMulticastEntry_Locked(ENATSC3Session *pSession, 
    LMTMulticastEntry *pNewEntry)
{
    LMTMulticastEntryInternal *pEntry;
    int32_t i;
    int32_t rval = 0;
    
    for (i = 0, pEntry = pSession->pLMTMulticastEntries; 
         i < pSession->nLMTMulticastEntriesAllocated; 
         i++, pEntry++)
    {
        if (!pEntry->fValid)
        {
            pEntry->entry = *pNewEntry;
            pEntry->fValid = 1;
            break;
        }
    }

    if (i == pSession->nLMTMulticastEntriesAllocated)
    {
        LMTMulticastEntryInternal *pTmp;

        pTmp = (LMTMulticastEntryInternal *)realloc(
            pSession->pLMTMulticastEntries, 
            sizeof(LMTMulticastEntryInternal) * 
            (pSession->nLMTMulticastEntriesAllocated + 16));
        if (!pTmp)
        {
            ERRORPRINT("Out of memory\n");
            rval = 1;
            goto out;
        }
        pSession->pLMTMulticastEntries = pTmp;

        for (i = pSession->nLMTMulticastEntriesAllocated, 
             pEntry = pSession->pLMTMulticastEntries + 
                 pSession->nLMTMulticastEntriesAllocated;
             i < pSession->nLMTMulticastEntriesAllocated + 16;
             i++, pEntry++)
        {
            pEntry->fValid = 0;
        }

        pSession->pLMTMulticastEntries[pSession->nLMTMulticastEntriesAllocated].entry = 
            *pNewEntry;
        pSession->pLMTMulticastEntries[pSession->nLMTMulticastEntriesAllocated].fValid = 1;
        pSession->nLMTMulticastEntriesAllocated += 16;
    }

out:
    return rval;
}

static inline void 
RemoveSLTEntriesForGroupByValidTime_Locked(ENATSC3Session *pSession, 
    uint8_t groupId, struct timespec *pValidTime)
{
    SLTEntryInternal *pEntry;
    int32_t i;
    
    for (i = 0, pEntry = pSession->pSLTEntries; 
         i < pSession->nSLTEntriesAllocated; 
         i++, pEntry++)
    {
        if (pEntry->fValid && pEntry->entry.groupId == groupId &&
            (pEntry->entry.tProcessed.tv_sec != pValidTime->tv_sec ||
             pEntry->entry.tProcessed.tv_nsec != pValidTime->tv_nsec))
        {
            pEntry->fValid = 0;
            LOGPRINT("SLT Entry for %u/%s %d-%d removed\n", 
                pEntry->entry.serviceId,
                pEntry->entry.strShortServiceName, 
                pEntry->entry.majorChannelNo,
                pEntry->entry.minorChannelNo);
        }
    }
}

static inline SLTEntryInternal *
FindSLTEntry_Locked(ENATSC3Session *pSession, uint8_t groupId, 
    uint16_t serviceId)
{
    SLTEntryInternal *pEntry;
    int32_t i;
    
    for (i = 0, pEntry = pSession->pSLTEntries; 
         i < pSession->nSLTEntriesAllocated; 
         i++, pEntry++)
    {
        if (pEntry->fValid && pEntry->entry.groupId == groupId &&
            pEntry->entry.serviceId == serviceId)
        {
            goto out;
        }
    }

    pEntry = NULL;

out:
    return pEntry;
}

// The LLS lock must be held
static int32_t
AddSLTEntry_Locked(ENATSC3Session *pSession, SLTEntry *pNewEntry,
    struct timespec *pMononow)
{
    SLTEntryInternal *pEntry;
    int32_t i;
    int32_t rval = 0;

    for (i = 0, pEntry = pSession->pSLTEntries; 
         i < pSession->nSLTEntriesAllocated; 
         i++, pEntry++)
    {
        if (pEntry->fValid)
        {
            if (pEntry->entry.groupId == pNewEntry->groupId &&
                pEntry->entry.serviceId == pNewEntry->serviceId)
            {
                if (!memcmp(&pEntry->entry, pNewEntry, 
                    offsetof(SLTEntry, tAdded)))
                {
                    LOGPRINT("SLT Entry for %u/%s %d-%d unchanged\n", 
                        pEntry->entry.serviceId,
                        pEntry->entry.strShortServiceName, 
                        pEntry->entry.majorChannelNo,
                        pEntry->entry.minorChannelNo);
                    pEntry->entry.tProcessed = *pMononow;
                }
                else
                {
                    LOGPRINT("SLT Entry for %u/%s %d-%d changed\n", 
                        pEntry->entry.serviceId,
                        pEntry->entry.strShortServiceName, 
                        pEntry->entry.majorChannelNo,
                        pEntry->entry.minorChannelNo);
                    pEntry->entry = *pNewEntry;
                    pEntry->entry.tAdded = *pMononow;
                    pEntry->entry.tProcessed = *pMononow;
                }
                goto out;
            }
        }
    }

    for (i = 0, pEntry = pSession->pSLTEntries; 
         i < pSession->nSLTEntriesAllocated; 
         i++, pEntry++)
    {
        if (!pEntry->fValid)
        {
            pEntry->entry = *pNewEntry;
            pEntry->fValid = 1;
            pEntry->entry.tAdded = *pMononow;
            pEntry->entry.tProcessed = *pMononow;
            LOGPRINT("SLT Entry for %u/%s %d-%d added\n", 
                pEntry->entry.serviceId,
                pEntry->entry.strShortServiceName, 
                pEntry->entry.majorChannelNo,
                pEntry->entry.minorChannelNo);
            goto out;
        }
    }

    if (i == pSession->nSLTEntriesAllocated)
    {
        SLTEntryInternal *pTmp;

        pTmp = (SLTEntryInternal *)realloc(pSession->pSLTEntries, 
            sizeof(SLTEntryInternal) * 
            (pSession->nSLTEntriesAllocated + 16));
        if (!pTmp)
        {
            ERRORPRINT("Out of memory\n");
            rval = 1;
            goto out;
        }
        pSession->pSLTEntries = pTmp;

        for (i = pSession->nSLTEntriesAllocated, pEntry = pSession->pSLTEntries + pSession->nSLTEntriesAllocated;
             i < pSession->nSLTEntriesAllocated + 16;
             i++, pEntry++)
        {
            pEntry->fValid = 0;
        }

        pEntry = pSession->pSLTEntries + pSession->nSLTEntriesAllocated;
        pEntry->entry = *pNewEntry;
        pEntry->fValid = 1;
        pEntry->entry.tAdded = *pMononow;
        pEntry->entry.tProcessed = *pMononow;

        pSession->nSLTEntriesAllocated += 16;

        LOGPRINT("SLT Entry for %u/%s %d-%d added\n", 
            pEntry->entry.serviceId,
            pEntry->entry.strShortServiceName, 
            pEntry->entry.majorChannelNo,
            pEntry->entry.minorChannelNo);
    }

out:
    return rval;
}

ENStatus
ENATSC3Initialize(ENATSC3SourceType sourceTypeMask)
{
    int32_t fBasebandFileSourceInitialized = 0;
#if ENABLE_RZR
    int32_t fBasebandRZRSDK2SourceInitialized = 0;
#endif
#if ENABLE_HDHR
    int32_t fJ288ALPHDHRSourceInitialized = 0;
#endif
    ENStatus rval = ENStatus_Success;

    if (sourceTypeMask & ENATSC3SourceType_BasebandFile)
    {
        if (ENATSC3BasebandFileSourceInitialize(
            &atsc3GlobalsG.nBasebandFileDevices))
        {
            ERRORPRINT("Failed to initialize Baseband file source\n");
            rval = ENStatus_Failure;
            goto errout;
        }
        fBasebandFileSourceInitialized = 1;
    }

#if ENABLE_RZR
    if (sourceTypeMask & ENATSC3SourceType_BasebandRZRSDK2)
    {
        if (ENATSC3BasebandRZRSDK2SourceInitialize(
            &atsc3GlobalsG.nBasebandRZRSDK2Devices))
        {
            ERRORPRINT("Failed to initialize RZR SDK2 source\n");
            rval = ENStatus_Failure;
            goto errout;
        }
        fBasebandRZRSDK2SourceInitialized = 1;
    }
#endif

#if ENABLE_HDHR
    if (sourceTypeMask & ENATSC3SourceType_J288ALPHDHR)
    {
        if (ENATSC3J288ALPHDHRSourceInitialize(
            &atsc3GlobalsG.nJ288ALPHDHRDevices))
        {
            ERRORPRINT("Failed to initialize J288 ALP HDHRsource\n");
            rval = ENStatus_Failure;
            goto errout;
        }
        fJ288ALPHDHRSourceInitialized = 1;
    }
#endif

    atsc3GlobalsG.sourceTypeMask = sourceTypeMask;

    atsc3GlobalsG.fInitialized = 1;
 
out:
    return rval;

errout:
    if (fBasebandFileSourceInitialized)
    {
        if (ENATSC3BasebandFileSourceShutdown())
        {
            ERRORPRINT("Failed to shut down Baseband file source\n");
        }
    }

#if ENABLE_HDHR
    if (fJ288ALPHDHRSourceInitialized)
    {
        if (ENATSC3J288ALPHDHRSourceShutdown())
        {
            ERRORPRINT("Failed to shut down J288 ALP source\n");
        }
    }
#endif
 
#if ENABLE_RZR
    if (fBasebandRZRSDK2SourceInitialized)
    {
        if (ENATSC3BasebandRZRSDK2SourceShutdown())
        {
            ERRORPRINT("Failed to shut down Baseband RZR source\n");
        }
    }
#endif
 
    goto out;
}

ENStatus
ENATSC3Shutdown(void)
{
    ENStatus rval = ENStatus_Success;

    if (atsc3GlobalsG.sourceTypeMask & ENATSC3SourceType_BasebandFile)
    {
        if (ENATSC3BasebandFileSourceShutdown())
        {
            ERRORPRINT("Failed to shut down Baseband File source\n");
        }
    }

#if ENABLE_RZR
    if (atsc3GlobalsG.sourceTypeMask & ENATSC3SourceType_BasebandRZRSDK2)
    {
        if (ENATSC3BasebandRZRSDK2SourceShutdown())
        {
            ERRORPRINT("Failed to shut down RZR source\n");
        }
    }
#endif

#if ENABLE_HDHR
    if (atsc3GlobalsG.sourceTypeMask & ENATSC3SourceType_J288ALPHDHR)
    {
        if (ENATSC3J288ALPHDHRSourceShutdown())
        {
            ERRORPRINT("Failed to shut down HDHR source\n");
        }
    }
#endif

    return rval;
}

ENStatus
SetPLPIds(ENATSC3Session *pSession, uint64_t plpIdMask)
{
    ENStatus rval = ENStatus_Success;

    LOGPRINT_NOCONSOLE("Setting PLP mask %"PRIx64"\n", plpIdMask);

    // Always have at least one plp
    if (plpIdMask == 0)
    {
        plpIdMask = 1;
    }

    if (pSession->pSource &&
        pSession->pSource->SetPLPIds(pSession->pSource, plpIdMask))
    {
        ERRORPRINT("Failed to set ATSC3 source PLPs to %"PRIx64"\n",
            plpIdMask);
        rval = ENStatus_Failure;
    }

    if (pSession->hATSC3Baseband)
    {
        if (ENATSC3BasebandSetPLPs(pSession->hATSC3Baseband, plpIdMask))
        {
            ERRORPRINT("Failed to set baseband processor PLP mask %x\n",
                plpIdMask);
        }
    }

    if (pSession->hATSC3ALP)
    {
        if (ENATSC3ALPSetPLPs(pSession->hATSC3ALP, plpIdMask))
        {
            ERRORPRINT("Failed to set set ALP processor PLPs to %"PRIx64"\n",
                plpIdMask);
            rval = ENStatus_Failure;
        }
    }

    return rval;
}

void
AddIPAddress(uint32_t addr, uint16_t port, void *pCbUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;

    if (EnIPFilterSetAddFilter(addr, port, pServiceSession->pIPFilterSet))
    {
        ERRORPRINT("%u/%s: Failed to add filter for %d.%d.%d.%d:%d\n",
            pServiceSession->serviceId, pServiceSession->strShortServiceName,
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
    }
}

void
RemoveIPAddress(uint32_t addr, uint16_t port, void *pCbUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;

    if (EnIPFilterSetRemoveFilter(addr, port, pServiceSession->pIPFilterSet))
    {
        ERRORPRINT("Failed to remove filter for %d.%d.%d.%d:%d\n",
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
    }
}

// The Link Layer Signaling lock must be held
static inline ENStatus
FindPLPForIPAddress_Locked(ENATSC3Session *pSession, uint32_t addr, 
    uint16_t port, uint8_t *pPLPId)
{
    LMTMulticastEntryInternal *pEntry;
    int32_t i;
    ENStatus rval = ENStatus_Failure;

    for (i = 0, pEntry = pSession->pLMTMulticastEntries; 
         i < pSession->nLMTMulticastEntriesAllocated; 
         i++, pEntry++)
    {
        if (pEntry->fValid)
        {
            LOGPRINT_NOCONSOLE("lmtplp %d %d:%d:%d:%d:%d plp %d\n",
                pEntry->entry.lmtPLPId, 
                (pEntry->entry.destAddr >> 24) & 0xff,
                (pEntry->entry.destAddr >> 16) & 0xff,
                (pEntry->entry.destAddr >> 8) & 0xff,
                (pEntry->entry.destAddr >> 9) & 0xff,
                pEntry->entry.destPort, 
                pEntry->entry.plpId);
        }

        if (pEntry->fValid && 
            pEntry->entry.destAddr == addr && pEntry->entry.destPort == port)
        {
            *pPLPId = pEntry->entry.plpId;
            rval = ENStatus_Success;
            break;
        }
    }

    return rval;
}

ENStatus
GetPLPForIPAddress(uint32_t addr, uint16_t port, uint8_t *pPLPId, 
    void *pCbUserData)
{
    ENStatus rval = ENStatus_Success;

    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;

    if (pthread_mutex_lock(&pServiceSession->pSession->linkLayerSignalingLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (FindPLPForIPAddress_Locked(pServiceSession->pSession, addr, port, 
        pPLPId))
    {
        ERRORPRINT("WARN: Couldn't find plp for %d.%d.%d.%d:%d\n",
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
        rval = ENStatus_Failure;
    }

    if (pthread_mutex_unlock(
        &pServiceSession->pSession->linkLayerSignalingLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    return rval;
}

typedef struct
{
    ServiceSession *pServiceSession;
    ENATSC3Session *pSession;
    uint64_t plpIdMask;
} SetPLPsUserData;

// The Link-Layer Signaling lock must be hed
ENStatus
SetPLPsIter_Locked(uint32_t addr, uint16_t port, void *pCbUserData)
{
    SetPLPsUserData *pUserData = (SetPLPsUserData *)pCbUserData;
    uint8_t plpId;
    ENStatus rval = ENStatus_Success;

    if (FindPLPForIPAddress_Locked(pUserData->pSession, addr, port, &plpId))
    {
        LOGPRINT_NOCONSOLE("%u:%s WARN: Couldn't find plp for %d.%d.%d.%d:%d\n",
            pUserData->pServiceSession->serviceId, 
            pUserData->pServiceSession->strShortServiceName,
            (addr >> 24) & 0xff, (addr >> 16) & 0xff, (addr >> 8) & 0xff,
            addr & 0xff, port);
        rval = ENStatus_Failure;
    }
    else
    {
        LOGPRINT_NOCONSOLE("%u/%s: Found PLP %d for IP %d.%d.%d.%d:%d\n", 
            pUserData->pServiceSession->serviceId, 
            pUserData->pServiceSession->strShortServiceName,
            plpId, (addr >> 24) & 0xff, (addr >> 16) & 0xff, 
            (addr >> 8) & 0xff, addr & 0xff, port);
        pUserData->plpIdMask |= (1ULL << plpId);
    }

    return rval;
}

// The Link-Layer Signaling lock must be held
// If pServiceSession is not null then its PLP set is evaluated and the 
// PLPs are set. Otherwise if pServiceSession is NULL then just set the 
// PLPs.
static inline ENStatus
GetServicePLPsFromIPFilterSet_LinkLayerSignalingLocked(
    ServiceSession *pServiceSession, int32_t *pFChanged)
{
    SetPLPsUserData userdata;
    uint64_t oldServicePLPIdMask;
    ENStatus rval = ENStatus_Success;

    oldServicePLPIdMask = pServiceSession->servicePLPIdMask;

    // Iterate over the IP filter set to determine which PLPs it 
    // corresponds to. 
    userdata.pServiceSession = pServiceSession;
    userdata.pSession = pServiceSession->pSession;
    userdata.plpIdMask = 0;
    if (EnIPFilterSetIter(pServiceSession->pIPFilterSet, SetPLPsIter_Locked, 
        &userdata))
    {
        LOGPRINT_NOCONSOLE(
            "WARN: Failed to map all IP address/ports to PLPs\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (userdata.plpIdMask != oldServicePLPIdMask)
    {
        pServiceSession->servicePLPIdMask = userdata.plpIdMask;
        *pFChanged = 1;
    }
    else
    {
        *pFChanged = 0;
    }

out:
    return rval;
}

// The Link-Layer Signaling lock must be held
// If pServiceSession is not null then its PLP set is evaluated and the 
// PLPs are set. Otherwise if pServiceSession is NULL then just set the 
// PLPs.
ENStatus
SetPLPsFromMasks(ENATSC3Session *pSession)
{
    ENStatus rval = ENStatus_Success;

    ServiceSession *pIter;
    uint64_t mask;

    if (pthread_mutex_lock(&pSession->plpsLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    mask = pSession->llsPLPIdMask;

    if (pthread_mutex_unlock(&pSession->plpsLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    if (pthread_mutex_lock(&pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pIter = pSession->pServiceSessions;
    while (pIter)
    {
        mask |= pIter->servicePLPIdMask;
        pIter = pIter->pNext;
    }

    if (pthread_mutex_unlock(&pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    rval = SetPLPIds(pSession, mask);

    return rval;
}

void
CommitIPAddresses(void *pCbUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;
    int32_t fChanged;

    if (pthread_mutex_lock(&pServiceSession->pSession->linkLayerSignalingLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (GetServicePLPsFromIPFilterSet_LinkLayerSignalingLocked(
        pServiceSession, &fChanged))
    {
        ERRORPRINT("WARN: Failed to get service PLPs for serviceId %d %d-%d\n",
            pServiceSession->serviceId, pServiceSession->channelMajor,
            pServiceSession->channelMinor);
        fChanged = 0;
    }

    if (pthread_mutex_unlock(
        &pServiceSession->pSession->linkLayerSignalingLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    if (fChanged)
    {
        if (SetPLPsFromMasks(pServiceSession->pSession))
        {
            ERRORPRINT("WARN: Failed to update PLPs\n");
        }
    }
}

static ENStatus
DASHFeederMediaFileAccept(struct timespec *pPacketTimeUTC, uint32_t tsi, 
    uint32_t toi, uint8_t codepoint, const char *pId, 
    ENATSC3ROUTESessionIdType idType, const char *pURL, const uint8_t *pPayload, 
    int32_t payloadLength, void *pCbData)
{
    ServiceSession *pServiceSession = (ServiceSession *)pCbData;
    struct ServiceSessionStream *pStream;
    int32_t i;
    int32_t rval = ENStatus_Success;

    for (i = 0, pStream = pServiceSession->vStreams;
         i < numof(pServiceSession->vStreams);
         i++, pStream++)
    {
        if (pStream->dash.tsi == 0)
        {
            if (strlen(pURL) >= sizeof(pStream->dash.strURL))
            {
                fprintf(stderr, "URL %s too long\n", pURL);
                rval = ENStatus_Failure;
                break;
            }
            pStream->dash.tsi = tsi;
            pStream->fd = -1;
            strcpy(pStream->dash.strURL, pURL);
            break;
        }
    }

    if (i == numof(pServiceSession->vStreams))
    {
        fprintf(stderr, "Failed to allocate stream for %s tsi %u\n", pURL,
            tsi);
        rval = ENStatus_Failure;
        goto out;
    }

out:
    return rval;
}

ENStatus
DASHFeederMediaFileBeginCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t tsi, uint32_t toi, uint8_t codepoint, 
    const char *pId, ENATSC3ROUTESessionIdType idType, const char *pURL, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t fError,
    void *pCbData)
{
    ServiceSession *pServiceSession = (ServiceSession *)pCbData;
    struct ServiceSessionStream *pStream = NULL;
    char strFilename[256];
    int32_t i;
    ENStatus rval = ENStatus_Success;

    if (DASHFeederMediaFileAccept(pPacketTimeUTC, tsi, toi, codepoint, pId, 
        idType, pURL, pPayload, payloadLength, pCbData))
    {
        rval = ENStatus_Dropped;
        goto out;
    }

    for (i = 0, pStream = pServiceSession->vStreams; 
         i < numof(pServiceSession->vStreams); 
         i++, pStream++)
    {
        if (!strcmp(pStream->dash.strURL, pURL))
        {
            break;
        }
    }

    if (i < numof(pServiceSession->vStreams))
    {
        if (snprintf(strFilename, sizeof(strFilename), "%s/%s-%06d-%s", 
            pServiceSession->pSession->pOutPath, 
            pServiceSession->strShortServiceName, 
            pServiceSession->mediaFileCounter++, pURL) >= sizeof(strFilename))
        {
            ERRORPRINT("Filename too long\n");
            rval = ENStatus_Failure;
            goto out;
        }

        pStream->fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (pStream->fd < 0)
        {
            fprintf(stderr, "Failed to open %s (%s)\n", strFilename, 
                strerror(errno));
            rval = ENStatus_Failure;
            goto out;
        }
    }

out:
    return rval;
}

void
DASHFeederMediaFileDataCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t tsi, uint32_t toi, const char *pURL,
    const uint8_t *pPayload, int32_t sPayload, const int64_t objectPosition, 
    int64_t objectSize, int32_t fError, void *pCbData)
{
    ServiceSession *pServiceSession = (ServiceSession *)pCbData;
    struct ServiceSessionStream *pStream;
    int32_t i;

    for (i = 0, pStream = pServiceSession->vStreams; 
         i < numof(pServiceSession->vStreams); 
         i++, pStream++)
    {
        if (!strcmp(pStream->dash.strURL, pURL))
        {
            break;
        }
    }

    if (i < numof(pServiceSession->vStreams))
    {
        if (pStream->fd >= 0)
        {
            if (lseek(pStream->fd, objectPosition, SEEK_SET) != objectPosition)
            {
                fprintf(stderr, "Failed to seek %s to %"PRId64"\n",
                    pURL, objectPosition); 
            }

            // pPayload can be null to indicate missing data 
            if (pPayload)
            {
                if (write(pStream->fd, pPayload, sPayload) != sPayload)
                {
                    fprintf(stderr, "Error writing to file: %s\n", 
                        strerror(errno));
                }
            }
            else
            {
                uint8_t vPad[1024] = {0};
                int32_t nToWrite;

                while (sPayload)
                {
                    nToWrite = MIN(sizeof(vPad), sPayload);

                    if (write(pStream->fd, vPad, nToWrite) != nToWrite)
                    {
                        fprintf(stderr, "Error padding file: %s\n", 
                            strerror(errno));
                    }

                    sPayload -= nToWrite;
                }
            }
        }
    }

    return;
}

void
DASHFeederMediaFileEndCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t tsi, uint32_t toi, const char *pURL, 
    int64_t objectSize, int64_t pos, void *pCbData)
{
    ServiceSession *pServiceSession = (ServiceSession *)pCbData;
    struct ServiceSessionStream *pStream;
    int32_t i;

    for (i = 0, pStream = pServiceSession->vStreams; 
         i < numof(pServiceSession->vStreams); 
         i++, pStream++)
    {
        if (!strcmp(pStream->dash.strURL, pURL))
        {
            break;
        }
    }

    if (i < numof(pServiceSession->vStreams))
    {
        if (pStream->fd >= 0)
        {
            if (close(pStream->fd))
            {
                fprintf(stderr, "Failed to close file: %s\n", strerror(errno));
            }
            pStream->fd = -1;
        }
        pStream->dash.tsi = 0;
    }
    else
    {
        fprintf(stderr, "Failed to end file %s tsi %u\n", pURL, tsi);
    }

    return;
}

void
DASHFeederMediaFileAbortCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t tsi, uint32_t toi, const char *pURL, 
    int64_t objectSize, int64_t pos, void *pCbData)
{
    ServiceSession *pServiceSession = (ServiceSession *)pCbData;
    struct ServiceSessionStream *pStream;
    int32_t i;

    for (i = 0, pStream = pServiceSession->vStreams; 
         i < numof(pServiceSession->vStreams); 
         i++, pStream++)
    {
        if (!strcmp(pStream->dash.strURL, pURL))
        {
            break;
        }
    }

    if (i < numof(pServiceSession->vStreams))
    {
        if (pStream->fd >= 0)
        {
            if (close(pStream->fd))
            {
                fprintf(stderr, "Failed to close file: %s\n", strerror(errno));
            }
            pStream->fd = -1;
        }
        pStream->dash.tsi = 0;
    }
    else
    {
        fprintf(stderr, "Failed to abort file %s tsi %u\n", pURL, tsi);
    }

    return;
}

static void
DASHFeederLCTChannelRemovedCallback(struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, uint32_t tsi, void *pCbData)
{
    ServiceSession *pServiceSession = (ServiceSession *)pCbData;
    struct ServiceSessionStream *pStream;
    int32_t i;

    for (i = 0, pStream = pServiceSession->vStreams; 
         i < numof(pServiceSession->vStreams); 
         i++, pStream++)
    {
        if (pStream->dash.tsi == tsi)
        {
            break;
        }
    }

    if (i < numof(pServiceSession->vStreams))
    {
        if (pStream->fd >= 0)
        {
            close(pStream->fd);
        }
        pStream->dash.tsi = 0;
    }
}

static int32_t
SetupROUTEDASHSession(ServiceSession *pServiceSession, 
    uint32_t addr, uint16_t port)
{
    int32_t fROUTEMPDAvailableCallbacksRegistered = 0;
    int32_t fROUTEHELDAvailableCallbacksRegistered = 0;
    int32_t fROUTESTSIDAvailableCallbacksRegistered = 0;
    int32_t fROUTEObjectCallbacksRegistered = 0;
    int32_t fROUTELCTChannelCallbacksRegistered = 0;
    int32_t rval = 0;

    if (ENATSC3ROUTESessionAllocate(ENATSC3ROUTESessionType_ATSC3_DASH, addr, 
        port, AddIPAddress, RemoveIPAddress, CommitIPAddresses, pServiceSession, 
        pServiceSession->strShortServiceName, &pServiceSession->hROUTESession))
    {
        ERRORPRINT("%s Failed to allocate ROUTE session\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }
    pServiceSession->routeSessionType = ENATSC3ROUTESessionType_ATSC3_DASH;

    if (ENATSC3ROUTESessionRegisterMPDAvailableCallback(
        pServiceSession->hROUTESession, ROUTEMPDAvailableCallback, 
        pServiceSession))
    {
        ERRORPRINT("%s: Failed to register ROUTE HELD available callback\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }
    fROUTEMPDAvailableCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterHELDAvailableCallback(
        pServiceSession->hROUTESession, ROUTEHELDAvailableCallback, 
        pServiceSession))
    {
        ERRORPRINT("%s Failed to register ROUTE HELD available callback\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }
    fROUTEHELDAvailableCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterSTSIDAvailableCallback(
        pServiceSession->hROUTESession, ROUTESTSIDAvailableCallback, 
        pServiceSession))
    {
        ERRORPRINT("%s Failed to register ROUTE STSID available callback\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }
    fROUTESTSIDAvailableCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterObjectCallbacks(
        pServiceSession->hROUTESession, ROUTEObjectDataCallback, 
        ROUTESessionResetCallback, pServiceSession))
    {
        ERRORPRINT("%s Failed to register ROUTE object callbacks\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }
    fROUTEObjectCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterLCTChannelCallbacks(
        pServiceSession->hROUTESession, LCTChannelAddedCallback, 
        LCTChannelRemovedCallback, pServiceSession))
    {
        ERRORPRINT("%s Failed to register LCT Channel callbacks\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto out;
    }
    fROUTELCTChannelCallbacksRegistered = 1;

    if (ENATSC3DASHFeederAllocate(pServiceSession->hROUTESession, 
        pServiceSession->strShortServiceName, &pServiceSession->hDASHFeeder))
    {
        ERRORPRINT("%s Failed to allocate DASH consumer\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }

    ENATSC3DASHFeederSetVerbosity(pServiceSession->hDASHFeeder, 
        pServiceSession->verbosity);

    if (ENATSC3DASHFeederRegisterCallbacks(pServiceSession->hDASHFeeder,
        DASHFeederMediaFileBeginCallback, 
        DASHFeederMediaFileEndCallback, 
        DASHFeederMediaFileAbortCallback, 
        DASHFeederMediaFileDataCallback, 
        DASHFeederLCTChannelRemovedCallback,
        pServiceSession))
    {
        ERRORPRINT("%s Failed to register ENATSC3DASHFeeder callbacks\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }

out:
    return rval;

errout:
    if (fROUTELCTChannelCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterLCTChannelCallbacks(
            pServiceSession->hROUTESession, LCTChannelAddedCallback, 
            LCTChannelRemovedCallback, pServiceSession))
        {
            ERRORPRINT("%s WARN: Failed to unregister LCT Channel callbacks\n",
                pServiceSession->strShortServiceName);
        }
    }

    if (fROUTEObjectCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterObjectCallbacks(
            pServiceSession->hROUTESession, ROUTEObjectDataCallback, 
            ROUTESessionResetCallback, pServiceSession))
        {
            ERRORPRINT("%s WARN: Failed to unregister ROUTE object callbacks\n",
                pServiceSession->strShortServiceName);
        }
    }

    if (fROUTESTSIDAvailableCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterSTSIDAvailableCallback(
            pServiceSession->hROUTESession, ROUTESTSIDAvailableCallback, 
            pServiceSession))
        {
            ERRORPRINT("%s WARN: Failed to unregister ROUTE STSID available callback\n", pServiceSession->strShortServiceName);
        }
    }

    if (fROUTEHELDAvailableCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterHELDAvailableCallback(
            pServiceSession->hROUTESession, ROUTEHELDAvailableCallback, 
            pServiceSession))
        {
            ERRORPRINT("%s WARN: Failed to unregister ROUTE HELD available callback\n", pServiceSession->strShortServiceName);
        }
    }

    if (fROUTEMPDAvailableCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterMPDAvailableCallback(
            pServiceSession->hROUTESession, ROUTEMPDAvailableCallback, 
            pServiceSession))
        {
            ERRORPRINT("%s WARN: Failed to unregister ROUTE HELD available callback\n", pServiceSession->strShortServiceName);
        }
    }

    if (pServiceSession->hDASHFeeder)
    {
        if (ENATSC3DASHFeederFree(pServiceSession->hDASHFeeder))
        {
            ERRORPRINT("%s Failed to free DASH consumer\n",
                pServiceSession->strShortServiceName);
        }
        pServiceSession->hDASHFeeder = NULL;
    }

    if (pServiceSession->hROUTESession)
    {
        if (ENATSC3ROUTESessionFree(pServiceSession->hROUTESession))
        {
            ERRORPRINT("%s Failed to free ROUTE session\n",
                pServiceSession->strShortServiceName);
        }
        pServiceSession->hROUTESession = NULL;
    }

    pServiceSession->routeSessionType = ENATSC3ROUTESessionType_Invalid;

    goto out;
}

static int32_t
SetupROUTEESGSession(ServiceSession *pServiceSession, 
    uint32_t addr, uint16_t port)
{
    int32_t fROUTEObjectCallbacksRegistered = 0;
    int32_t fROUTELCTChannelCallbacksRegistered = 0;
    int32_t fESGCallbacksRegistered = 0;
    int32_t rval = 0;

    if (ENATSC3ROUTESessionAllocate(ENATSC3ROUTESessionType_ATSC3_ESG, addr, 
        port, AddIPAddress, RemoveIPAddress, CommitIPAddresses, pServiceSession, 
        pServiceSession->strShortServiceName, &pServiceSession->hROUTESession))
    {
        ERRORPRINT("%s Failed to allocate ROUTE session\n", 
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }
    
    pServiceSession->routeSessionType = ENATSC3ROUTESessionType_ATSC3_ESG;

    if (ENATSC3ROUTESessionRegisterObjectCallbacks(
        pServiceSession->hROUTESession, ROUTEObjectDataCallback, 
        ROUTESessionResetCallback, pServiceSession))
    {
        ERRORPRINT("%s Failed to register ROUTE object callbacks\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }
    fROUTEObjectCallbacksRegistered = 1;

    if (ENATSC3ROUTESessionRegisterLCTChannelCallbacks(
        pServiceSession->hROUTESession, LCTChannelAddedCallback, 
        LCTChannelRemovedCallback, pServiceSession))
    {
        ERRORPRINT("%s Failed to register LCT Channel callbacks\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }
    fROUTELCTChannelCallbacksRegistered = 1;

    if (ENATSC3ESGSessionAllocate(pServiceSession->hROUTESession,
        pServiceSession->strShortServiceName, &pServiceSession->hESGSession))
    {
        ERRORPRINT("%s Failed to allocate ESG session\n", 
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }

    if (ENATSC3ESGSessionRegisterCallbacks(pServiceSession->hESGSession,
        ESGServiceFragmentCallback, ESGScheduleFragmentCallback, 
        ESGContentFragmentCallback, pServiceSession))
    {
        ERRORPRINT("%s Failed to register ESG callbacks\n", 
            pServiceSession->strShortServiceName);
        rval = 1;
        goto errout;
    }

out:
    return rval;

errout:
    if (fESGCallbacksRegistered)
    {
        if (ENATSC3ESGSessionUnregisterCallbacks(pServiceSession->hESGSession,
            ESGServiceFragmentCallback, ESGScheduleFragmentCallback, 
            ESGContentFragmentCallback, pServiceSession))
        {
            ERRORPRINT("%s WARN: Failed to unregister ESG callbacks\n",
                pServiceSession->strShortServiceName);
        }
    }

    if (pServiceSession->hESGSession)
    {
        if (ENATSC3ESGSessionFree(pServiceSession->hESGSession))
        {
            ERRORPRINT("%s Failed to free ESG session\n",
                pServiceSession->strShortServiceName);
        }
    }

    if (fROUTELCTChannelCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterLCTChannelCallbacks(
            pServiceSession->hROUTESession, LCTChannelAddedCallback, 
            LCTChannelRemovedCallback, pServiceSession))
        {
            ERRORPRINT("%s WARN: Failed to unregister LCT Channel callbacks\n",
                pServiceSession->strShortServiceName);
        }
    }

    if (fROUTEObjectCallbacksRegistered)
    {
        if (ENATSC3ROUTESessionUnregisterObjectCallbacks(
            pServiceSession->hROUTESession, ROUTEObjectDataCallback, 
            ROUTESessionResetCallback, pServiceSession))
        {
            ERRORPRINT("%s WARN: Failed to unregister ROUTE object callbacks\n",
                pServiceSession->strShortServiceName);
        }
    }

    if (pServiceSession->hROUTESession)
    {
        if (ENATSC3ROUTESessionFree(pServiceSession->hROUTESession))
        {
            ERRORPRINT("%s: Failed to free ROUTE session\n",
                pServiceSession->strShortServiceName);
        }
        pServiceSession->hROUTESession = NULL;
    }

    pServiceSession->routeSessionType = ENATSC3ROUTESessionType_Invalid;

    goto out;
}

void
MMTPUSBDAvailableCallback(struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, 
    const char *pUSBD, int32_t sUSBD, int32_t version, void *pCbUserData)
{
    ServiceSession *pServiceSession = (ServiceSession *)pCbUserData;
    ATSC3MMTPSLSUSBDBundleDescriptionNode parsedUSBD;
    int32_t fHaveParsedUSBD = 0;

    if (ENATSC3MMTPSLSUSBDParse(pUSBD, sUSBD, &parsedUSBD))
    {
        ERRORPRINT("%s: WARN: Failed to parse USBD\n", 
            pServiceSession->strShortServiceName);
        goto out;
    }
    fHaveParsedUSBD = 1;

    if ((parsedUSBD.base.presentFlags &
         ATSC3MMTPSLSUSBDBundleDescriptionRouteChildNode_userServiceDescription) &&
        (parsedUSBD.userServiceDescription.base.presentFlags &
         ATSC3MMTPSLSUSBDUserServiceDescriptionChildNode_ComponentInfo))
    {
        ATSC3MMTPSLSUSBDComponentInfoNode *pComponentInfos;
        struct ServiceSessionStream *pStream;
        int32_t nComponentInfos;
        int32_t i;

        pComponentInfos = parsedUSBD.userServiceDescription.vComponentInfos;
        nComponentInfos = parsedUSBD.userServiceDescription.nComponentInfos;

        for (i = 0; i < nComponentInfos; i++, pComponentInfos++)
        {
            char *pComponentId;
            int32_t j;

            if (!(pComponentInfos->base.presentFlags &
                ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentType) ||
                !(pComponentInfos->base.presentFlags &
                ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentRole) ||
                !(pComponentInfos->base.presentFlags &
                ATSC3MMTPSLSUSBDComponentInfoNodeAttribute_componentId))
            {
                ERRORPRINT("%s: ComponentInfo Role/Type/Id missing\n",
                    pServiceSession->strShortServiceName);
                continue;
            }

            // XXX We should handle non-complete-main audio and non-primary
            // video.
            if (pComponentInfos->componentType ==
                ATSC3MMTPLLSUSBDComponentType_Audio)
            {
                if (pComponentInfos->componentRole !=
                    ATSC3MMTPLLSUSBDComponentRole_AudioCompleteMain)
                {
                    continue;
                }
            }
            else if (pComponentInfos->componentType ==
                ATSC3MMTPLLSUSBDComponentType_Video)
            {
                if (pComponentInfos->componentRole !=
                    ATSC3MMTPLLSUSBDComponentRole_VideoPrimary)
                {
                    continue;
                }
            }
            else if (pComponentInfos->componentType ==
                ATSC3MMTPLLSUSBDComponentType_Caption)
            {
                if (pComponentInfos->componentRole !=
                    ATSC3MMTPLLSUSBDComponentRole_CaptionNormal)
                {
                    continue;
                }
            }
            else
            {
                ERRORPRINT("Unhandled MMTP component type %d\n", 
                    pComponentInfos->componentType);
                continue;
            }

            pComponentId = (char *)EnXMLParserDataBufferGetPointer(
                &parsedUSBD.dataBuffer, 
                pComponentInfos->componentIdDataOffset);
            if (strlen(pComponentId) >= sizeof(pStream->mmtp.strAssetId))
            {
                ERRORPRINT("%s: Asset Id %s too long, ignoring component\n", 
                    pServiceSession->strShortServiceName, pComponentId);
                continue;
            }

            for (j = 0, pStream = pServiceSession->vStreams; 
                 j < numof(pServiceSession->vStreams); 
                 j++, pStream++)
            {
                if (pStream->mmtp.fActive && 
                    !strcmp(pStream->mmtp.strAssetId, pComponentId))
                {
                    break;
                }
            }

            if (j < numof(pServiceSession->vStreams))
            {
                LOGPRINT_NOFILE("%s: Existing stream for component %s\n",
                    pServiceSession->strShortServiceName, pComponentId);
                continue;
            }

            for (j = 0, pStream = pServiceSession->vStreams; 
                 j < numof(pServiceSession->vStreams); 
                 j++, pStream++)
            {
                if (!pStream->mmtp.fActive)
                {
                    break;
                }
            }

            if (j == numof(pServiceSession->vStreams))
            {
                ERRORPRINT("%s: Too many streams, dropping component %s\n",
                    pServiceSession->strShortServiceName, pComponentId);
                continue;
            }

            LOGPRINT_NOFILE("%s: added component: %s\n", 
                pServiceSession->strShortServiceName, pComponentId);
            strcpy(pStream->mmtp.strAssetId, pComponentId);
            pStream->mmtp.fActive = 1;

            pStream->mmtp.fHaveStreamSignalingInfo = 0;
        }

        for (i = 0, pStream = pServiceSession->vStreams; 
             i < numof(pServiceSession->vStreams); 
             i++, pStream++)
        {
            if (pStream->mmtp.fActive)
            {
                pServiceSession->fHaveMMTPUSBD = 1;
                break;
            }
        }
    }

out:
    if (fHaveParsedUSBD)
    {
        if (ENATSC3MMTPSLSUSBDFree(&parsedUSBD))
        {
            ERRORPRINT("%s: Failed to free parsed USBD\n",
                pServiceSession->strShortServiceName);
        }
    }

    return;
}

void
MMTPHELDAvailableCallback(struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, 
    const char *pHELD, int32_t sHELD, int32_t version, void *pUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pUserData;
    ENATSC3SLSHELDNode previousHELD;
    int32_t fHavePreviousHELD = 0;

    if (pthread_mutex_lock(&pServiceSession->slsLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    if (pServiceSession->fHaveHELD)
    {
        if (pServiceSession->heldVersion == version)
        {
            goto out;
        }

        LOGPRINT("NEW HELD Version %d -> %d\n", pServiceSession->heldVersion,
            version);
        previousHELD = pServiceSession->currentHELD;
        fHavePreviousHELD = 1;
    }

    if (ENATSC3SLSHELDParse(pHELD, sHELD, &pServiceSession->currentHELD))
    {
        ERRORPRINT("Failed to parse HELD\n");
        if (fHavePreviousHELD)
        {
            pServiceSession->currentHELD = previousHELD;
            fHavePreviousHELD = 0;
        }
        goto out;
    }

    ENATSC3SLSHELDHTMLEntryPackageNode *pEntryPackageNode;
    struct _HTMLEntryPackageInfo *pInfo;
    int32_t i;

    for (i = 0, pInfo = pServiceSession->vHTMLEntryPackageInfos; 
         i < pServiceSession->currentHELD.htmlEntryPackages.nOffsetsUsed; 
         i++, pInfo++)
    {
        if (i == numof(pServiceSession->vHTMLEntryPackageInfos))
        {
            ERRORPRINT("Too many HTMLEntryPakages (%d > %d)\n",
                pServiceSession->currentHELD.htmlEntryPackages.nOffsetsUsed,
                numof(pServiceSession->vHTMLEntryPackageInfos));
            break;
        }

        pEntryPackageNode = EnXMLParserDataOffsetArrayGetPointer(
            &pServiceSession->currentHELD.htmlEntryPackages, i);

        if (pEntryPackageNode->base.presentFlags & 
            ENATSC3SLSHELDHTMLEntryPackage_appContextId)
        {
            pInfo->pAppContextId = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer, 
                pEntryPackageNode->appContextIdDataOffset);
        }
        else
        {
            pInfo->pAppContextId = NULL;
        }

        if (pEntryPackageNode->base.presentFlags & 
            ENATSC3SLSHELDHTMLEntryPackage_requiredCapabilities)
        {
            pInfo->pRequiredCapabilities = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer, 
                pEntryPackageNode->requiredCapabilitiesDataOffset);
        }
        else
        {
            pInfo->pRequiredCapabilities = NULL;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_appRendering)
        {
            pInfo->fAppRendering = pEntryPackageNode->appRendering;
        }
        else
        {
            pInfo->fAppRendering = 0;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_clearAppContextCacheDate)
        {
            pInfo->clearAppContextCacheDateSecondsUTC = 
                pEntryPackageNode->clearAppContextCacheDateSecondsUTC;
        }
        else
        {
            pInfo->clearAppContextCacheDateSecondsUTC = 0x7fffffffffffffff;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_bcastEntryPackageUrl)
        {
            pInfo->pBroadcastEntryPackageUrl = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer,
                pEntryPackageNode->bcastEntryPackageUrlDataOffset);
        }
        else
        {
            pInfo->pBroadcastEntryPackageUrl = NULL;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_bcastEntryPageUrl)
        {
            pInfo->pBroadcastEntryPageUrl = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer,
                pEntryPackageNode->bcastEntryPageUrlDataOffset);
        }
        else
        {
            pInfo->pBroadcastEntryPageUrl = NULL;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_bbandEntryPageUrl)
        {
            pInfo->pBroadbandEntryPageUrl = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer,
                pEntryPackageNode->bbandEntryPageUrlDataOffset);
        }
        else
        {
            pInfo->pBroadbandEntryPageUrl = NULL;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_validFrom)
        {
            pInfo->validFromSecondsUTC = pEntryPackageNode->validFromSecondsUTC;
        }
        else
        {
            pInfo->validFromSecondsUTC = 0;
        }

        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_validUntil)
        {
            pInfo->validUntilSecondsUTC 
                = pEntryPackageNode->validUntilSecondsUTC;
        }
        else
        {
            pInfo->validUntilSecondsUTC = 0x7fffffffffffffff;
        }

        pInfo->nTSIs = 0;
        if (pEntryPackageNode->base.presentFlags &
            ENATSC3SLSHELDHTMLEntryPackage_lctTSIRef)
        {
            char *pTSIs;
            
            pTSIs = EnXMLParserDataBufferGetPointer(
                &pServiceSession->currentHELD.dataBuffer,
                pEntryPackageNode->lctTSIRefDataOffset);
           
            while (pInfo->nTSIs < numof(pInfo->vTSIs))
            {
                char *pEndPtr;

                pInfo->vTSIs[pInfo->nTSIs] = strtol(pTSIs, &pEndPtr, 10);
                if (pEndPtr == pTSIs)
                {
                    // No number found
                    break;
                }

                pInfo->nTSIs++;
                pTSIs = pEndPtr;
            }
        }
    }

    pServiceSession->nHTMLEntryPackageInfos = i;
    pServiceSession->fHaveHELD = 1;
    pServiceSession->heldVersion = version;
    pServiceSession->tHELDUTC = *pPacketTimeUTC;

out:
    if (pthread_mutex_unlock(&pServiceSession->slsLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    if (fHavePreviousHELD)
    {
        if (ENATSC3SLSHELDFree(&previousHELD))
        {
            ERRORPRINT("Failed to free HELD\n");
        }
    }
}

void
MMTPVSPDAvailableCallback(struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, 
    const uint8_t *pVSPD, int32_t sVSPD, int32_t version, void *pCbUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;
    int32_t nAssets;
    int32_t i;

    nAssets = *(pVSPD++);
    LOGPRINT_NOFILE("%s: IMPLEMENT ME VSPD Number of assets: %d\n", 
        pServiceSession->strShortServiceName, nAssets);
     
    for (i = 0; i < nAssets; i++)
    {
        int32_t j;
        uint32_t assetIdLength;
        
        assetIdLength = (pVSPD[0] << 24) | (pVSPD[1] << 16) | 
            (pVSPD[2] << 8) | pVSPD[3];
        pVSPD += 4;
        for (j = 0; j < assetIdLength; j++)
        {
            printf("%.2x", *(pVSPD++));
        }
        printf("\n");
    }

    return;
}

void
MMTPASPDAvailableCallback(struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, 
    const uint8_t *pASPD, int32_t sASPD, int32_t version, void *pCbUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;
    int32_t nAssets;
    int32_t i;

    nAssets = *(pASPD++);
    LOGPRINT_NOFILE("%s: IMPLEMENT ME ASPD Number of assets: %d\n", 
        pServiceSession->strShortServiceName, nAssets);
     
    for (i = 0; i < nAssets; i++)
    {
        int32_t j;
        uint32_t assetIdLength;
        
        assetIdLength = (pASPD[0] << 24) | (pASPD[1] << 16) | 
            (pASPD[2] << 8) | pASPD[3];
        pASPD += 4;
        for (j = 0; j < assetIdLength; j++)
        {
            printf("%.2x", *(pASPD++));
        }
        printf("\n");
    }

    return;
}

void
MMTPCADAvailableCallback(struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, const ENATSC3MMTPATSC3SignalingInfo *pSignalingInfo, 
    const uint8_t *pCAD, int32_t sCAD, int32_t version, void *pCbUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;
    int32_t nAssets;
    int32_t i;

    nAssets = *(pCAD++);
    LOGPRINT_NOFILE("%s: IMPLEMENT ME CAD Number of assets: %d\n", 
        pServiceSession->strShortServiceName, nAssets);
     
    for (i = 0; i < nAssets; i++)
    {
        int32_t j;
        uint32_t assetIdLength;
        
        assetIdLength = (pCAD[0] << 24) | (pCAD[1] << 16) | 
            (pCAD[2] << 8) | pCAD[3];
        pCAD += 4;
        for (j = 0; j < assetIdLength; j++)
        {
            printf("%.2x", *(pCAD++));
        }
        printf("\n");
    }

    return;
}

void
MMTPMPTSignalingMessageCallback(struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, const ENATSC3MMTPMPTSignalingInfo *pSignalingInfo, 
    void *pCbUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;
    const ENATSC3MMTPMPTSignalingInfoAsset *pAsset;
    struct ServiceSessionStream *pStream;
    int32_t i;
    int32_t j;

    // We need to process a USBD first so that we know which audio / video
    // assets we wish to consume.
    if (!pServiceSession->fHaveMMTPUSBD)
    {
        LOGPRINT_NOFILE("Ignoring MMTP message - no usbd\n");
        goto out;
    }

    // Table id 0x20 is a complete MP table. 
    // Table id 0x11 is subset_0 which "contains the minimum information
    //     required for Package consumption," i.e. the package id and 
    //     table descriptors.
    if (pSignalingInfo->table_id != 0x11 &&
        pSignalingInfo->table_id != 0x20)
    {
        goto out;
    }

    if (pSignalingInfo->message_version == 
        pServiceSession->mmtpMPTTableVersion &&
        pSignalingInfo->payloadCRC32 == 
        pServiceSession->mmtpMPTTableCRC32)
    {
        goto out;
    }

    if (pServiceSession->mmtpMPTTableVersion != -1)
    {
        LOGPRINT("%s: MMTP reconfiguration on MPT table version change %d/%.8X -> %d/%.8X\n", 
            pServiceSession->strShortServiceName, 
            pServiceSession->mmtpMPTTableVersion, 
            pServiceSession->mmtpMPTTableCRC32,
            pSignalingInfo->message_version,
            pSignalingInfo->payloadCRC32);
    }

    for (i = 0, pStream = pServiceSession->vStreams; 
         i < numof(pServiceSession->vStreams); 
         i++, pStream++)
    {
        pStream->mmtp.fHaveStreamSignalingInfo = 0;
    }

    for (i = 0, pAsset = pSignalingInfo->vAssets;
         i < pSignalingInfo->nAssets;
         i++, pAsset++)
    {
        int32_t streamIdx = -1;

        if (pAsset->Identifier_mapping.type != 0)
        {
            ERRORPRINT("%s: Unhandled identifier mapping type %d\n",
                pServiceSession->strShortServiceName,
                pAsset->Identifier_mapping.type);
            continue;
        }

        for (streamIdx = 0, pStream = pServiceSession->vStreams; 
             streamIdx < numof(pServiceSession->vStreams); 
             streamIdx++, pStream++)
        {
            if (pStream->mmtp.fActive && strlen(pStream->mmtp.strAssetId) == 
                pAsset->Identifier_mapping.asset.asset_id_length &&
                !strncmp(pStream->mmtp.strAssetId, 
                (char *)pAsset->Identifier_mapping.asset.pAssetId,
                pAsset->Identifier_mapping.asset.asset_id_length)) 
            {
                break;
            }
        }

        if (streamIdx == numof(pServiceSession->vStreams))
        {
            LOGPRINT_NOFILE("Couldn't find stream for %.*s\n",
                pAsset->Identifier_mapping.asset.asset_id_length,
                (char *)pAsset->Identifier_mapping.asset.pAssetId);
            continue;
        }

        pStream->mmtp.fHaveStreamSignalingInfo = 1;

        for (j = 0; j < pAsset->nLocations; j++)
        {
            if (pAsset->vLocations[j].location_type == 0)
            {
                pStream->mmtp.packet_id = pAsset->vLocations[j].packet_id;
                LOGPRINT("%s: Selected packet_id %d for %s\n",
                    pServiceSession->strShortServiceName,
                    pAsset->vLocations[j].packet_id, 
                    pStream->mmtp.strAssetId);
                break;
            }
        }
    }

    for (i = 0, pStream = pServiceSession->vStreams; 
         i < numof(pServiceSession->vStreams); 
         i++, pStream++)
    {
        if (pStream->mmtp.fActive && !pStream->mmtp.fHaveStreamSignalingInfo)
        {
            ERRORPRINT("%s: WARN: No asset for stream %s, omitting\n", 
                pServiceSession->strShortServiceName, pStream->mmtp.strAssetId);
            continue;
        }
    }

    pServiceSession->mmtpMPTTableVersion = pSignalingInfo->message_version;
    pServiceSession->mmtpMPTTableCRC32 = pSignalingInfo->payloadCRC32;

out:
    return;
}

void
MMTPMediaDataCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t packet_id, const char *pAssetId, int32_t fragment_type, 
    uint32_t sequence_number, uint64_t mpuPresentationTime, 
    uint32_t mpuDuration, uint64_t timeScale, const uint8_t *pData, 
    int32_t sData, void *pCbUserData)
{
    ServiceSession *pServiceSession = 
        (ServiceSession *)pCbUserData;
    struct ServiceSessionStream *pStream;
    int32_t i;

    for (i = 0, pStream = pServiceSession->vStreams; 
         i < numof(pServiceSession->vStreams);
         i++, pStream++)
    {
        if (pStream->mmtp.fActive)
        {
            // We are checking the packet_id first and then the asset id if the
            // packet_id is not valid. While it is more efficient to compare 
            // the packet_id, sometimes the USBD is delivered before a type 0 
            // fragment while the full MPT is delivered after. In that case we 
            // would not accept the type 0 fragment since we didn't know the 
            // packet_id (which comes from the MPT), although we do know the 
            // asset name (from the USBD).
            if (packet_id == pStream->mmtp.packet_id ||
                (pStream->mmtp.packet_id == -1 && 
                 !strcmp(pStream->mmtp.strAssetId, pAssetId)))
            {
                break;
            }
        }
    }

    if (i == numof(pServiceSession->vStreams))
    {
        goto out;
    }

    // For each file we expect a fragment of type 0 to appear first, followed
    // by a type 1 fragment and finally a type 2. 
    if (fragment_type == 0)
    {
        if (pStream->fd >= 0)
        {
            close(pStream->fd);
            pStream->fd = -1;
        }

        char strFilename[256];
        if (snprintf(strFilename, sizeof(strFilename), "%s/%s-%06d-%s", 
            pServiceSession->pSession->pOutPath, 
            pServiceSession->strShortServiceName, 
            pServiceSession->mediaFileCounter++, pStream->mmtp.strAssetId) >= 
            sizeof(strFilename))
        {
            ERRORPRINT("Filename too long\n");
            goto out;
        }
        pStream->fd = open(strFilename, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (pStream->fd < 0)
        {
            fprintf(stderr, "Failed to open %s (%s)\n", strFilename, 
                strerror(errno));
            goto out;
        }
    }

    if (pStream->fd >= 0)
    {
        if (write(pStream->fd, pData, sData) != sData)
        {
            fprintf(stderr, "Failed to write to file: %s\n", 
                strerror(errno));
        }
    }

out:
    return;
}

static int32_t
SetupMMTPSession(ServiceSession *pServiceSession, uint32_t addr, 
    uint16_t port)
{
    int32_t i;
    int32_t rval = 0;

    if (ENATSC3MMTPSessionAllocate(addr, port, AddIPAddress, RemoveIPAddress, 
        pServiceSession, pServiceSession->strShortServiceName, 
        &pServiceSession->hMMTPSession))
    {
        ERRORPRINT("%s: Failed to allocate MMTP session\n",
            pServiceSession->strShortServiceName);
        rval = 1;
        goto out;
    }

    if (ENATSC3MMTPSessionRegisterCallbacks(pServiceSession->hMMTPSession,
        MMTPUSBDAvailableCallback, MMTPHELDAvailableCallback, 
        MMTPVSPDAvailableCallback, MMTPASPDAvailableCallback, 
        MMTPCADAvailableCallback, MMTPMPTSignalingMessageCallback, 
        MMTPMediaDataCallback, pServiceSession))
    {
        ERRORPRINT("Failed to register MMTP callbacks\n");
        rval = 1;
        goto errout;
    }

    pServiceSession->fHaveMMTPUSBD = 0;
    pServiceSession->mmtpMPTTableVersion = -1;
    for (i = 0; i < numof(pServiceSession->vStreams); i++)
    {
        pServiceSession->vStreams[i].mmtp.packet_id = -1;
    }

out:
    return rval;

errout:
    if (pServiceSession->hMMTPSession)
    {
        if (ENATSC3MMTPSessionFree(pServiceSession->hMMTPSession))
        {
            ERRORPRINT("WARN: Failed to free MMTP session\n");
        }
        pServiceSession->hMMTPSession = NULL;
    }

    goto out;
}

int32_t
SetupService(ServiceSession *pServiceSession, 
    ATSC3LLSSLTServiceCategory category, ATSC3LLSSLTSLSProtocol protocol,
    struct in_addr inaddr, uint32_t port)
{
    int32_t rval = 0;

    switch (category)
    {
        case ATSC3LLSSLTServiceCategory_LinearAV:
        case ATSC3LLSSLTServiceCategory_LinearAudioOnly:
        case ATSC3LLSSLTServiceCategory_AppBased:
            switch (protocol)
            {
                case ATSC3LLSSLTSLSProtocol_ROUTE:
                    if (SetupROUTEDASHSession(pServiceSession, 
                        htonl(inaddr.s_addr), port))
                    {
                        ERRORPRINT(
                            "Failed to set up ROUTE/DASH session\n");
                        rval = 1;
                        break;
                    }
                    break;
                case ATSC3LLSSLTSLSProtocol_MMTP:
                    if (SetupMMTPSession(pServiceSession, htonl(inaddr.s_addr), 
                        port))
                    {
                        ERRORPRINT(
                            "Failed to set up ROUTE/DASH session\n");
                        rval = 1;
                        break;
                    }
                    break;
                default:
                    ERRORPRINT("Unhandled SLS Protocol %d\n", protocol);
                    rval = 1;
                    break;
            }
            break;
        case ATSC3LLSSLTServiceCategory_ESG:
            if (SetupROUTEESGSession(pServiceSession, htonl(inaddr.s_addr), 
                port))
            {
                ERRORPRINT("Failed to set up ESG session\n");
                rval = 1;
                break;
            }
            break;
        case ATSC3LLSSLTServiceCategory_EAS:
        case ATSC3LLSSLTServiceCategory_Reserved:
            ERRORPRINT("Unhandled service category %d\n", category);
            rval = 1;
            break;
    }

    return rval;
}

int32_t
ProcessSLT(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const ENATSC3LLSTableHeader *pLLSTableHeader, const uint8_t *pPayload, 
    int32_t payloadLength, uint8_t sltTableVersion, ENATSC3Session *pSession)
{
    char strXML[65536];
    int32_t xmlLength;
    ATSC3LLSSLT_SLTNode parsedSLT;
    ServiceSession*pServiceSession;
    int32_t fHaveParsedSLT = 0;
    SLTEntry vSLTEntries[32];
    SLTEntry *pSLTEntry;
    struct timespec mononow;
    int32_t nSLTEntriesUsed = 0;
    uint32_t gzipCRC32;
    uint32_t gzipILEN;
    const uint8_t *pGZIPFooter;
    int32_t i;
    int32_t rval = 0;

    pGZIPFooter = pPayload + payloadLength - 8;
    gzipCRC32 = (pGZIPFooter[3] << 24) | (pGZIPFooter[2] << 16) |
        (pGZIPFooter[1] << 8) | pGZIPFooter[0];
    gzipILEN = (pGZIPFooter[7] << 24) | (pGZIPFooter[6] << 16) |
        (pGZIPFooter[5] << 8) | pGZIPFooter[4];

    if (EnGunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength) != Z_OK)
    {
        ERRORPRINT("Failed to unzip\n");
        rval = 1;
        goto out;
    }

    if (!xmlLength)
    {
        ERRORPRINT("Invalid SLT length %d\n", xmlLength);
        rval = 1;
        goto out;
    }

    if (xmlLength != gzipILEN)
    {
        ERRORPRINT("WARN: unzip'd length mismatch %u != %u\n", gzipILEN,
            xmlLength);
    }
    else
    {
        uint32_t crc32;

        crc32 = ENUtils_CalcCRC32LSB((uint8_t *)strXML, xmlLength);
        if (crc32 != gzipCRC32)
        {
            ERRORPRINT("WARN: unzip'd crc32 mismatch %.8X != %.8X\n",
                crc32, gzipCRC32);
        }
    }

    if (ATSC3LLSParseSLT(strXML, xmlLength,  &parsedSLT))
    {
        ERRORPRINT("%"PRId64" Failed to parse SLT\n", packetNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedSLT = 1;

    clock_gettime(CLOCK_MONOTONIC, &mononow);

    // This first pass is for statistics gathering
    memset(vSLTEntries, 0, sizeof(vSLTEntries));

    for (i = 0;
         i < parsedSLT.services.nOffsetsUsed && 
         nSLTEntriesUsed < numof(vSLTEntries); 
         i++)
    {
        ATSC3LLSSLT_ServiceNode *pServiceNode;

        pServiceNode = (ATSC3LLSSLT_ServiceNode *)
            EnXMLParserDataOffsetArrayGetPointer(&parsedSLT.services, i);
        
        if (pServiceNode->base.presentFlags & 
             ATSC3LLSSLTServiceNodeAttribute_serviceId)
        {
            pSLTEntry = vSLTEntries + nSLTEntriesUsed;
            pSLTEntry->groupId = pLLSTableHeader->lls_group_id; 
            pSLTEntry->serviceId = pServiceNode->serviceId;
            pSLTEntry->tAdded = mononow;
            pSLTEntry->presentMask = 0;

            pSLTEntry->presentMask |= SLTEntryPresentFlag_ServiceId;

            if (pServiceNode->base.presentFlags & 
                ATSC3LLSSLTServiceNodeAttribute_serviceCategory)
            {
                pSLTEntry->serviceCategory = 
                    pServiceNode->serviceCategory;
                pSLTEntry->presentMask |= 
                    SLTEntryPresentFlag_ServiceCategory;
            }

            if (pServiceNode->base.presentFlags & 
                ATSC3LLSSLTServiceNodeAttribute_protected)
            {
                if (pServiceNode->protected)
                {
                    pSLTEntry->flags |= SLTEntryFlag_Protected;
                }
                pSLTEntry->presentMask |= SLTEntryPresentFlag_Protected;
            }

            if (pServiceNode->base.presentFlags & 
                ATSC3LLSSLTServiceNodeAttribute_hidden)
            {
                if (pServiceNode->hidden)
                {
                    pSLTEntry->flags |= SLTEntryFlag_Hidden;
                }
                pSLTEntry->presentMask |= SLTEntryPresentFlag_Hidden;
            }

            if (pServiceNode->base.presentFlags & 
                ATSC3LLSSLTServiceNodeAttribute_essential)
            {
                if (pServiceNode->essential)
                {
                    pSLTEntry->flags |= SLTEntryFlag_Essential;
                }
                pSLTEntry->presentMask |= SLTEntryPresentFlag_Essential;
            }

            if (pServiceNode->base.presentFlags & 
                ATSC3LLSSLTServiceNodeAttribute_broadbandAccessRequired)
            {
                if (pServiceNode->broadbandAccessRequired)
                {
                    pSLTEntry->flags |= SLTEntryFlag_BroadbandAccessRequired;
                }
                pSLTEntry->presentMask |= 
                    SLTEntryPresentFlag_BroadbandAccessRequired;
            }

            if (pServiceNode->base.presentFlags & 
                ATSC3LLSSLTServiceNodeAttribute_majorChannelNo)
            {
                pSLTEntry->majorChannelNo = pServiceNode->majorChannelNo; 
                pSLTEntry->presentMask |= SLTEntryPresentFlag_MajorChannelNo;
            }

            if (pServiceNode->base.presentFlags & 
                ATSC3LLSSLTServiceNodeAttribute_minorChannelNo)
            {
                pSLTEntry->minorChannelNo = pServiceNode->minorChannelNo; 
                pSLTEntry->presentMask |= SLTEntryPresentFlag_MinorChannelNo;
            }

            if (pServiceNode->base.presentFlags & 
                ATSC3LLSSLTServiceNodeAttribute_shortServiceName)
            {
                strncpy(pSLTEntry->strShortServiceName, 
                    EnXMLParserDataBufferGetPointer(&parsedSLT.dataBuffer, 
                        pServiceNode->shortServiceNameDataOffset), 
                    sizeof(pSLTEntry->strShortServiceName));
                pSLTEntry->strShortServiceName[
                    sizeof(pSLTEntry->strShortServiceName) - 1] = 0;
                pSLTEntry->presentMask |= SLTEntryPresentFlag_ShortServiceName;
            }

            if (pServiceNode->broadcastSvcSignaling.base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsProtocol)
            {
                pSLTEntry->slsProtocol = 
                    pServiceNode->broadcastSvcSignaling.slsProtocol; 
                pSLTEntry->presentMask |= SLTEntryPresentFlag_SLSProtocol;
            }

            if (pServiceNode->broadcastSvcSignaling.base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationIpAddress)
            {
                struct in_addr inaddr;

                if (!inet_aton(
                    pServiceNode->broadcastSvcSignaling.slsDestinationIpAddress,
                    &inaddr))
                {
                    ERRORPRINT("Invalid IP address %s\n", 
                        pServiceNode->broadcastSvcSignaling.slsDestinationIpAddress);
                    continue;
                }

                pSLTEntry->slsDestinationIpAddress = htonl(inaddr.s_addr);
                pSLTEntry->presentMask |= 
                    SLTEntryPresentFlag_DestinationIPAddress;
            }

            if (pServiceNode->broadcastSvcSignaling.base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationUdpPort)
            {
                pSLTEntry->slsDestinationPort = 
                    pServiceNode->broadcastSvcSignaling.slsDestinationUdpPort;
                pSLTEntry->presentMask |= 
                    SLTEntryPresentFlag_DestinationUdpPort;
            }

            nSLTEntriesUsed++;
        }
        else
        {
            ERRORPRINT("WARN: SLT Entry missing service id");
        }
    }

    if (pthread_mutex_lock(&pSession->llsLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    for (i = 0; i < nSLTEntriesUsed; i++)
    {
        if (AddSLTEntry_Locked(pSession, vSLTEntries + i, &mononow))
        {
            ERRORPRINT("Failed to add SLT entry\n");
        }
    }

    RemoveSLTEntriesForGroupByValidTime_Locked(pSession, 
        pLLSTableHeader->lls_group_id, &mononow); 

    if (pthread_mutex_unlock(&pSession->llsLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    // Remove sessions which are no longer in the SLT.
    if (pthread_mutex_lock(&pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pServiceSession = pSession->pServiceSessions;
    while (pServiceSession)
    {
        if (pServiceSession->lls_group_id == 
            pLLSTableHeader->lls_group_id &&
            !FindSLTEntry_Locked(pSession, pServiceSession->lls_group_id,
            pServiceSession->serviceId))
        {
            // Unlink the service session from the list of active sessions
            RemoveServiceSession_ServiceSessionListLocked(pSession, 
                pServiceSession);

            // We cannot hold the serviceRecontinerSessionList lock 
            // while stopping the service since stopping the service will
            // result in the IP filter set changing, which will grab
            // the serviceRecontinerSessionList lock. 
            if (pthread_mutex_unlock(
                &pSession->serviceSessionsListLock))
            {
                ERRORPRINT("Failed to unlock\n");
            }

            LOGPRINT("Removing service %d %s %d-%d no longer in SLT\n", 
                pServiceSession->serviceId, 
                pServiceSession->strShortServiceName,
                pServiceSession->channelMajor,
                pServiceSession->channelMajor);

            // And shut it down
            if (ENATSC3SessionServiceSessionStop(pServiceSession))
            {
                ERRORPRINT("Failed to stop service %d %s %d-%d\n",
                    pServiceSession->serviceId, 
                    pServiceSession->strShortServiceName,
                    pServiceSession->channelMajor, 
                    pServiceSession->channelMinor);
            }

            if (ENATSC3SessionServiceSessionFree(pServiceSession))
            {
                ERRORPRINT("Failed to free service %d %s %d-%d\n",
                    pServiceSession->serviceId, 
                    pServiceSession->strShortServiceName,
                    pServiceSession->channelMajor,
                    pServiceSession->channelMinor);
            }

            if (pthread_mutex_lock(
                &pSession->serviceSessionsListLock))
            {
                ERRORPRINT("Failed to lock\n");
            }

            // Restart from the beginning just in case the list has
            // changed.
            pServiceSession = pSession->pServiceSessions;
            continue;
        }

        pServiceSession = pServiceSession->pNext;
    }

    if (pthread_mutex_unlock(
        &pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    // This second pass is for actually tuning to the service we are 
    // interested in. See if we can combine the loops.
    for (i = 0; i < parsedSLT.services.nOffsetsUsed; i++)
    {
        ATSC3LLSSLT_ServiceNode *pServiceNode;
        char *pShortServiceName = NULL;
        char *pGlobalServiceID = NULL;
        uint32_t majorChannelNo;
        uint32_t minorChannelNo;
        uint32_t slsProtocol;
        uint32_t serviceCategory;
        uint32_t serviceId;
        uint32_t slsDestinationUdpPort;
        char *slsDestinationIpAddress;

        pServiceNode = (ATSC3LLSSLT_ServiceNode *)
            EnXMLParserDataOffsetArrayGetPointer(&parsedSLT.services, i);

        // We need to know what type of service this is...
        if (!(pServiceNode->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_serviceCategory))
        {
            ERRORPRINT("Service missing service category in SLT!\n");
            continue;
        }
        serviceCategory = pServiceNode->serviceCategory;

        if (!(pServiceNode->base.presentFlags & 
            ATSC3LLSSLTServiceNodeAttribute_serviceId))
        {
            ERRORPRINT("Service missing service id in SLT!\n");
            continue;
        }
        serviceId = pServiceNode->serviceId;

        if (!(pServiceNode->base.presentFlags & 
            ATSC3LLSSLTServiceNodeChildNode_BroadcastSvcSignaling))
        {
            ERRORPRINT("Service missing broadcastSvcSignaling in SLT\n");
            continue;
        }

        if (!(pServiceNode->broadcastSvcSignaling.base.presentFlags &
            ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsProtocol))
        {
            ERRORPRINT("Service missing SLS protocol in SLT\n");
            continue;
        }
        slsProtocol = pServiceNode->broadcastSvcSignaling.slsProtocol;

        if (!(pServiceNode->broadcastSvcSignaling.base.presentFlags &
              ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationIpAddress) ||
            !(pServiceNode->broadcastSvcSignaling.base.presentFlags &
              ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsDestinationUdpPort))
        {
            ERRORPRINT("Service missing IP address or port in SLT\n");
            continue;
        }
        slsDestinationUdpPort = 
            pServiceNode->broadcastSvcSignaling.slsDestinationUdpPort;
        slsDestinationIpAddress =
            pServiceNode->broadcastSvcSignaling.slsDestinationIpAddress;

        if (pServiceNode->base.presentFlags &
            ATSC3LLSSLTServiceNodeAttribute_shortServiceName)
        {
            pShortServiceName = EnXMLParserDataBufferGetPointer(
                &parsedSLT.dataBuffer,
                pServiceNode->shortServiceNameDataOffset);
        }

        if (pServiceNode->base.presentFlags &
            ATSC3LLSSLTServiceNodeAttribute_globalServiceID)
        {
            pGlobalServiceID = EnXMLParserDataBufferGetPointer(
                &parsedSLT.dataBuffer,
                pServiceNode->globalServiceIDDataOffset);
        }

        // Linear, audio-only and app-based services must have a major/minor 
        // channel number, destination IP address/port, and SLS protocol aa
        // to be processed further.
        //
        // ESG services must have a destination IP address/port, and SLS 
        // protocol.
        //
        // We don't handle EAS services at present.
        if (IsLinearService(serviceCategory) ||
            IsAppBasedService(serviceCategory))
        {
            if (!(pServiceNode->base.presentFlags & 
                  ATSC3LLSSLTServiceNodeAttribute_majorChannelNo) ||
                !(pServiceNode->base.presentFlags &
                  ATSC3LLSSLTServiceNodeAttribute_minorChannelNo))
            {
                continue;
            }

            majorChannelNo = pServiceNode->majorChannelNo;
            minorChannelNo = pServiceNode->minorChannelNo;
        }
        else if (IsESGService(serviceCategory))
        {
            // ESG doesn't have a major/minor channel number, so just make one
            // up.
            majorChannelNo = 0xffffffff;
            minorChannelNo = 0xffffffff;
        }
        else
        {
            continue;
        }

        // If requestedMajorChannelNo is -1 then we handle all services, 
        // otherwise majorChannelNo and minorChannelNo specify the channel to 
        // handle. We also handle ESG when a specific service is requested.
        if (pSession->requestedChannelMajor != -1)
        {
            if (IsLinearService(serviceCategory) ||
                IsAppBasedService(serviceCategory))
            {
                if (majorChannelNo != pSession->requestedChannelMajor ||
                    minorChannelNo != pSession->requestedChannelMinor)
                {
                    continue;
                }
            }
            else if (!IsESGService(serviceCategory))
            {
                continue;
            }
        }

        if (pthread_mutex_lock(&pSession->serviceSessionsListLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        pServiceSession = pSession->pServiceSessions;
        while (pServiceSession)
        {
            if (pServiceSession->serviceId == serviceId)
            {
                break;
            }

            pServiceSession = pServiceSession->pNext;
        }

        // Check if this is a reconfiguration of the channel
        if (pServiceSession)
        {
            if ((pServiceSession->serviceGroupId != 
                 pLLSTableHeader->lls_group_id) ||
                (majorChannelNo != pServiceSession->channelMajor) ||
                (minorChannelNo != pServiceSession->channelMinor) ||
                (slsProtocol != pServiceSession->slsProtocol))
            {
                LOGPRINT("Service changed groupId %d ch %d-%d protocol %d -> groupId %d ch %d-%d protocol %d\n",
                    pServiceSession->serviceGroupId, 
                    pServiceSession->channelMajor, 
                    pServiceSession->channelMinor,
                    pServiceSession->slsProtocol,
                    pLLSTableHeader->lls_group_id,
                    majorChannelNo,
                    minorChannelNo,
                    slsProtocol);

                // Unlink the service session from the list
                // of active sessions
                RemoveServiceSession_ServiceSessionListLocked(
                    pSession, pServiceSession);

                // XXX We cannot hold the serviceRecontinerSessionList lock 
                // while stopping the service since stopping the service will
                // result in the IP filter set changing, which will grab
                // the serviceRecontinerSessionList lock.
                if (pthread_mutex_unlock(
                    &pSession->serviceSessionsListLock))
                {
                    ERRORPRINT("Failed to unlock\n");
                }

                LOGPRINT("Removing service %d %s %d-%d due to reconfigure\n", 
                    pServiceSession->serviceId, 
                    pShortServiceName ? pShortServiceName : "", 
                    pServiceSession->channelMajor,
                    pServiceSession->channelMajor);

                // And shut it down
                if (ENATSC3SessionServiceSessionStop( pServiceSession))
                {
                    ERRORPRINT("Failed to stop service %d %s %d-%d\n",
                        pServiceSession->serviceId, 
                        pShortServiceName ? pShortServiceName : "",
                        pServiceSession->channelMajor, 
                        pServiceSession->channelMinor);
                }

                if (ENATSC3SessionServiceSessionFree( pServiceSession))
                {
                    ERRORPRINT("Failed to free service %d %s %d-%d\n",
                        pServiceSession->serviceId, 
                        pShortServiceName ? pShortServiceName : "",
                        pServiceSession->channelMajor,
                        pServiceSession->channelMinor);
                }

                pServiceSession = NULL;
            }
            else
            {
                if (pthread_mutex_unlock(
                    &pSession->serviceSessionsListLock))
                {
                    ERRORPRINT("Failed to unlock\n");
                }
            }
        }
        else
        {
            if (pthread_mutex_unlock(
                &pSession->serviceSessionsListLock))
            {
                ERRORPRINT("Failed to unlock\n");
            }
        }

        // If this is a new service create it.
        if (!pServiceSession)
        {
            struct in_addr inaddr;

            if (!inet_aton( slsDestinationIpAddress, &inaddr))
            {
                ERRORPRINT("Invalid IP address %s\n", slsDestinationIpAddress);
                continue;
            }

            if (ENATSC3SessionServiceSessionCreate( &pServiceSession))
            {
                ERRORPRINT(
                    "Failed to create session for serviceId %d %s %d-%d\n",
                    serviceId, 
                    pShortServiceName ? pShortServiceName : "",
                    majorChannelNo,
                    minorChannelNo);
                continue;
            }

            pServiceSession->fRunning = 1;
            pServiceSession->pSession = pSession;
            pServiceSession->verbosity = pSession->verbosity;
            pServiceSession->sltTableVersion = sltTableVersion;
            pServiceSession->serviceId = serviceId;
            pServiceSession->serviceCategory = serviceCategory;
            pServiceSession->lls_group_id = pLLSTableHeader->lls_group_id;
            pServiceSession->sltSvcSeqNum = pServiceNode->sltSvcSeqNum;
            if (pGlobalServiceID)
            {
                strncpy(pServiceSession->strGlobalServiceID, pGlobalServiceID, 
                    sizeof(pServiceSession->strGlobalServiceID));
                if (pServiceSession->strGlobalServiceID[
                    sizeof(pServiceSession->strGlobalServiceID) - 1])
                {
                    ERRORPRINT("Service name %s too long\n", pGlobalServiceID); 
                    pServiceSession->strGlobalServiceID[0] = 0;
                }
            }
            else
            {
                pServiceSession->strGlobalServiceID[0] = 0;
            }
            pServiceSession->serviceGroupId = pLLSTableHeader->lls_group_id;
            pServiceSession->channelMajor = majorChannelNo;
            pServiceSession->channelMinor = minorChannelNo;
            pServiceSession->slsProtocol = slsProtocol;
            if (pShortServiceName)
            {
                strncpy(pServiceSession->strShortServiceName, 
                    pShortServiceName,  
                    sizeof(pServiceSession->strShortServiceName) - 1); 
                pServiceSession->strShortServiceName[
                    sizeof(pServiceSession->strShortServiceName) - 1] = 0;
            }
            else
            {
                pServiceSession->strShortServiceName[0] = 0;
            }

            if (pServiceNode->broadcastSvcSignaling.base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsMajorProtocolVersion)
            {
                pServiceSession->slsMajorProtocolVersion = 
                    pServiceNode->broadcastSvcSignaling.slsMajorProtocolVersion;
            }

            if (pServiceNode->broadcastSvcSignaling.base.presentFlags &
                ATSC3LLSSLTBroadcastSvcSignalingNodeAttribute_slsMinorProtocolVersion)
            {
                pServiceSession->slsMinorProtocolVersion = 
                    pServiceNode->broadcastSvcSignaling.slsMinorProtocolVersion;
            }

            LOGPRINT("Adding session for serviceId %d %s %d-%d\n",
                serviceId, 
                pShortServiceName ? pShortServiceName : "",
                majorChannelNo, minorChannelNo);

            // Adding the service to the list makes it active
            if (pthread_mutex_lock(
                &pSession->serviceSessionsListLock))
            {
                ERRORPRINT("Failed to lock\n");
            }

            InsertServiceSession_ServiceSessionListLocked(pSession, 
                pServiceSession);

            if (pthread_mutex_unlock(
                &pSession->serviceSessionsListLock))
            {
                ERRORPRINT("Failed to unlock\n");
            }

            if (SetupService(pServiceSession, serviceCategory, slsProtocol, 
                inaddr, slsDestinationUdpPort))
            {
                ERRORPRINT("Failed to setup service %s\n",
                    pServiceSession->strShortServiceName);
            }
        }
    }

out:
    if (fHaveParsedSLT)
    {
        if (ATSC3LLSFreeParsedSLT(&parsedSLT))
        {
            ERRORPRINT("Failed to free parsed SLT\n");
        }
    }

    return rval;
}

int32_t
ProcessAEAT(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const ENATSC3LLSTableHeader *pLLSTableHeader, const uint8_t *pPayload, 
    int32_t payloadLength, uint8_t aeatTableVersion, 
    ENATSC3Session *pSession)
{
    char strXML[65536];
    int32_t xmlLength;
    ATSC3LLSAEAT_AEATNode parsedAEAT;
    int32_t fHaveParsedAEAT = 0;
    int32_t rval = 0;

    if (EnGunzipUnpipe((uint8_t *)pPayload, payloadLength, 
        (unsigned char *)strXML, sizeof(strXML), &xmlLength) != Z_OK)
    {
        ERRORPRINT("%"PRId64" Failed to unzip AEAT\n", packetNumber);
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseAEAT(strXML, xmlLength,  &parsedAEAT))
    {
        ERRORPRINT("%"PRId64" Failed to parse AEAT\n", packetNumber);
        rval = 1;
        goto out;
    }
    fHaveParsedAEAT = 1;

    // The app can do something with the parsed AEAT

out:
    if (fHaveParsedAEAT)
    {
        if (ATSC3LLSFreeParsedAEAT(&parsedAEAT))
        {
            ERRORPRINT("Failed to free parsed AEAT\n");
        }
    }

    return rval;
}

int32_t
ProcessCertificationDataTable(struct timespec *pPacketTimeUTC,
    uint64_t packetNumber, const ENATSC3LLSTableHeader *pLLSTableHeader,
    const uint8_t *pPayload, int32_t payloadLength, uint8_t sltTableVersion,
    ENATSC3Session *pSession)
{
    char strXML[16384];
    int32_t xmlLength;
    ATSC3LLSCDT_CertificationDataNode parsedCertificationData;
    int32_t fHaveParsedCertificationData = 0;
    struct timespec tMonoTimeout;
    int ret;
    int32_t rval = 0;

    ret = EnGunzipUnpipe((uint8_t *)pPayload, payloadLength,
        (unsigned char*)strXML, sizeof(strXML), &xmlLength);
    if (ret != Z_OK)
    {
        fprintf(stderr, "Failed to unzip: %d\n", ret);
        rval = 1;
        goto out;
    }

    if (ATSC3LLSParseCDT(strXML, xmlLength, &parsedCertificationData))
    {
        printf("%"PRId64" Failed to parse CertificationData\n",
            packetNumber);
        rval = 1;
        goto out;
    }

    fHaveParsedCertificationData = 1;

    clock_gettime(CLOCK_MONOTONIC, &tMonoTimeout);
    tMonoTimeout.tv_sec += 5;
    if (ENATSC3SSPVerifyAndLoadCertificationData(pSession->hSSP,
        strXML, xmlLength, &parsedCertificationData, &tMonoTimeout, 
        pSession->fSSPDontVerifySigningCertificate,
        pSession->fSSPIgnoreOCSPResponseExpired))
    {
        printf("Failed to verify and load certification data\n");
        rval = 1;
        goto out;
    }

out:
    if (fHaveParsedCertificationData)
    {
        if (ATSC3LLSFreeParsedCDT(&parsedCertificationData))
        {
            fprintf(stderr, "Failed to free parsed CDT\n");
        }
    }

    return rval;
}

void
SignedLLSTableCallback(struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    ENATSC3LLSTableHeader *pLLSTableHeader, const uint8_t *pPayload,
    int32_t payloadLength, int32_t fSigned, void *pUserData)
{
    ENATSC3Session *pSession = (ENATSC3Session *)pUserData;

    ProcessLLSTable(pPacketTimeUTC, packetNumber, pLLSTableHeader,
        pPayload, payloadLength, fSigned, pSession);
}

int32_t
ProcessLLSTable(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    ENATSC3LLSTableHeader *pLLSTableHeader, const uint8_t *pPayload, 
    int32_t payloadLength, int32_t fSigned, ENATSC3Session *pSession)
{
    struct LLSInfo *pLLS;
    uint32_t crc32;
    int32_t rval = 0;

    if (pLLSTableHeader->lls_group_id > numof(pSession->vLLS))
    {
        ERRORPRINT("Invalid group id %d\n", pLLSTableHeader->lls_group_id);
        rval = 1;
        goto out;
    }

    if (payloadLength < 8)
    {
        ERRORPRINT("Invalid lls payload length %d\n", payloadLength);
        rval = 1;
        goto out;
    }

    pLLS = pSession->vLLS + pLLSTableHeader->lls_group_id;

    switch (pLLSTableHeader->lls_table_id)
    {
        case 0x01: // SLT
            crc32 = ENUtils_CalcCRC32(pPayload, payloadLength);
            if (pLLS->slt.version != pLLSTableHeader->lls_table_version ||
                pLLS->slt.crc32 != crc32 || pLLS->slt.length != payloadLength)
            {
                if (pLLS->slt.version == -1)
                {
                    if (fSigned)
                    {
                        pLLS->slt.tSignedFirstObserved = *pPacketTimeUTC;
                    }
                    else
                    {
                        pLLS->slt.tUnsignedFirstObserved = *pPacketTimeUTC;
                    }
                }
                else if (pLLS->slt.version !=
                    pLLSTableHeader->lls_table_version)
                {
                    LOGPRINT("SLT group %d version changed from %d to %d\n",
                        pLLSTableHeader->lls_group_id,
                        pLLS->slt.version,
                        pLLSTableHeader->lls_table_version);
                }
                else
                {
                    LOGPRINT("SLT changed unexpectedly (%X ?= %X  %X ?= %X)\n",
                        pLLS->slt.crc32, crc32, pLLS->slt.length,
                        payloadLength);
                }

                rval = ProcessSLT(pPacketTimeUTC, packetNumber,
                    pLLSTableHeader, pPayload, payloadLength,
                    pLLSTableHeader->lls_table_version, pSession);
                if (!rval)
                {
                    pLLS->slt.version = pLLSTableHeader->lls_table_version;
                    pLLS->slt.crc32 = crc32;
                    pLLS->slt.length = payloadLength;
                    if (fSigned)
                    {
                        pLLS->slt.tSignedMostRecentlyAccepted =
                            *pPacketTimeUTC;
                    }
                    else
                    {
                        pLLS->slt.tUnsignedMostRecentlyAccepted =
                            *pPacketTimeUTC;
                    }
                }
            }
            if (fSigned)
            {
                pLLS->slt.tSignedMostRecentlyObserved = *pPacketTimeUTC;
            }
            else
            {
                pLLS->slt.tUnsignedMostRecentlyObserved = *pPacketTimeUTC;
            }
            break;
        case 0x04: // AEAT
            crc32 = ENUtils_CalcCRC32(pPayload, payloadLength);

            if (pLLS->aeat.version != pLLSTableHeader->lls_table_version ||
                pLLS->aeat.crc32 != crc32 || pLLS->aeat.length != payloadLength)
            {
                if (pLLS->aeat.version == -1)
                {
                    if (fSigned)
                    {
                        pLLS->aeat.tSignedFirstObserved = *pPacketTimeUTC;
                    }
                    else
                    {
                        pLLS->aeat.tUnsignedFirstObserved = *pPacketTimeUTC;
                    }
                }
                else if (pLLS->aeat.version !=
                    pLLSTableHeader->lls_table_version)
                {
                    LOGPRINT("AEAT group %d version changed from %d to %d\n",
                        pLLSTableHeader->lls_group_id,
                        pLLS->aeat.version,
                        pLLSTableHeader->lls_table_version);
                }
                else
                {
                    LOGPRINT("AEAT changed unexpectedly (%X ?= %X  %X ?= %X)\n",
                        pLLS->aeat.crc32, crc32, pLLS->aeat.length,
                        payloadLength);
                }

                rval = ProcessAEAT(pPacketTimeUTC, packetNumber,
                    pLLSTableHeader, pPayload, payloadLength,
                    pLLSTableHeader->lls_table_version, pSession);
                if (!rval)
                {
                    pLLS->aeat.version = pLLSTableHeader->lls_table_version;
                    pLLS->aeat.crc32 = crc32;
                    pLLS->aeat.length = payloadLength;
                    if (fSigned)
                    {
                        pLLS->aeat.tSignedMostRecentlyAccepted =
                            *pPacketTimeUTC;
                    }
                    else
                    {
                        pLLS->aeat.tUnsignedMostRecentlyAccepted =
                            *pPacketTimeUTC;
                    }
                }
            }
            if (fSigned)
            {
                pLLS->aeat.tSignedMostRecentlyObserved = *pPacketTimeUTC;
            }
            else
            {
                pLLS->aeat.tUnsignedMostRecentlyObserved = *pPacketTimeUTC;
            }
            break;
        case 0x06: // CDT
            crc32 = ENUtils_CalcCRC32(pPayload, payloadLength);

            if (pLLS->cdt.version != pLLSTableHeader->lls_table_version ||
                pLLS->cdt.crc32 != crc32 || pLLS->cdt.length != payloadLength)
            {
                if (pLLS->cdt.version == -1)
                {
                    pLLS->cdt.tFirstObserved = *pPacketTimeUTC;
                }
                else if (pLLS->cdt.version !=
                    pLLSTableHeader->lls_table_version)
                {
                    LOGPRINT("CDT group %d version changed from %d to %d\n",
                        pLLSTableHeader->lls_group_id,
                        pLLS->cdt.version,
                        pLLSTableHeader->lls_table_version);
                }
                else
                {
                    LOGPRINT("CDT changed unexpectedly (%X ?= %X  %X ?= %X)\n",
                        pLLS->cdt.crc32, crc32, pLLS->cdt.length,
                        payloadLength);
                }

                rval = ProcessCertificationDataTable(pPacketTimeUTC,
                    packetNumber, pLLSTableHeader, pPayload, payloadLength,
                    pLLSTableHeader->lls_table_version, pSession);
                if (!rval)
                {
                    pLLS->cdt.version = pLLSTableHeader->lls_table_version;
                    pLLS->cdt.crc32 = crc32;
                    pLLS->cdt.length = payloadLength;
                    pLLS->cdt.tMostRecentlyAccepted = *pPacketTimeUTC;
                }
            }
            pLLS->cdt.tMostRecentlyObserved = *pPacketTimeUTC;
            break;
        case 0xfe: // SMT
            crc32 = ENUtils_CalcCRC32(pPayload, payloadLength);

            if (pLLS->smt.version != pLLSTableHeader->lls_table_version ||
                pLLS->smt.crc32 != crc32 || pLLS->smt.length != payloadLength)
            {
                if (pLLS->smt.version == -1)
                {
                    pLLS->smt.tFirstObserved = *pPacketTimeUTC;
                }
                else if (pLLS->smt.version !=
                    pLLSTableHeader->lls_table_version)
                {
                    LOGPRINT("SMT group %d version changed from %d to %d\n",
                        pLLSTableHeader->lls_group_id,
                        pLLS->smt.version,
                        pLLSTableHeader->lls_table_version);
                }
                else
                {
                    LOGPRINT("SMT changed unexpectedly (%X ?= %X  %X ?= %X)\n",
                        pLLS->smt.crc32, crc32, pLLS->smt.length,
                        payloadLength);
                }

                if (ENATSC3SSPVerifySignedMultitableSignature(pSession->hSSP,
                    pPayload, payloadLength, 
                    pSession->fSSPDontVerifySigningCertificate))
                {
                    ERRORPRINT("Failed to verify SignedMultitable\n");
                    rval = 1;
                    goto out;
                }

                rval = ENATSC3LLSProcessSignedMultiTable(pPacketTimeUTC,
                   packetNumber, pLLSTableHeader, pPayload, payloadLength,
                   SignedLLSTableCallback, pSession);
                if (!rval)
                {
                    pLLS->smt.version = pLLSTableHeader->lls_table_version;
                    pLLS->smt.crc32 = crc32;
                    pLLS->smt.length = payloadLength;
                    pLLS->smt.tMostRecentlyAccepted = *pPacketTimeUTC;
                }
            }
            pLLS->smt.tMostRecentlyObserved = *pPacketTimeUTC;
            break;
        default:
            break;
    }
  
out:
    return rval;
}

int32_t
ProcessLLSUDPPacket(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    const uint8_t *pPayload, int32_t payloadLength, 
    ENATSC3Session *pSession)
{
    ENATSC3LLSTableHeader *pLLSTableHeader;
    int32_t rval = 0;

    if (payloadLength < 4)
    {
        ERRORPRINT("%"PRId64" LLS payload too short (%d)\n", packetNumber,
            payloadLength);
        rval = 1;
        goto out;
    }
  
    pLLSTableHeader = (ENATSC3LLSTableHeader *)pPayload;

    rval = ProcessLLSTable(pPacketTimeUTC, packetNumber, pLLSTableHeader, 
        pPayload + 4, payloadLength - 4, 0 /* unsigned */, pSession);
  
out:
    return rval;
}

int32_t
HandleUDPPacket(struct timespec *pPacketTimeUTC, uint64_t packetNumber, 
    int32_t plpId, uint32_t sourceAddress, uint32_t destinationAddress, 
    const uint8_t *pPacket, int32_t packetLength, 
    ENATSC3Session *pSession)
{
    int32_t sourcePort;
    int32_t destPort;
    int32_t length;
    int32_t checksum;
    ENInternetChecksumContext iccontext;
    uint32_t sum;
    int32_t rval = 0;

    if (packetLength < 8)
    {
        ERRORPRINT("UDP too small, dropping\n");
        rval = 1;
        goto out;
    }

    length = pPacket[4] << 8 | pPacket[5];

    if (packetLength != length || length < 8)
    {
        DEBUGPRINT_NOCONSOLE(
            "Not enough data for UDP packet: %d < %d\n", packetLength, 
            length);
        rval = 1;
        goto out;
    }

    sourcePort = pPacket[0] << 8 | pPacket[1];
    destPort = pPacket[2] << 8 | pPacket[3];
    checksum = pPacket[6] << 8 | pPacket[7];

    (void)sourcePort;

    // For IPv4 the checksum is optional. If the sender did not compute the
    // checksum then it should have set checksum to 0.
    if (checksum)
    {
        struct
        {
            uint32_t sourceAddress;
            uint32_t destinationAddress;
            uint8_t zero;
            uint8_t protocol;
            uint8_t lengthHi;
            uint8_t lengthLo;
        } __attribute__((packed)) pseudoHeader = 
        {
            .sourceAddress = htonl(sourceAddress),
            .destinationAddress = htonl(destinationAddress),
            .zero = 0,
            .protocol = 17,
            .lengthHi = packetLength >> 8,
            .lengthLo = packetLength & 0xff,
        };

        ENInternetChecksumInitialize(&iccontext);
        ENInternetChecksumCalc((uint8_t *)&pseudoHeader, sizeof(pseudoHeader), 
            &iccontext);
        ENInternetChecksumCalc(pPacket, packetLength, &iccontext);
        sum = ENInternetChecksumFinalize(&iccontext);

        if (sum  && !pSession->fIgnoreUDPChecksumErrors)
        {
            LOGPRINT("  UDP CHECKSUM ERROR: %x\n", sum);
        }
    }
    else
    {
        sum = 0;
    }

    if (sum == 0 || pSession->fIgnoreUDPChecksumErrors)
    {
        // Check for LLS
        if (destinationAddress == 0xE000173C && destPort == 4937)
        {
            // Skip over the UDP header in the first 8 bytes
            if (ProcessLLSUDPPacket(pPacketTimeUTC, packetNumber, pPacket + 8, 
                packetLength - 8, pSession))
            {
                LOGPRINT("%"PRId64" Failed to process UDP packet\n",
                    packetNumber);
            }
        }
        else
        {
            ServiceSession *pIter;

            pIter = pSession->pServiceSessions;
            while (pIter)
            {
                if (pIter->slsProtocol == ATSC3LLSSLTSLSProtocol_ROUTE)
                {
                    if (!EnIPFilterSetFilter(destinationAddress, destPort,
                        pIter->pIPFilterSet))
                    {
                        if (ENATSC3ROUTESessionProcessUDPPacket(
                            pIter->hROUTESession, pPacketTimeUTC, packetNumber, 
                            plpId, pPacket + 8, packetLength - 8))
                        {
                            LOGPRINT_NOCONSOLE(
                                "WARN: Failed to process UDP packet\n");
                            goto out;
                        }
                    }
                }
                else if (pIter->slsProtocol == ATSC3LLSSLTSLSProtocol_MMTP)
                {
                    if (!EnIPFilterSetFilter(destinationAddress, destPort,
                        pIter->pIPFilterSet))
                    {
                        if (ENATSC3MMTPSessionProcessUDPPacket(
                            pIter->hMMTPSession, pPacketTimeUTC, packetNumber, 
                            pPacket + 8, packetLength - 8))
                        {
                            LOGPRINT_NOCONSOLE(
                                "WARN: Failed to process UDP packet\n");
                            goto out;
                        }
                    }
                }

                pIter = pIter->pNext;
            }
        }
    }

out:
    return rval;
}

// Callback function from the IPV4 processor that a network packet has been
// received.
void
NetworkPacketCallback(int32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, uint32_t sourceAddress, uint32_t destinationAddress, 
    uint16_t protocol, const uint8_t *pPacket, int32_t sPacket, void *pCbData)
{
    ENATSC3Session *pSession = (ENATSC3Session *)pCbData;

    switch (protocol)
    {
        case 17:
            HandleUDPPacket(pPacketTimeUTC, packetNumber, plpId, sourceAddress, 
                destinationAddress, pPacket, sPacket, pSession);
            break;
        default:
            // ERRORPRINT("Unhandled protocol %d\n", protocol);
            break;
    }
}

// Callback function from the ALP processor notifying us that an IPV4 
// packet has been received.
void
IPV4PacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPayload, int32_t payloadLength, int32_t SID, void *pCbData)
{
    ENATSC3Session *pSession = (ENATSC3Session *)pCbData;

    ENATSC3IPV4ProcessIPV4Packet(pSession->hATSC3IPV4, plpId, pPacketTimeUTC,
        packetNumber, pPayload, payloadLength);

    return;
}

void
HandleLinkMappingTable(ENATSC3ALPLinkLayerSignalingHeader *pLLSHeader,
    uint32_t plpId, struct timespec *pPacketTimeUTC, const uint8_t *pPayload,
    int32_t payloadLength, ENATSC3Session *pSession)
{
    const uint8_t *pIter;
    int32_t nIter;
    uint32_t numPLPs;
    LMTMulticastEntry vLMTMulticasts[32];
    LMTMulticastEntry *pLMTMulticast;
    int32_t nLMTMulticasts = 0;
    int i;
    int j;

    if (plpId > numof(pSession->vLMTs))
    {
        ERRORPRINT("Unsupported PLP Id %d\n", plpId);
        goto out;
    }

    if (payloadLength > pSession->vLMTs[plpId].lmtBytesAllocated)
    {
        uint8_t *pTmp;

        pTmp = (uint8_t *)realloc(pSession->vLMTs[plpId].pLMT, 
            payloadLength + 1024);
        if (!pTmp)
        {
            ERRORPRINT("Out of memory\n");
            goto out;
        }
        pSession->vLMTs[plpId].pLMT = pTmp;
        pSession->vLMTs[plpId].lmtBytesAllocated = payloadLength + 1024;
    }

    if (pSession->vLMTs[plpId].lmtVersion == pLLSHeader->signaling_version)
    {
        if (pSession->vLMTs[plpId].lmtLength != payloadLength)
        {
            DEBUGPRINT_NOCONSOLE(
                "LMT length changed unexpectedly (%d != %d)\n", 
                pSession->vLMTs[plpId].lmtLength, payloadLength);
        }
        else if (!pSession->vLMTs[plpId].pLMT || 
            pSession->vLMTs[plpId].lmtBytesAllocated < payloadLength)
        {
            ERRORPRINT("SNH: Invalid stored PLP length\n");
        }
        else if (memcmp(pSession->vLMTs[plpId].pLMT + 5, pPayload + 5, 
            payloadLength - 5))
        {
            LOGPRINT_NOCONSOLE("LMT payload changed unexpectedly\n");
        }
        else
        {
            // Nothing has changed
            pSession->vLMTs[plpId].tMostRecentlyObserved = *pPacketTimeUTC;
            goto out;
        }
    }

    // Skip over signaling header
    pIter = pPayload + 5;
    nIter = payloadLength - 5;

    if (nIter <= 0)
    {
        DEBUGPRINT_NOCONSOLE("SHORT LMT PACKET\n");
        goto out;
    }
    numPLPs = (pIter[0] >> 2) + 1;
    pIter++;
    nIter--;

    pLMTMulticast = vLMTMulticasts;
    for (i = 0; i < numPLPs; i++)
    {
        uint32_t mcastPlpId;
        uint32_t numMulticasts;

        if (nIter < 2)
        {
            DEBUGPRINT_NOCONSOLE("SHORT LMT PACKET\n");
            goto out;
        }
        mcastPlpId = pIter[0] >> 2;
        numMulticasts = pIter[1];
        pIter += 2;
        nIter -= 2;

        for (j = 0; j < numMulticasts; j++)
        {
            uint32_t srcIPAddr;
            uint32_t destIPAddr;
            uint16_t srcUDPPort;
            uint16_t destUDPPort;
            uint8_t sidFlag;
            uint8_t compressedFlag;
            uint8_t SID;
            uint8_t contextId;

            if (nLMTMulticasts == numof(vLMTMulticasts))
            {
                LOGPRINT_NOCONSOLE("Too many multicasts in LMT\n");
                // This is fatal - stop processing all PLPs
                i = numPLPs;
                break;
            }

            if (nIter < 13)
            {
                DEBUGPRINT_NOCONSOLE("SHORT LMT PACKET\n");
                goto out;
            }

            srcIPAddr = (pIter[0] << 24) | (pIter[1] << 16) |
                (pIter[2] << 8) | pIter[3];
            destIPAddr = (pIter[4] << 24) | (pIter[5] << 16) |
                (pIter[6] << 8) | pIter[7];
            srcUDPPort = (pIter[8] << 8) | pIter[9];
            destUDPPort = (pIter[10] << 8) | pIter[11];
            sidFlag = pIter[12] & 0x80;
            compressedFlag = pIter[12] & 0x40;

            pIter += 13;
            nIter -= 13;

            if (sidFlag)
            {
                if (nIter < 1)
                {
                    DEBUGPRINT_NOCONSOLE("SHORT LMT PACKET\n");
                    goto out;
                }
                SID = pIter[0];
                pIter++;
                nIter--;
            }
            else
            {
                SID = 0;
            }

            if (compressedFlag)
            {
                if (nIter < 1)
                {
                    DEBUGPRINT_NOCONSOLE("SHORT LMT PACKET\n");
                    goto out;
                }
                contextId = pIter[0];
                pIter++;
                nIter--;
            }
            else
            {
                contextId = 0;
            }

            pLMTMulticast->lmtPLPId = plpId;
            pLMTMulticast->plpId = mcastPlpId;
            pLMTMulticast->srcAddr = srcIPAddr;
            pLMTMulticast->srcPort = srcUDPPort;
            pLMTMulticast->destAddr = destIPAddr;
            pLMTMulticast->destPort = destUDPPort;
            pLMTMulticast->sidFlag = sidFlag;
            pLMTMulticast->compressedFlag = compressedFlag;
            pLMTMulticast->contextId = contextId;
            pLMTMulticast->SID = SID;
            pLMTMulticast->tAdded = *pPacketTimeUTC;

            pLMTMulticast++;
            nLMTMulticasts++;
        }
    }

    // We only accept the LMT into the session structure if it has been
    // successfully parsed.
    pSession->vLMTs[plpId].lmtVersion = pLLSHeader->signaling_version;
    pSession->vLMTs[plpId].lmtLength = payloadLength;
    pSession->vLMTs[plpId].tFirstObserved = *pPacketTimeUTC;
    pSession->vLMTs[plpId].tMostRecentlyObserved = *pPacketTimeUTC;
    memcpy(pSession->vLMTs[plpId].pLMT, pPayload, payloadLength);

    if (pthread_mutex_lock(&pSession->linkLayerSignalingLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    RemoveMulticastEntriesForLMTPLP_Locked(pSession, plpId);
    for (i = 0, pLMTMulticast = vLMTMulticasts; 
         i < nLMTMulticasts; 
         i++, pLMTMulticast++)
    {
        if (AddMulticastEntry_Locked(pSession, pLMTMulticast))
        {
            ERRORPRINT("Failed to add multicast entry\n");
        }
    }

    LOGPRINT_NOCONSOLE("New LMT, regenerating PLP masks...\n");

    if (pthread_mutex_lock(&pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    ServiceSession *pServiceSessionIter;
    pServiceSessionIter = pSession->pServiceSessions;
    while (pServiceSessionIter)
    {
        int32_t fChanged;

        if (GetServicePLPsFromIPFilterSet_LinkLayerSignalingLocked(
            pServiceSessionIter, &fChanged))
        {
            LOGPRINT_NOCONSOLE(
                "WARN: Failed to get PLPs for service %d %d-%d\n",
                pServiceSessionIter->serviceId, 
                pServiceSessionIter->channelMajor, 
                pServiceSessionIter->channelMinor);
        }

        pServiceSessionIter = pServiceSessionIter->pNext;
    }

    if (pthread_mutex_unlock(&pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    if (pthread_mutex_unlock(&pSession->linkLayerSignalingLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    if (SetPLPsFromMasks(pSession))
    {
        ERRORPRINT("WARN: Failed to update PLPs from new LMT\n");
    }

out:
    return;
}

// Callback function from the ALP processor notifying us that a link-layer
// signaling packet has been received.
void
LinkLayerSignalingPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity, const uint8_t *pPayload, 
    int32_t payloadLength, void *pCbData)
{
    ENATSC3Session *pSession = (ENATSC3Session *)pCbData;
    ENATSC3ALPLinkLayerSignalingHeader llsheader;

    llsheader.signaling_type = pPayload[0];
    llsheader.signaling_type_extension = (pPayload[1] << 8) | pPayload[2];
    llsheader.signaling_version = pPayload[3];
    llsheader.signaling_format = (pPayload[4] >> 6) & 0x3;
    llsheader.signaling_encoding = (pPayload[4] >> 4) & 0x3;
    
    switch (llsheader.signaling_type)
    {
        case ENATSC3ALPSignalingType_LinkMappingTable:
            HandleLinkMappingTable(&llsheader, plpId, pPacketTimeUTC,
                pPayload, payloadLength, pSession);
            break;
        default:
            LOGPRINT_NOCONSOLE("Unhandled LLS signaling type %d\n",
                llsheader.signaling_type);
            break;
    }

    return;
}

// Callback function from the baseband packet processor notifying us that an
// ALP packet has been received.
void
ALPPacketCallback(uint32_t plpId, struct timespec *pPacketTimeUTC, 
    uint64_t packetNumber, int32_t fError, int32_t fDiscontinuity, 
    const uint8_t *pPacket, int32_t sPacket, void *pUserData)
{
    ENATSC3Session *pSession = (ENATSC3Session *)pUserData;

    if (ENATSC3ALPProcessALPPacket(pSession->hATSC3ALP, plpId,
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPacket, 
        sPacket))
    {
        DEBUGPRINT("Failed to process ALP packet\n");
    }
}

// Callback function from an ATSC3 Source notifying us that the PLPs 
// carrying LLS information has changed.
void
ENATSC3SourceLLSPLPsChangedCallback(uint64_t plpIdMask, 
    struct timespec *pPacketTimeUTC, uint64_t packetNumber, void *pUserData)
{
    ENATSC3Session *pSession = (ENATSC3Session *)pUserData;
    uint64_t mask;

    LOGPRINT("LLS PLPs changed to 0x%X\n", plpIdMask);

    if (pthread_mutex_lock(&pSession->plpsLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    mask = pSession->llsPLPIdMask;

    if (pthread_mutex_unlock(&pSession->plpsLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    if (mask != plpIdMask)
    {
        ServiceSession *pIter;

        pSession->llsPLPIdMask = plpIdMask;

        mask = plpIdMask;

        if (pthread_mutex_lock(&pSession->serviceSessionsListLock))
        {
            ERRORPRINT("Failed to lock\n");
        }

        pIter = pSession->pServiceSessions;
        while (pIter)
        {
            mask |= pIter->servicePLPIdMask;
            pIter = pIter->pNext;
        }

        if (pthread_mutex_unlock(&pSession->serviceSessionsListLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }

        SetPLPIds(pSession, mask);
    }
}

// Callback function from an ATSC3 Source notifying us that a baseband 
// packet has been received.
void 
ENATSC3SourceBasebandPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity, 
    const ENATCS3L1Time *pL1Time, const uint8_t *pPayload, int32_t sPayload, 
    void *pUserData)
{
    ENATSC3Session *pSession = (ENATSC3Session *)pUserData;

    if (!pSession->fXMLErrorHandlerRegistered)
    {
        xmlSetGenericErrorFunc(pSession, xmlGenericErrorLogHandler);
        pSession->fXMLErrorHandlerRegistered = 1;
    }

    if (ENATSC3BasebandProcessBasebandPacket(pSession->hATSC3Baseband, plpId, 
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pL1Time, 
        pPayload, sPayload))
    {
        ERRORPRINT("Failed to process baseband packet\n");
    }
}

// Callback function from an ATSC3 Source notifying us that an ALP
// packet has been received.
void 
ENATSC3SourceALPPacketCallback(uint32_t plpId,
    struct timespec *pPacketTimeUTC, uint64_t packetNumber,
    int32_t fError, int32_t fDiscontinuity, 
    const ENATCS3L1Time *pL1Time, const uint8_t *pPayload, 
    int32_t sPayload, void *pUserData)
{
    ENATSC3Session *pSession = (ENATSC3Session *)pUserData;

    if (!pSession->fXMLErrorHandlerRegistered)
    {
        xmlSetGenericErrorFunc(pSession, xmlGenericErrorLogHandler);
        pSession->fXMLErrorHandlerRegistered = 1;
    }

    if (ENATSC3ALPProcessALPPacket(pSession->hATSC3ALP, plpId,
        pPacketTimeUTC, packetNumber, fError, fDiscontinuity, pPayload,
        sPayload))
    {
        DEBUGPRINT("Failed to process ALP packet\n");
    }
}

ENStatus 
SSPLoadCertificatesCallback(ENATSC3SSPHandle_t hSSP, 
    STACK_OF(X509) *pCertificatesStack, void *pUserData)
{
    ENStatus rval = ENStatus_Success;

    // Load any certificates required by the SSP here

    return rval;
}

ENStatus
RunSession(ENATSC3Session *pSession)
{
    int32_t i;
    ENStatus rval = ENStatus_Success;
    
    pSession->pSource = NULL;
    pSession->hATSC3ALP = NULL;
    pSession->hSSP = NULL;
    pSession->llsPLPIdMask = 0;
    for (i = 0; i < numof(pSession->vLMTs); i++)
    {
        pSession->vLMTs[i].lmtVersion = -1;
    }

    for (i = 0; i < numof(pSession->vLLS); i++)
    {
        pSession->vLLS[i].slt.version = -1;
        pSession->vLLS[i].smt.version = -1;
    }

    for (i = 0; i < pSession->nLMTMulticastEntriesAllocated; i++)
    {
        pSession->pLMTMulticastEntries[i].fValid = 0;
    }

    for (i = 0; i < pSession->nSLTEntriesAllocated; i++)
    {
        pSession->pSLTEntries[i].fValid = 0;
    }

    if (ENATSC3SSPCreate(SSPLoadCertificatesCallback, NULL, &pSession->hSSP))
    {
        ERRORPRINT("Failed to create SSP\n");
        rval = ENStatus_Failure;
        goto out;
    }

    if (pSession->sourceType == ENATSC3SourceType_BasebandFile 
#if ENABLE_RZR
        || pSession->sourceType == ENATSC3SourceType_BasebandRZRSDK2
#endif
        )
    {
        if (ENATSC3BasebandOpen(&pSession->hATSC3Baseband))
        {
            ERRORPRINT("Failed to open baseband processor\n");
            rval = ENStatus_Failure;
            goto out;
        }

        if (ENATSC3BasebandRegisterCallbacks(pSession->hATSC3Baseband,
            ALPPacketCallback, pSession))
        {
            ERRORPRINT("Failed to register baseband callbacks\n");
            rval = ENStatus_Failure;
            goto out;
        }
    }

    if (pSession->sourceType == ENATSC3SourceType_BasebandFile 
#if ENABLE_RZR
        || pSession->sourceType == ENATSC3SourceType_BasebandRZRSDK2
#endif
#if ENABLE_HDHR
        || pSession->sourceType == ENATSC3SourceType_J288ALPHDHR
#endif
        )
    {
        if (ENATSC3ALPOpen(&pSession->hATSC3ALP))
        {
            ERRORPRINT("Failed to open ALP processor\n");
            rval = ENStatus_Failure;
            goto out;
        }

        if (ENATSC3ALPRegisterCallbacks(pSession->hATSC3ALP, 
            IPV4PacketCallback, LinkLayerSignalingPacketCallback, pSession))
        {
            ERRORPRINT("Failed to register callbacks\n");
            rval = ENStatus_Failure;
            goto out;
        }
    }

    if (ENATSC3IPV4Open(&pSession->hATSC3IPV4))
    {
        ERRORPRINT("Failed to open ATSC3IPV4 session\n");
        rval = ENStatus_Failure;
        goto out;
    }

    ENATSC3IPV4SetVerbosity(pSession->hATSC3IPV4, pSession->verbosity);

    if (ENATSC3IPV4RegisterCallbacks(pSession->hATSC3IPV4,
        NetworkPacketCallback, pSession))
    {
        ERRORPRINT("Failed to register ATSC3IPV4 callback\n");
        rval = ENStatus_Failure;
        goto out;
    }

    pSession->ifd = -1;
    switch (pSession->sourceType)
    {
        case ENATSC3SourceType_BasebandFile:
            pSession->ifd = open(pSession->pInFilename, O_RDONLY);
            if (pSession->ifd < 0)
            {
                fprintf(stderr, "Failed to open %s: %s\n", 
                    pSession->pInFilename, strerror(errno));
            }
            if (ENATSC3BasebandFileSourceOpen(pSession->deviceNumber, 
                pSession->ifd, 0, &pSession->pSource))
            {
                ERRORPRINT("Failed to open baseband file source\n");
                rval = ENStatus_Failure;
                goto out;
            }
            break;
#if ENABLE_RZR
        case ENATSC3SourceType_BasebandRZRSDK2:
            if (ENATSC3BasebandRZRSDK2SourceOpen(pSession->deviceNumber, 
                ENATSC3BasebandRZRSDK2SourceOperatingMode_ATSC3,
                pSession->frequency, &pSession->pSource))
            {
                ERRORPRINT("Failed to open baseband RZR source\n");
                rval = ENStatus_Failure;
                goto out;
            }
            break;
#endif
#if ENABLE_HDHR
        case ENATSC3SourceType_J288ALPHDHR:
            if (ENATSC3J288ALPHDHRSourceOpen(pSession->deviceNumber, 
                ENATSC3J288ALPHDHRSourceOperatingMode_ATSC3,
                pSession->frequency, &pSession->pSource))
            {
                ERRORPRINT("Failed to open J288 ALP HDHR source\n");
                rval = ENStatus_Failure;
                goto out;
            }
            break;
#endif
        default:
            ERRORPRINT("Unhandled source type %d\n", pSession->sourceType);
            rval = ENStatus_Failure;
            goto out;
    }

    if (pSession->pSource->RegisterCallbacks(pSession->pSource, 
        ENATSC3SourceLLSPLPsChangedCallback, 
        ENATSC3SourceBasebandPacketCallback, 
        ENATSC3SourceALPPacketCallback,
        pSession))
    {
        ERRORPRINT("Failed to register callbacks\n");
    }

    // We are going to kick things off by initializing PLP 0. For some devices
    // getting the PLP LLS/valid masks are very slow unless at least one valid
    // PLP is selected. While we can't guarantee that PLP 0 is valid, there is
    // a very good chance that if the emissions has any PLP it will have PLP 0.
    pSession->llsPLPIdMask = 1;
    if (SetPLPIds(pSession, 1))
    {
        ERRORPRINT("WARN: Failed to select PLP 0\n");
    }

    rval = pSession->pSource->Start(pSession->pSource, &pSession->fRunning);

out:
    pSession->llsPLPIdMask = 0;
    if (SetPLPIds(pSession, 0))
    {
        ERRORPRINT("Failed to remove PLPs\n");
    }

    if (pSession->pSource)
    {
        if (pSession->pSource->Close(pSession->pSource))
        {
            ERRORPRINT("Failed to close source\n");
        }
        pSession->pSource = NULL;
    }

    if (pSession->ifd >= 0)
    {
        close(pSession->ifd);
    }

    if (pthread_mutex_lock(&pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    // Stop all services 
    while (pSession->pServiceSessions)
    {
        ServiceSession *pServiceSession =
            pSession->pServiceSessions;

        // Unlink the service from the list of active services. 
        RemoveServiceSession_ServiceSessionListLocked(pSession, 
            pServiceSession);

        //  We need to temporarily unlock the serviceSessionList 
        // lock since it will be grabbed when the service  is stopped.
        if (pthread_mutex_unlock(&pSession->serviceSessionsListLock))
        {
            ERRORPRINT("Failed to unlock\n");
        }

        if (ENATSC3SessionServiceSessionStop(
            pServiceSession))
        {
            ERRORPRINT("Failed to stop serviceId %d %d-%d\n",
                pServiceSession->serviceId, pServiceSession->channelMajor,
                pServiceSession->channelMinor);
        }

        if (ENATSC3SessionServiceSessionFree(pServiceSession))
        {
            ERRORPRINT("Failed to free serviceId %d %d-%d\n",
                pServiceSession->serviceId, pServiceSession->channelMajor,
                pServiceSession->channelMinor);
        }

        if (pthread_mutex_lock(&pSession->serviceSessionsListLock))
        {
            ERRORPRINT("Failed to lock\n");
        }
    }

    if (pthread_mutex_unlock(&pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    if (pSession->hATSC3IPV4)
    {
        if (ENATSC3IPV4Close(pSession->hATSC3IPV4))
        {
            ERRORPRINT("Failed to close ATSC3IPV4 session\n");
        }
        pSession->hATSC3IPV4 = NULL;
    }

    if (pSession->hATSC3ALP)
    {
        if (ENATSC3ALPClose(pSession->hATSC3ALP))
        {
            ERRORPRINT("Failed to close ALP processor\n");
        }
        pSession->hATSC3ALP = NULL;
    }

    if (pSession->hATSC3Baseband)
    {
        if (ENATSC3BasebandClose(pSession->hATSC3Baseband))
        {
            ERRORPRINT("Failed to close baseband processor\n");
        }
        pSession->hATSC3Baseband = NULL;
    }

    if (pSession->hSSP)
    {
        if (ENATSC3SSPDestroy(pSession->hSSP))
        {
            ERRORPRINT("Failed to destroy SSP\n");
        }
    }

    return rval;
}

ENStatus
ENATSC3SessionCreate(ENATSC3SourceType sourceType, 
    int32_t deviceNumber, char *pInFilename, char *pOutPath,
    int32_t frequency, int32_t channelMajor, int32_t channelMinor, 
    int32_t fIgnoreUDPChecksumErrors, 
    int32_t fSSPDontVerifySigningCertificate,
    int32_t fSSPIgnoreOCSPResponseExpired,
    int32_t verbosity, ENATSC3SessionHandle_t *phSession)
{
    ENATSC3Session *pSession = NULL;
    int32_t fHavePLPsLock = 0;
    int32_t fHaveLinkLayerSignalingLock = 0;
    int32_t fHaveLLSLock = 0;
    int32_t fHaveServiceSessionsListLock = 0;
    int32_t fHaveIDsLock = 0;
    int32_t i;
    ENStatus rval = ENStatus_Success;

    LOGPRINT("Starting ATSC 3 session for rf %d vc %d-%d\n", frequency, 
        channelMajor, channelMinor);

    pSession = (ENATSC3Session *)calloc( sizeof(ENATSC3Session), 1);
    if (!pSession)
    {
        ERRORPRINT("Out of memory\n");
        rval = ENStatus_Failure;
        goto errout;
    }

    if (pthread_mutex_init(&pSession->plpsLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHavePLPsLock = 1;

    if (pthread_mutex_init(&pSession->linkLayerSignalingLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveLinkLayerSignalingLock = 1;

    if (pthread_mutex_init(&pSession->llsLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveLLSLock = 1;

    if (pthread_mutex_init(&pSession->serviceSessionsListLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveServiceSessionsListLock = 1;

    if (pthread_mutex_init(&pSession->idAllocator.idsLock, NULL))
    {
        ERRORPRINT("Failed to init lock\n");
        rval = ENStatus_Failure;
        goto errout;
    }
    fHaveIDsLock = 1;

    pSession->sourceType = sourceType;
    pSession->deviceNumber = deviceNumber;

    pSession->fSSPDontVerifySigningCertificate = 
        fSSPDontVerifySigningCertificate;
    pSession->fSSPIgnoreOCSPResponseExpired = 
        fSSPIgnoreOCSPResponseExpired;

    pSession->pInFilename = pInFilename;
    pSession->pOutPath = pOutPath;

    for (i = 0; i < numof(pSession->vLMTs); i++)
    {
        pSession->vLMTs[i].lmtVersion = -1;
        pSession->vLMTs[i].lmtLength = 0;
        pSession->vLMTs[i].pLMT = NULL;
        pSession->vLMTs[i].lmtBytesAllocated = 0;
    }
    pSession->frequency = frequency;
    pSession->requestedChannelMajor = channelMajor;
    pSession->requestedChannelMinor = channelMinor;
    pSession->fIgnoreUDPChecksumErrors = fIgnoreUDPChecksumErrors;
    pSession->verbosity = verbosity;

    for (i = 0; i < numof(pSession->vLLS); i++)
    {
        pSession->vLLS[i].slt.version = -1;
        pSession->vLLS[i].smt.version = -1;
    }

    pSession->pLMTMulticastEntries = NULL;
    pSession->nLMTMulticastEntriesAllocated = 0;

    pSession->pSLTEntries = NULL;
    pSession->nSLTEntriesAllocated = 0;

    *phSession = pSession;

out:
    return rval;

errout:
    if (pSession)
    {
        if (fHaveIDsLock)
        {
            if (pthread_mutex_destroy(&pSession->idAllocator.idsLock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        if (fHaveServiceSessionsListLock)
        {
            if (pthread_mutex_destroy(
                &pSession->serviceSessionsListLock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        if (fHaveLLSLock)
        {
            if (pthread_mutex_destroy(&pSession->llsLock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        if (fHaveLinkLayerSignalingLock)
        {
            if (pthread_mutex_destroy(&pSession->linkLayerSignalingLock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        if (fHavePLPsLock)
        {
            if (pthread_mutex_destroy(&pSession->plpsLock))
            {
                ERRORPRINT("Failed to destroy lock\n");
            }
        }

        free(pSession);
    }

    goto out;
}

ENStatus
ENATSC3SessionStart(ENATSC3SessionHandle_t hSession) 
{
    ENATSC3Session *pSession = (ENATSC3Session *)hSession;
    ENStatus rval = ENStatus_Success;

    pSession->fRunning = 1;

    RunSession(pSession);

    return rval;
}

ENStatus
ENATSC3SessionStop(ENATSC3SessionHandle_t hSession)
{
    ENATSC3Session *pSession = (ENATSC3Session *)hSession;
    ServiceSession *pIter;
    ENStatus rval = ENStatus_Success;

    pSession->fRunning = 0;

    if (pthread_mutex_lock(&pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to lock\n");
    }

    pIter = pSession->pServiceSessions;
    while (pIter)
    {
        if (ENATSC3SessionServiceSessionQuiesce(pIter))
        {
            ERRORPRINT("Failed to quiesce service %s\n", 
                pIter->strShortServiceName);
        }

        pIter = pIter->pNext;
    }

    if (pthread_mutex_unlock(&pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to unlock\n");
    }

    return rval;
}

ENStatus
ENATSC3SessionDestroy(ENATSC3SessionHandle_t hSession)
{
    ENATSC3Session *pSession = (ENATSC3Session *)hSession;
    int32_t i;
    ENStatus rval = ENStatus_Success;
    
    if (pSession->fRunning)
    {
        ERRORPRINT("Destroy called without stop!\n");
        if (ENATSC3SessionStop(pSession))
        {
            ERRORPRINT("Stop failed\n");
        }
    }

    for (i = 0; i < numof(pSession->vLMTs); i++)
    {
        free(pSession->vLMTs[i].pLMT);
    }

    free(pSession->pLMTMulticastEntries);

    free(pSession->pSLTEntries);

    if (pthread_cond_destroy(&pSession->muxStatsCond))
    {
        ERRORPRINT("Failed to destroy cond\n");
    }

    if (pthread_mutex_destroy(&pSession->idAllocator.idsLock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    if (pthread_mutex_destroy(&pSession->serviceSessionsListLock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    if (pthread_mutex_destroy(&pSession->llsLock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    if (pthread_mutex_destroy(&pSession->linkLayerSignalingLock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    if (pthread_mutex_destroy(&pSession->plpsLock))
    {
        ERRORPRINT("Failed to destroy lock\n");
    }

    if (pSession)
    {
        free(pSession);
    }

    return rval;
}

struct UserData
{
    ENATSC3SourceType sourceType;

    int32_t fInputEOF;
    int32_t verbose;
    int32_t fIgnoreUDPChecksumErrors;
    int32_t fSSPDontVerifySigningCertificate;
    int32_t fSSPIgnoreOCSPResponseExpired;

    int32_t frequency;

    // The channel the user requested
    int32_t channelMajor;
    int32_t channelMinor;

    // ATSC 3 to MPEGTS session
    ENATSC3SessionHandle_t hENATSC3;
} userdata;

int32_t
parseArgs(int argc, char **argv, char **ppInFilename, 
    ENATSC3SourceType *pSourceType, char **ppOutPath, 
    int32_t *pChannelMajor, int32_t *pChannelMinor, int32_t *pVerbose, 
    int32_t *pFIgnoreUDPChecksumErrors, int32_t *pFIgnoreOCSPExpired,
    int32_t *pFNoSigningCertVerify, int32_t *pFrequency)
{
    int32_t err = 0;
    int32_t fHaveMajorChannelNo = 0;
    int32_t fHaveMinorChannelNo = 0;
    int32_t fHaveFrequency = 0;
    int32_t i;

    *ppInFilename = NULL;
    *ppOutPath = NULL;
    *pSourceType = ENATSC3SourceType_Invalid;

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-bb"))
        {
            if (i + 1 < argc)
            {
                *ppInFilename = argv[i + 1];
                *pSourceType = ENATSC3SourceType_BasebandFile;
                i++;
            }
            else err = 1;
        }
#if ENABLE_RZR
        else if (!strcmp(argv[i], "-br"))
        {
            *pSourceType = ENATSC3SourceType_BasebandRZRSDK2;
        }
#endif
#if ENABLE_HDHR
        else if (!strcmp(argv[i], "-bh"))
        {
            *pSourceType = ENATSC3SourceType_J288ALPHDHR;
        }
#endif
        else if (!strcmp(argv[i], "-o"))
        {
            if (i + 1 < argc)
            {
                *ppOutPath = argv[i + 1];
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-major"))
        {
            if (i + 1 < argc)
            {
                *pChannelMajor = atol(argv[i + 1]);
                fHaveMajorChannelNo = 1;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-minor"))
        {
            if (i + 1 < argc)
            {
                *pChannelMinor = atol(argv[i + 1]);
                fHaveMinorChannelNo = 1;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-f"))
        {
            if (i + 1 < argc)
            {
                *pFrequency = atol(argv[i + 1]);
                fHaveFrequency = 1;
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-v"))
        {
            if (i + 1 < argc)
            {
                *pVerbose = atol(argv[i + 1]);
                i++;
            }
            else err = 1;
        }
        else if (!strcmp(argv[i], "-ignoreudpchecksum"))
        {
            *pFIgnoreUDPChecksumErrors = 1;
        }
        else if (!strcmp(argv[i], "-ignoreocspexpired"))
        {
            *pFIgnoreOCSPExpired = 1;
        }
        else if (!strcmp(argv[i], "-nosigningcertverify"))
        {
            *pFNoSigningCertVerify = 1;
        }
        else
        {
            fprintf(stderr, "Unrecognized option %s\n", argv[i]);
            err = 1;
        }
    }

    if (!*ppInFilename && *pSourceType == ENATSC3SourceType_BasebandFile)
    {
        err = 1;
    }

#if ENABLE_RZR || ENABLE_HDHR
    if ((*pSourceType == ENATSC3SourceType_BasebandRZRSDK2 ||
         *pSourceType == ENATSC3SourceType_J288ALPHDHR) && 
        !fHaveFrequency)
    {
        err = 1;
    }
#else
    // This is just to squash a variable-set-but-not-used compiler warning
    (void)fHaveFrequency;
#endif

    if (!*ppOutPath || !fHaveMajorChannelNo || !fHaveMinorChannelNo ||
        *pSourceType == ENATSC3SourceType_Invalid)
    {
        err = 1;
    }

    if (err)
    {
        fprintf(stderr, "Usage: %s -bb basebandfile | -br | -bh -o outpath -major majorchannelno -minor minorchannelno [-v verbosity] [-ignoreudpchecksum]  [-ignoreocspexpired] [-nosigningcertverify] [ -f frequency]\n", argv[0]);
    }

    return err;
}

void
sighandler(int signum)
{
    if (signum == SIGINT)
    {
        if (userdata.hENATSC3)
        {
            userdata.hENATSC3->fRunning = 0;
        }
    }
}

int
main(int argc, char **argv)
{
    int32_t fEnLogInitialized = 0;
    int32_t fXML2Initialized = 0;
    int32_t fENATSC3Initialized = 0;
    int32_t fENATSC3Started = 0;
    char *pInFilename = NULL;
    char *pOutPath = NULL;
    struct sigaction act;
    int rval = 0;

    memset(&userdata, 0, sizeof(userdata));

    // Graceful exit on ctrl-c
    act.sa_handler = sighandler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    sigaction(SIGINT, &act, NULL);

    if (EnLogInitialize(1))
    {
        fprintf(stderr, "Failed to initialize EnLog\n");
        rval = -1;
        goto out;
    }
    fEnLogInitialized = 1;

    xmlInitParser();
    fXML2Initialized = 1;

    if (parseArgs(argc, argv, &pInFilename, &userdata.sourceType,
        &pOutPath, &userdata.channelMajor, &userdata.channelMinor,
        &userdata.verbose, &userdata.fIgnoreUDPChecksumErrors, 
        &userdata.fSSPIgnoreOCSPResponseExpired,
        &userdata.fSSPDontVerifySigningCertificate, &userdata.frequency))
    {
        rval = 1;
        goto out;
    }

    if (ENATSC3Initialize(userdata.sourceType))
    {
        fprintf(stderr, "ENATSC3Initialize failed\n");
        rval = 1;
        goto out;
    }
    fENATSC3Initialized = 1;

    if (ENATSC3SessionCreate(userdata.sourceType,
        0 /* deviceNumber */, pInFilename, pOutPath,
        userdata.frequency, userdata.channelMajor, userdata.channelMinor,
        userdata.fIgnoreUDPChecksumErrors, 
        userdata.fSSPDontVerifySigningCertificate,
        userdata.fSSPIgnoreOCSPResponseExpired,
        userdata.verbose, 
        &userdata.hENATSC3))
    {
        fprintf(stderr, "Failed to create ATSC 3 to MPEG2-TS session\n");
        rval = 1;
        goto out;
    }

    if (ENATSC3SessionStart(userdata.hENATSC3))
    {
        fprintf(stderr, "Failed to start ATSC 3 session\n");
        rval = 1;
        goto out;
    }
    fENATSC3Started = 1;

out:
    if (fENATSC3Initialized)
    {
        if (userdata.hENATSC3)
        {
            if (fENATSC3Started)
            {
                if (ENATSC3SessionStop(userdata.hENATSC3))
                {
                    fprintf(stderr, "Failed to stop ATSC 3 session\n");
                }
            }

            if (ENATSC3SessionDestroy(userdata.hENATSC3))
            {
                fprintf(stderr, "Failed to close ATSC 3 session\n");
            }
        }

        if (ENATSC3Shutdown())
        {
            fprintf(stderr, "Failed to shut down ATSC 3\n");
        }
    }

    if (fXML2Initialized)
    {
        xmlCleanupCharEncodingHandlers();
        xmlDictCleanup();
        xmlCleanupParser();
    }

    if (fEnLogInitialized)
    {
        if (EnLogShutdown())
        {
            fprintf(stderr, "Failed to shut down EnLog\n");
        }
    }

    return rval;
}

